/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "context.h"
#include "utils.h"


Treeptr talloc(int N, int *parContext, int nParContext, int newContext);
void tdel(Treeptr node);
void fillTree(Treeptr node, int N, int order, int *thisContext, int thisValue);
void fillWholeTree(Treeptr root, int *data, int n, int k, int N, int position, int memStart, int order);
void printNodes(Treeptr node, int N);
int pruneSmallNodes(Treeptr node, int minCount);
int pruneTreeCutoff(Treeptr node, int *parCount, int parTotal, double K);
double pruneTreeIC(Treeptr node, double penalty);
double entropy(Treeptr node);
int getNpars(Treeptr node);
double scoreContext(int *context, int order, int val, Treeptr node);
double getLogProb(int *x, int n, int y);

/* a few globals */
static int nodeCount;

/* 
** data is passed in by row and is of dimension n by k.
** N is the number of chars in the alphabet (ie for DNA N=4).
** position is the index of the position being modeled (0 <= position < k).
** memStart is the index of the first position of the context, order is its order.
** K is the cutoff value for B&W-style cutoff pruning.
** nFinal will hold the number of nodes in the pruned tree.
** nPars will hold the number of parameters estimated.
** logProbs will be an N^order * N array of log probabilities (AA, AC, AG, ... TT)
** verbose is the verbose option.
** pruneMethod should be one of CUTOFF,AIC or BIC.
*/
void fitContext(int *data, int n, int k, int N, int position, int memStart, int order, double K, int *nFinal, int *nPars, double *logProbs, int verbose, int pruneMethod) {
  Treeptr root;
  int origContext = 0;
  int minCount = 2;
  int i,j,iTemp,lastNodeCount;
  double bestIC,penalty;
  int *thisContext;
  int nContexts,divisor;

  nodeCount = 0;
  root = talloc(N,&origContext,0,0);
  fillWholeTree(root,data,n,k,N,position,memStart,order);
  if (verbose)
    printf("Have %d nodes in initial tree\n",nodeCount);

  pruneSmallNodes(root,minCount);
  if (verbose)
    printf("Have %d nodes after pruning small ones (counts less than %d)\n",nodeCount,minCount);

  /*
  ** Do the pruning, according to one of three possible methods:
  **  cutoff (the context algorithm, as in Buhlmann & Wyner)
  **  AIC    (Akaike information criterion)
  **  BIC    (Bayesian information criterion)
  */
  lastNodeCount = nodeCount;

  if(pruneMethod==CUTOFF) {
    /* Buhlmann & Wyner's cutoff method */
    if(verbose)
      printf("  doing cutoff-based pruning\n");
    for(i=0; i<k; i++) {
      pruneTreeCutoff(root,root->count,root->total,K);
      if(verbose)
        printf("  step %d, %d nodes remain\n",i+1,nodeCount);
      if(nodeCount == lastNodeCount) {
        if(verbose)
          printf("  (pruning finished early)\n");
        break;
      } else {
        lastNodeCount = nodeCount;
      }
    }
  } else if( (pruneMethod==AIC) || (pruneMethod==BIC) ) {
    /* AIC or BIC differ only in the penalty term they use */
    if(verbose)
      printf("  doing penalized likelihood pruning\n");
    penalty = (pruneMethod==AIC) ? 2.0 : log(n);
    bestIC = pruneTreeIC(root,penalty);
    if(verbose)
      printf("  %d nodes remain\n",nodeCount);
  } else if (pruneMethod!=MM) {
    /* We shouldn't be in here */
    fprintf(stderr,"Bad pruneMethod (%d) passed\n",pruneMethod);
  }
  /* Note that if pruneMethod==MM then we don't actually do any pruning */


  if(verbose) {
    printf("Final tree:\n");
    printNodes(root,N);
  }

  /*
  ** Use the final tree to compute log probabilities
  ** for all contexts.
  */
  thisContext = (int *) myAlloc(N * sizeof(int));
  nContexts = (int) pow((double) N, (double) order);
  for(i=0; i<nContexts; i++) {
    iTemp = i;
    for(j=0; j<order; j++) {
      divisor = (int) pow((double) N, (double) order-(j+1));
      thisContext[j] = iTemp/divisor + 1;
      iTemp %= divisor;
    }

    for(k=0; k<N; k++)
      logProbs[N*i+k] = scoreContext(thisContext,order,k+1,root);
  }
  free(thisContext);

  /* record the number of nodes and parameters */
  *nFinal = nodeCount;
  *nPars  = getNpars(root);

  /* delete the tree */
  tdel(root);

  return;
}


/*
** Prune tree according to AIC/BIC
*/
double pruneTreeIC(Treeptr node, double penalty) {
  double *nodeIC,*childIC;
  double bestIC,bestNodeIC,bestChildIC,tempIC;
  int *kidCount;
  int i,j,nLostKids;

  if(node->nChildren == 0) {
    /* terminal node, just return the IC contribution */
    bestIC  = (2.0 * node->total * entropy(node));
    bestIC += penalty * (node->maxChildren -1);
  } else {
    /* non-terminal node, need to determine best IC */
    nodeIC   = (double *) myAlloc(node->maxChildren * sizeof(double));
    childIC  = (double *) myAlloc(node->maxChildren * sizeof(double));
    kidCount = (int *) myAlloc(node->maxChildren * sizeof(int));

    bestChildIC = 0;
    bestNodeIC  = penalty * (node->maxChildren -1);
    /*
    ** Account for "lost children" (children deleted by pruneSmallNodes)
    ** Otherwise the final value of bestNodeIC may be a little to small.
    */
    for(i=0; i<node->maxChildren; i++)
      kidCount[i] = 0;
    for(i=0; i<node->maxChildren; i++) {
      if(node->child[i] == NULL)
        continue;
      for(j=0; j<node->maxChildren; j++)
        kidCount[j] += node->child[i]->count[j];
    }
    for(i=0; i<node->maxChildren; i++) {
      if((nLostKids = node->count[i] - kidCount[i]) > 0)
        bestNodeIC -= 2.0 * nLostKids * log(node->count[i] / (double) node->total);
    }

    /*
    ** Compute nodeIC and childIC for each child.
    ** bestNodeIC is the sum of the minima of these.
    ** bestChildIC is the sum of the childICs
    */
    for(i=0; i<node->maxChildren; i++) {
      nodeIC[i] = 0;

      if(node->child[i] == NULL) {
        /*
        ** The ith node doesn't exist.  hack: childIC[i] is
        ** set to nodeIC[i]+1, to make sure nodeIC[i] is the min.
        */
        childIC[i] = nodeIC[i]+1;
        continue;
      }

      /* IC term for the node, corresponds to pruning the kid */
      for(j=0; j<node->maxChildren; j++) {
        if(node->child[i]->count[j] > 0) {
          tempIC  = node->child[i]->count[j];
          tempIC *= log(node->count[j] / (double) node->total);
          nodeIC[i] -= 2.0 * tempIC;
        }
      }
      /* optimal IC term for the child, computed recursively */
      childIC[i] = pruneTreeIC(node->child[i],penalty);
      bestNodeIC  += MIN(nodeIC[i],childIC[i]);
      bestChildIC += childIC[i];
    }


    if( (node->nChildren == node->maxChildren) && (bestChildIC < bestNodeIC) ) {
      /* No pruning, keep the full set of children */
      bestIC = bestChildIC;
    } else {
      /* prune as required */
      for(i=0; i<node->maxChildren; i++) {
        if(node->child[i] == NULL) {
          continue;
        } else if (nodeIC[i] < childIC[i]) {
          tdel(node->child[i]);
	  node->child[i] = NULL;
	  node->nChildren--;
        }
      }
      bestIC = bestNodeIC;
    }

    free(nodeIC);
    free(childIC);
    free(kidCount);
  }

  return(bestIC);
}


/*
** Computes the entropy of the distribution at a node (required by
** the function pruneTreeIC).
*/
double entropy(Treeptr node) {
  int i;
  double ent;

  if(node->total <= 0) {
    fprintf(stderr,"Request to compute entropy of an empty node, can't do\n");
    return(-1);
  } else {
    for(ent=0,i=0; i<node->maxChildren; i++)
      if(node->count[i] > 0)
        ent -= node->count[i] * log((double) node->count[i] / (double) node->total);
    return((double) ent / (double) node->total);
  }
}


/* 
** Allocate memory for a new node, returns a pointer to it.
*/
Treeptr talloc(int maxKids, int *parContext, int nParContext, int newContext) {
  int i;
  Treeptr newnode;

  /* increment the global counter of nodes */
  nodeCount++;

  newnode = (Treeptr)myAlloc(sizeof(Treenode));

  newnode->count = (int *)myAlloc(maxKids * sizeof(int));
  for(i=0; i<maxKids; i++)
    newnode->count[i] = 0;
  newnode->total = 0;

  newnode->nContext = (newContext == 0) ? nParContext : nParContext+1;
  newnode->context = (int *)myAlloc((newnode->nContext)*sizeof(int));
  for(i=0; i<nParContext; i++)
    newnode->context[i] = parContext[i];
  if(newContext > 0)
    newnode->context[nParContext] = newContext;

  newnode->nChildren = 0;
  newnode->maxChildren = maxKids;
  newnode->child = (Treeptr *)myAlloc(maxKids * sizeof(Treeptr));
  for(i=0; i<maxKids; i++)
    newnode->child[i] = NULL;

  return(newnode);
}


/* 
** Frees all memory for a node, (descends recursively to its
** children and frees them first).
*/
void tdel(Treeptr node) {
  int i;

  /* free the children first */
  for(i=0; i<node->maxChildren; i++)
    if(node->child[i] != NULL)
      tdel(node->child[i]);

  /* free the mem that was declared for this node */
  free(node->child);
  free(node->context);
  free(node->count);

  /* free the node itself and decrement the global node counter */
  free(node);
  nodeCount--;

  return;
}


/*
** Recursive function to fill out the initial tree.
*/
void fillTree(Treeptr node, int N, int order, int *thisContext, int thisValue) {
  int category;

  node->total++;
  node->count[thisValue-1]++;
  if(order > 0) {
    category = thisContext[order-1]-1;
    if(NULL == node->child[category]) {
      node->child[category] = talloc(N,node->context,node->nContext,category+1);
      node->nChildren++;
    }
    fillTree(node->child[category],N,order-1,thisContext,thisValue);
  }
  return;
}


/*
** Goes through all n data and enters each one to the tree using the 
** (recursive) function fillTree.
*/
void fillWholeTree(Treeptr root, int *data, int n, int k, int N, int position, int memStart, int order) {
  int i;
  int *thisContext;
  int thisValue;

  for(i=0; i<n; i++) {
    thisContext = data+(i*k+memStart);
    thisValue   = data[i*k+position];
    /* printf("about to enter ");
    for(j=order-1; j>=0; j--)
      printf("%d",thisContext[j]);
    printf(" in to the tree\n"); */
    fillTree(root,N,order,thisContext,thisValue);
  }
}


/* prints out all nodes of a tree (not just the terminal ones) */
void printNodes(Treeptr node, int N) {
  int i;

  for(i=0; i < node->nContext; i++)
    printf("%d",node->context[i]);
  printf("\t %5d ",node->total);
  for(i=0; i < N; i++)
    printf(" %4d",node->count[i]);
  printf("\n");
  if(node->nChildren > 0)
    for(i=0; i<N; i++)
      if(node->child[i] != NULL)
        printNodes(node->child[i],N);
  return;
}


/*
** Remove nodes with counts of less than minCount
*/
int pruneSmallNodes(Treeptr node, int minCount) {
  int i;

  if(node->total < minCount) {
    tdel(node);
    return(1);
  } else {
    for(i=0; i<node->maxChildren; i++) {
      if(node->child[i] != NULL) {
        if(pruneSmallNodes(node->child[i],minCount)) {;
	  node->child[i] = NULL;
	  node->nChildren--;
	}
      }
    }
    return(0);
  }
}


/*
** Implement step 2 of Buhlmann & Wyners' context algorithm.  Only operate on terminal
** nodes, ie nodes with no children.
** Returns 1 if there was at a terminal node that should be pruned, otherwise 0.
*/
int pruneTreeCutoff(Treeptr node, int *parCount, int parTotal, double K) {
  int i,doPrune;
  double delta,penalty;

  if(node->nChildren == 0) {
    /* printf("Looking at terminal node ");
    for(i=0; i<node->nContext; i++)
      printf("%d",node->context[i]);
    printf("\n"); */
 
    delta = node->total * log((double) parTotal / (double) node->total);
    for(i=0; i<node->maxChildren; i++) {
      if( (parCount[i] > 0) && (node->count[i] > 0) )
        delta += node->count[i] * log((double) node->count[i] / (double) parCount[i]);
    }

    penalty = K;

    doPrune = (delta - penalty < 0);
    /* printf("  delta = %g at ",delta);
    for(i=0; i<node->nContext; i++)
      printf("%d",node->context[i]);
    printf("\n");
    if(doPrune)
      printf("  pruning node\n"); */
    return(doPrune);
  } else {
    /* for(i=0; i<node->nContext; i++)
      printf("%d",node->context[i]);
    printf(" not a terminal node (%d children)\n",node->nChildren); */
    for(i=0; i<node->maxChildren; i++) {
      doPrune = 0;
      if(node->child[i] != NULL)
        doPrune = pruneTreeCutoff(node->child[i],node->count,node->total,K);
      if(doPrune) {
        tdel(node->child[i]);
	node->child[i] = NULL;
	node->nChildren--;
      }
    }
    return(0);
  }
}


/*
** Compute (recursively) the effective number of parameters of the tree.
*/
int getNpars(Treeptr node) {
  int i,nPars;

  nPars = 0;
  for(i=0; i < node->maxChildren; i++)
    if(node->child[i] != NULL)
      nPars += getNpars(node->child[i]);

  if(node->nChildren == 0) {
    return(node->maxChildren-1);
  } else if(node->nChildren < node->maxChildren) {
    return(nPars + node->maxChildren - 1);
  } else {
    return(nPars);
  }
}


/*
** Return the log probability of (1-based) val, given context *context of length
** order (written in forward-time order), according to the tree rooted at node.
** If the full context is not in the tree then we descend as far as possible.
** Zero counts get incremented to 1.
*/
double scoreContext(int *context, int order, int val, Treeptr node) {
  double ans;
  int useThisNode=0;
  int category;

  if(order > 0) {
    category = context[order-1]-1;
    if(NULL == node->child[category]) {
      /* can't descend any further, use the value from here */
      useThisNode = 1;
    } else {
      ans = scoreContext(context,order-1,val,node->child[category]);
    }
  } else {
    useThisNode = 1;
  }

  if(useThisNode)
    ans = getLogProb(node->count,node->maxChildren,val-1);

  return(ans);
}


/*
** Given array x of n integers, this function normalizes the array and returns
** the log probability of element y (0-based).  Zero counts get incremented to 1.
*/
double getLogProb(int *x, int n, int y) {
  double sum;
  int i;
  double numerator;

  for(sum=0,i=0; i<n; i++) {
    if(x[i] > 0)
      sum += x[i];
    else
      ++sum;
  }
  numerator = (x[y] > 0) ? (double) x[y] : 1.0;
  return(log(numerator/sum));
}
