/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "utils.h"
#include "context.h"

#define MAX_SEQ_NAME_LEN 100
#define N_DNA            4

enum charTypes{OTHER, BASE_A, BASE_C, BASE_G, BASE_T, NEXT, FILE_END};


/*
** The option structure
*/
typedef struct {
  char *inFile;        /* the input fasta file */
  int order;           /* order of the VLMC/MM */
  int probs;           /* print probabilities rather than log probabilities */
  int modelType;       /* type of model (MM, VLMC with BIC, VLMC with AIC, ...) */
  int upStrAvailable;  /* the number of upstream bases in the input file */
  int dnStrAvailable;  /* the number of dnstream bases in the input file */
  int upStr;           /* the number of upstream bases to use in the model */
  int dnStr;           /* the number of dnstream bases to use in the model */
  int verbose;         /* run in verbose mode */
} Option;

Option processArgs(int argc, char *argv[]);
int getModelType(char *s);
void checkOptions(Option *option);
int *readData(Option *option, int *n, int *k);
char getNextBase(FILE **seqFile);
int maxNseqs(char *fileName);
void writeHeader(Option *option);
void fitModel(int *spliceData, int n, int k, int thisPos, Option *option);

int main(int argc,char *argv[]) {
  Option option;   /* The option structure contains the running options */
  int *spliceData;
  int thisPos;
  int n,k;

  /*
  ** Process and check command-line options.
  */
  option = processArgs(argc,argv);
  checkOptions(&option);

  /*
  ** Read in the splice sites.
  */
  spliceData = readData(&option,&n,&k);

  /*
  for(i=0; i<n; i++) {
    for(thisPos=0; thisPos<k; thisPos++)
      fprintf(stderr,"%c",'A'+spliceData[i*k+thisPos]);
    fprintf(stderr,"\n");
  }
  */

  /*
  ** Write header of output.
  */
  writeHeader(&option);

  /*
  ** Fit a model to each position in the site.
  */
  for(thisPos=0; thisPos<k; thisPos++)
    fitModel(spliceData,n,k,thisPos,&option);

  exit(EXIT_SUCCESS);
}


/*
** Process command line arguments and return them
** in an Option structure.
*/
Option processArgs(int argc, char *argv[]) {
  Option opt;
  char *thisarg;
  int i;
  int gotInFile=0;
  int gotUpStr  = 0;
  int gotDnStr  = 0;
  int gotUpStrA = 0;
  int gotDnStrA = 0;

  /* 
  ** Set default values for options
  */
  opt.inFile         = NULL;
  opt.order          = 0;
  opt.probs          = 0;
  opt.modelType      = BIC;
  opt.upStrAvailable = 0;
  opt.dnStrAvailable = 0;
  opt.upStr          = 0;
  opt.dnStr          = 0;
  opt.verbose        = 0;

  for (i=1; i<argc; i++) {
    thisarg = argv[i];
    if(thisarg[0] == '-') {
      if (0==strcmp(thisarg+1,"v")) {
        opt.verbose = 1;
      } else if (0==strcmp(thisarg+1,"p")) {
        opt.probs = 1;
      } else if (0==strcmp(thisarg+1,"m")) {
	i++;
	if (i>=argc)
	  error((char *)"must provide a model type with option -m\n");
	else
          opt.modelType = getModelType(argv[i]);
      } else if (0==strcmp(thisarg+1,"o")) {
	i++;
	if (i>=argc)
	  error((char *)"must provide a non-negative integer with option -o\n");
	else
          opt.order = atoi(argv[i]);
      } else if (0==strcmp(thisarg+1,"5a")) {
	i++;
	if (i>=argc)
	  error((char *)"must provide a non-negative number with option -5a\n");
	else
          opt.upStrAvailable = atoi(argv[i]);
        gotUpStrA = 1;
      } else if (0==strcmp(thisarg+1,"3a")) {
        i++;
        if(i >= argc)
	  error((char *)"must provide a non-negative number with option -3a\n");
        else
          opt.dnStrAvailable = atoi(argv[i]);
        gotDnStrA = 1;
      } else if (0==strcmp(thisarg+1,"5")) {
        i++;
        if(i >= argc)
	  error((char *)"must provide a non-negative number with option -5\n");
        else
          opt.upStr = atoi(argv[i]);
        gotUpStr = 1;
      } else if (0==strcmp(thisarg+1,"3")) {
        i++;
        if(i >= argc)
	  error((char *)"must provide a non-negative number with option -3\n");
        else
          opt.dnStr = atoi(argv[i]);
        gotDnStr = 1;
      } else
        error((char *)"unknown option %s\n",thisarg);
    } else {
      if(!gotInFile) {
        opt.inFile = (char *) myAlloc((strlen(thisarg)+1) * sizeof(char));
        strcpy(opt.inFile,thisarg);
        gotInFile = 1;
      } else {
        fprintf(stderr,"Warning: ignoring command line option %s\n",thisarg);
      }
    }
  }

  /* make sure we got and input file and -5 and -3 options */
  if(!gotInFile)
    error((char *)"no input sequence file was specified.\n");
  if(!gotUpStr)
    error((char *)"must specify number of upstream bases with -5 option.\n");
  if(!gotDnStr)
    error((char *)"must specify number of downstream bases with -3 option.\n");

  if(!gotUpStrA) {
    fprintf(stderr,"Number of upstream bases available not specified, setting it to %d\n",opt.upStr);
    opt.upStrAvailable = opt.upStr;
  }
  if(!gotDnStrA) {
    fprintf(stderr,"Number of downstream bases available not specified, setting it to %d\n",opt.dnStr);
    opt.dnStrAvailable = opt.dnStr;
  }

  return opt;
}


/*
** Function used by processArgs to determine the requested model type.
*/
int getModelType(char *s) {
  if (0==strcmp(s,"cutoff"))
    return(CUTOFF);
  else if (0==strcmp(s,"aic"))
    return(AIC);
  else if (0==strcmp(s,"bic"))
    return(BIC);
  else if (0==strcmp(s,"mm"))
    return(MM);
  else
    error((char *)"Bad model type supplied with -m option.\n  Should be one of [cutoff, aic, bic, mm]\n");

  return(0);   /* won't be reached, keeps a fussy compiler quiet */
}


/*
** Check to make sure specified options are consistent
*/
void checkOptions(Option *option) {

  if(option->upStr < 0)
    error((char *)"option -5 must specify a non-negative integer\n",option->upStr);
  if(option->dnStr < 0)
    error((char *)"option -3 must specify a non-negative integer\n",option->dnStr);
  if(option->upStrAvailable < 0)
    error((char *)"option -5a must specify a non-negative integer\n",option->upStrAvailable);
  if(option->dnStrAvailable < 0)
    error((char *)"option -3a must specify a non-negative integer\n",option->dnStrAvailable);

  if(option->upStr > option->upStrAvailable)
    error((char *)"number of upstream bases shouldn't exceed the number available\n");
  if(option->dnStr > option->dnStrAvailable)
    error((char *)"number of downstream bases shouldn't exceed the number available\n");

  if(option->upStr + option->dnStr == 0)
    error((char *)"zero bases selected\n");

  if(option->order < 0)
    error((char *)"Model order can't be negative\n");

  return;
}


/*
** Read in the splice site data into an n x k matrix, stored by row.
** (A,C,G,T) are stored as (0,1,2,3)
*/
int *readData(Option *option, int *nSites, int *siteLen) {
  FILE *seqFile;
  int *spliceData;
  int *spliceDataPtr;
  int *thisSite;
  char seqName[MAX_SEQ_NAME_LEN];
  char c;
  int i,firstBase,lastBase,lenAvailable,len,siteOK,n,nGood,nBad;

  lenAvailable = option->upStrAvailable + option->dnStrAvailable;  /* total # bases in site */
  len          = option->upStr + option->dnStr;                    /* # bases we keep */
  firstBase    = option->upStrAvailable - option->upStr;           /* # bases to skip @ start */
  lastBase     = firstBase + len - 1;                              /* last base to use */

  n = maxNseqs(option->inFile);
  spliceData = (int *) myAlloc(n*len * sizeof(int));
  spliceDataPtr = spliceData;
  nGood = nBad = 0;

  thisSite = (int *) myAlloc(len * sizeof(int));

  if ((seqFile = fopen(option->inFile,"r")) == NULL)
    error((char *)"readData: can't open %s for read, exiting\n",option->inFile);

  c = getNextBase(&seqFile);
  if(c == FILE_END)
    error((char *)"file %s apparently contains no sequence\n",option->inFile);
  if(c != NEXT)
    error((char *)"bad format in %s (expecting fasta format)\n");

  while(c != FILE_END) {
    if(NULL == fgets(seqName,MAX_SEQ_NAME_LEN,seqFile))
      error((char *)"readData: problem reading %s\n",option->inFile);
    for(i=0, c = getNextBase(&seqFile), siteOK=1;
        ((c != NEXT) && (c != FILE_END));
        c = getNextBase(&seqFile), i++)
    {
      if((i >= firstBase) && (i <= lastBase) && siteOK) {
        if(c == OTHER)
          siteOK = 0;
        else
          thisSite[i-firstBase] = c;
      }
    }
    if(!siteOK) {
      fprintf(stderr,"site rejected because of non-ACGT bases: %s",seqName);
      ++nBad;
    } else if(i != lenAvailable) {
      error((char *)"site not of expected length: %s",seqName);
    } else {
      for(i=0; i<len; i++)
        *spliceDataPtr++ = thisSite[i];
      ++nGood;
    }
  }
  fclose(seqFile);
  free(thisSite);

  if(option->verbose) {
    fprintf(stderr,"Read %d sites\n",nGood+nBad);
    if(nBad > 0)
      fprintf(stderr,"  Rejectiong %d (because of non-ACGT bases)\n",nBad);
  }
  *nSites = nGood;
  *siteLen = len;

  return(spliceData);
}


/*
** Scans through the next characters of seqFile, skipping whitespace,
** digits, "*" and "-" till it reaches something different, in which
** case it returns....
**
** BASE_A   for 'A' or 'a'
** BASE_C   for 'C' or 'c'
** BASE_G   for 'G' or 'g'
** BASE_T   for 'T' or 't'
** NEXT     for '>'
** FILE_END for EOF
** OTHER    for anything else
*/
char getNextBase(FILE **seqFile) {
  char c=-1;

  c = fgetc(*seqFile);
  for( ; ; c = fgetc(*seqFile) ) {
    if((c == '>') || (c == EOF)) {
      break;
    } else if( isdigit(c) || isspace(c) || (c=='*') || (c=='-') ) {
      continue;
    } else {
      break;
    }
  }

  switch(c) {
    case 'A':
    case 'a':
      return(BASE_A);
    case 'C':
    case 'c':
      return(BASE_C);
    case 'G':
    case 'g':
      return(BASE_G);
    case 'T':
    case 't':
      return(BASE_T);
    case '>':
      return(NEXT);
    case EOF:
      return(FILE_END);
    default:
      return(OTHER);
  }
}

/*
** Simple function to count the number of lines beginning with '>' in
** the specified file, to give an upperbound to the number of valid sequences
** in a fasta file.
*/
int maxNseqs(char *fileName) {
  FILE *seqFile;
  char line[MAX_SEQ_NAME_LEN];
  int nSeq = 0;

  if ((seqFile = fopen(fileName,"r")) == NULL)
    error((char *)"openSeq: can't open %s for read, exiting\n",fileName);
  while(NULL != fgets(line,MAX_SEQ_NAME_LEN,seqFile)) {
    if(line[0] == '>')
      ++nSeq;
  }
  fclose(seqFile);
  return(nSeq);
}


/*
** Write the header of the output.
*/
void writeHeader(Option *option) {

  switch(option->modelType) {
    case MM:
      printf("MM\n");
    break;

    default:
      /*
      ** At the moment we print out VLMCs in the form of a MM.  So
      ** even though we're making a VLMC here we'll write MM in the 
      ** header.
      */
      printf("MM\n");
      /* printf("VLMC\n"); */
    break;
  }
  printf("order %d\n",option->order);
  printf("upstream %d\n",option->upStr);
  printf("downstream %d\n",option->dnStr);

  return;
}


/*
** Fit a model to a single location in the splice site.
*/
void fitModel(int *spliceData, int n, int k, int thisPos, Option *option) {
  int posOut,order,nContexts,i,j;
  int bwCutoff=10;
  int nFinal,nPars;
  double *logProbs;
  int veryVerbose=0;

  posOut = thisPos - option->upStr;
  order = (thisPos >= option->order) ? option->order : thisPos;
  nContexts = (int) pow((double) N_DNA,(double) order);
  logProbs = (double *) myAlloc(nContexts * N_DNA * sizeof(double));

  fitContext(spliceData,n,k,N_DNA,thisPos,thisPos-order,order,bwCutoff,
    &nFinal,&nPars,logProbs,veryVerbose,option->modelType);
  if(option->verbose)
    fprintf(stderr,"  Pos %3d: fitted tree has %d nodes (%d parameters)\n",posOut,nFinal,nPars);
  for(i=0; i<nContexts; i++) {
    printf("pos %2d    ",posOut);
    for(j=0; j<N_DNA; j++) {
      if(option->probs)
        printf("%1.4f ",exp(logProbs[i*N_DNA+j]));
      else
        printf("%1.4f ",logProbs[i*N_DNA+j]);
    }
    printf("\n");
  }
  printf("\n");

  free(logProbs);

  return;
}
