/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "utils.h"


/*
** The option structure
*/
typedef struct {
  char *inFile;        /* the input fasta file */
  int minLen;          /* minimum length considered */
  int maxLen;          /* maximum length considered */
  int verbose;         /* run in verbose mode */
  int lenOnly;         /* only print out the lengths, then quit */
} Option;


/*
** enums and defines.
*/ 
#define MAX_SEQ_NAME_LEN 100
#define MAX_MULTIPLIER   1.3
#define MIN_MULTIPLIER   0.7
#define ALMOST_ZERO      1e-6
enum charTypes{OTHER, BASE_A, BASE_C, BASE_G, BASE_T, NEXT, FILE_END};


/*
** Functions in this file.
*/
Option processArgs(int argc, char *argv[]);
void estGammaPars(double *gam_a, double *gam_b, double *gam_c, Option *option, int *minLen, int *maxLen);
void cmFromFasta(double *cm1, double *cm2, double *cm3, Option *option, int *minLen, int *maxLen);
char getNextBase(FILE **seqFile);
void printGammaDist(double gam_a, double gam_b, double gam_c, Option *option);
double dgamma(double x, double a, double b);

int main(int argc,char *argv[]) {
  Option option;       /* Option structure contains running options */
  double gam_a,gam_b,gam_c; /* the three gamma parameters */
  int maxLen,minLen;        /* Biggest/smallest lengths in observed data */

  option = processArgs(argc,argv);

  estGammaPars(&gam_a,&gam_b,&gam_c,&option,&minLen,&maxLen);

  /* If we only need to print out lengths then we're done */
  if(option.lenOnly)
    exit(EXIT_SUCCESS);

  /*
  ** If option.maxLen wasn't specified on the command line then
  ** set it to MAX_MULTIPLIER times the largest observed length.
  ** Similarly for option.minLen.
  */
  if(option.maxLen <= 0)
    option.maxLen = (int) MAX_MULTIPLIER * maxLen;
  if(option.minLen <= 0)
    option.minLen = (int) MIN_MULTIPLIER * minLen;

  /* Lengths less than gam_c are impossible */
  if(option.minLen < gam_c)
    option.minLen = (int) ceil(gam_c);

  /* Do a check to make sure maxLen > minLen */
  if(option.maxLen <= option.minLen)
    error((char *)"After automatically altering minLen and/or maxLen, ended up with maxLen <= minLen\n  Have a closer look at the data\n");

  /* print out the estimated gamma density */
  printGammaDist(gam_a,gam_b,gam_c,&option);

  if(option.verbose) {
    fprintf(stderr,"Analyzing lengths in %s\n",option.inFile);
    fprintf(stderr,"  gam_a = %f\n",gam_a);
    fprintf(stderr,"  gam_b = %f\n",gam_b);
    fprintf(stderr,"  gam_c = %f\n",gam_c);
    fprintf(stderr,"  minLen = %d, option.minLen = %d\n",minLen,option.minLen);
    fprintf(stderr,"  maxLen = %d, option.maxLen = %d\n",maxLen,option.maxLen);
  }

  exit(EXIT_SUCCESS);
}


/*
** Process command line arguments and return them
** in an Option structure.
*/
Option processArgs(int argc, char *argv[]) {
  Option opt;
  char *thisarg;
  int i;
  int gotInFile=0;

  /* 
  ** Set default values for options
  */
  opt.inFile         = NULL;
  opt.maxLen         = 0;
  opt.minLen         = 0;
  opt.verbose        = 0;
  opt.lenOnly        = 0;

  for (i=1; i<argc; i++) {
    thisarg = argv[i];
    if(thisarg[0] == '-') {
      if (0==strcmp(thisarg+1,"v")) {
        opt.verbose = 1;
      } else if (0==strcmp(thisarg+1,"l")) {
        opt.lenOnly = 1;
      } else if (0==strcmp(thisarg+1,"min")) {
	i++;
	if (i>=argc)
	  error((char *)"must provide a minimum length with option -min\n");
	else
          opt.minLen = atoi(argv[i]);
      } else if (0==strcmp(thisarg+1,"max")) {
	i++;
	if (i>=argc)
	  error((char *)"must provide a maximum length with option -max\n");
	else
          opt.maxLen = atoi(argv[i]);
      } else
        error((char *)"unknown option %s\n",thisarg);
    } else {
      if(!gotInFile) {
        opt.inFile = (char *) myAlloc((strlen(thisarg)+1) * sizeof(char));
        strcpy(opt.inFile,thisarg);
        gotInFile = 1;
      } else {
        fprintf(stderr,"Warning: ignoring command line option %s\n",thisarg);
      }
    }
  }

  /* make sure we got and input file and -5 and -3 options */
  if(!gotInFile)
    error((char *)"no input sequence file was specified.\n");

  /* make sure maxLen > minLen, if they were specified */
  if((opt.maxLen > 0) && (opt.minLen > 0) && (opt.minLen >= opt.maxLen))
    error((char *)"minLen must be less than maxLen\n");

  return opt;
}


/*
** Function to get the first three central moments of the length distribution
** of sequences in the fasta file in option->inFile.  We do a two-pass approach
** over the sequence file, computing the mean on the first pass and then the
** other moments on the next passes.  It could be done in one pass, but computing
** the third (non-central) moment could lead to overflow.
*/
void cmFromFasta(double *cm1, double *cm2, double *cm3, Option *option, int *minLen, int *maxLen) {
  FILE *seqFile;
  char seqName[MAX_SEQ_NAME_LEN];
  char c;
  int len;
  int nSeqs=0;
  double diff;

  /* open sequence file */
  if ((seqFile = fopen(option->inFile,"r")) == NULL)
    error((char *)"cmFromFasta: can't open %s for read (pass 1), exiting\n",option->inFile);

  /* read the first base or character */
  c = getNextBase(&seqFile);
  if(c == FILE_END)
    error((char *)"file %s apparently contains no sequence\n",option->inFile);
  if(c != NEXT)
    error((char *)"bad format in %s (expecting fasta format)\n");

  /* read through all sequences */
  nSeqs = 0;
  *cm1 = 0;
  *maxLen = 0;
  *maxLen = 0;
  while(c != FILE_END) {
    if(NULL == fgets(seqName,MAX_SEQ_NAME_LEN,seqFile))
      error((char *)"readData: problem reading %s\n",option->inFile);
    for(len=0, c = getNextBase(&seqFile);
        ((c != NEXT) && (c != FILE_END));
        c = getNextBase(&seqFile), ++len);
    *cm1 += len;
    if(option->lenOnly) printf("%d\n",len);
    *maxLen = (len > *maxLen) ? len : *maxLen;
    *minLen = ((len < *minLen) || (*minLen == 0)) ? len : *minLen;
    ++nSeqs;
  }
  *cm1 /= (double) nSeqs;

  /* close the file */
  fclose(seqFile);


  /*
  ** Now do a second pass over the file, this time getting
  ** the higher central moments.
  */

  /* open sequence file */
  if ((seqFile = fopen(option->inFile,"r")) == NULL)
    error((char *)"cmFromFasta: can't open %s for read (pass 2), exiting\n",option->inFile);

  /* read the first base or character */
  c = getNextBase(&seqFile);
  if(c == FILE_END)
    error((char *)"file %s apparently contains no sequence\n",option->inFile);
  if(c != NEXT)
    error((char *)"bad format in %s (expecting fasta format)\n");

  /* read through all sequences */
  *cm2 = *cm3 = 0;
  while(c != FILE_END) {
    if(NULL == fgets(seqName,MAX_SEQ_NAME_LEN,seqFile))
      error((char *)"readData: problem reading %s\n",option->inFile);
    for(len=0, c = getNextBase(&seqFile);
        ((c != NEXT) && (c != FILE_END));
        c = getNextBase(&seqFile), ++len);
    diff = (double) len - *cm1;
    *cm2 += pow(diff, (double) 2.0);
    *cm3 += pow(diff, (double) 3.0);
  }
  *cm2 /= (double) nSeqs;
  *cm3 /= (double) nSeqs;

  /* close the file */
  fclose(seqFile);

  return;
}


/*
** Scans through the next characters of seqFile, skipping whitespace,
** digits, "*" and "-" till it reaches something different, in which
** case it returns....
**
** BASE_A   for 'A' or 'a'
** BASE_C   for 'C' or 'c'
** BASE_G   for 'G' or 'g'
** BASE_T   for 'T' or 't'
** NEXT     for '>'
** FILE_END for EOF
** OTHER    for anything else
*/
char getNextBase(FILE **seqFile) {
  char c=-1;

  c = fgetc(*seqFile);
  for( ; ; c = fgetc(*seqFile) ) {
    if((c == '>') || (c == EOF)) {
      break;
    } else if( isdigit(c) || isspace(c) || (c=='*') || (c=='-') ) {
      continue;
    } else {
      break;
    }
  }

  switch(c) {
    case 'A':
    case 'a':
      return(BASE_A);
    case 'C':
    case 'c':
      return(BASE_C);
    case 'G':
    case 'g':
      return(BASE_G);
    case 'T':
    case 't':
      return(BASE_T);
    case '>':
      return(NEXT);
    case EOF:
      return(FILE_END);
    default:
      return(OTHER);
  }
}


/*
** Estimates the three gamma parameters by the method of moments.
*/
void estGammaPars(double *gam_a, double *gam_b, double *gam_c, Option *option, int *minLen, int *maxLen) {
  double cm1,cm2,cm3;

  cmFromFasta(&cm1,&cm2,&cm3,option,minLen,maxLen);

  *gam_a = 4 * pow(cm2,(double)3.0) / pow(cm3,(double)2.0);
  *gam_b = 0.5 * cm3 / cm2;
  *gam_c = cm1 - (2 * pow(cm2,(double)2.0) / cm3);

  return;
}

/*
** Writes the program's output to the screen.  The format is simply the
** number of entries, followed by the gamma density out to that number, with
** entries for values less than option->minLen truncated to ALMOST_ZERO.
*/
void printGammaDist(double gam_a, double gam_b, double gam_c, Option *option) {
  int i;

  printf("%d\n",option->maxLen);

  for(i=1; i < option->minLen; i++)
    printf("%g\n",ALMOST_ZERO);
  for(i=option->minLen; i <= option->maxLen; i++)
    printf("%g\n",dgamma(((double)i-gam_c),gam_a,gam_b));

  return;
}


/*
** The gamma density function.
** x,a and b must all be positive.
*/
double dgamma(double x, double a, double b) {
  x = x/b;
  return(pow(x,a-1) * exp(-x) / (b * gamma(a)));
}
