#!/usr/local/bin/perl -w

########################################################################
########################################################################
##                                                                    ##
## Phat - a system for predicting gene structures in DNA.             ##
## Copyright (C) 2000 Simon Cawley and Anthony Wirth                  ##
##                                                                    ##
## This program is free software; you can redistribute it and/or      ##
## modify it under the terms of the GNU General Public License        ##
## as published by the Free Software Foundation; either version 2     ##
## of the License, or (at your option) any later version.             ##
##                                                                    ##
## This program is distributed in the hope that it will be useful,    ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of     ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ##
## GNU General Public License for more details.                       ##
##                                                                    ##
## You should have received a copy of the GNU General Public License  ##
## along with this program; if not, write to the Free Software        ##
## Foundation, Inc.,                                                  ##
## 59 Temple Place - Suite 330,                                       ##
## Boston, MA  02111-1307, USA.                                       ##
##                                                                    ##
########################################################################
########################################################################

##
## buildModels.pl estimates gene-finding parameters from genbank files.
##
## Usage: buildModels.pl -o outDir [options] genbankFiles
##
## Option  Value    Effect
## -v               Run in verbose mode.
## -k               Keep temporary data files on exit.
## -f      fileList Take names of input Genbank sequences from fileList.
## -o      outDir   Write output files to outDir.
## -ends            Keep intergenic sequences from ends of each sequence.
## -5up    n        Use n bp for upstream of 5' splice sites.
## -5dn    n        Use n bp for downstream of 5' splice sites.
## -3up    n        Use n bp for upstream of 3' splice sites.
## -3dn    n        Use n bp for downstream of 3' splice sites.
## -cdsOrd n        Set order of coding exon model to n.
## -inOrd  n        Set order of intron model to n.
## -igOrd  n        Set order of intergene exon model to n.
##

use Getopt::Long;
use File::Basename;
use Bio::Seq;
use seq2freq;
use buildStopcodonModel;
use buildIstateModel;
use buildTransMat;
use Context::buildSiteModel;
use Lengths::buildLengthModel;


##
## The required executables.
##
$exe{"seqExtr"} = "./seqExtract.pl";
$exe{"fitmm"}   = "./Context/fitmm";
$exe{"estLen"}  = "./Lengths/estLengths";
foreach $exe (keys %exe) {
  die "Can't find $exe{$exe} (did you make it?)\n"    if(! -e $exe{$exe});
  die "Can't execute $exe{$exe}\n" if (! -x $exe{$exe});
}


##
## Command line options.
##
$option{"k"}       = 0;      # Keep the temporary files that seqExtract.pl makes.
$option{"v"}       = 0;      # verbose
$option{"f"}       = "";     # File from which to read entries.
$option{"o"}       = "";     # Directory for output.
$option{"ends"}    = 0;      # Keep intergenic regions from "ends" of each sequence.
$option{"5up"}     = "3";    # Number of bases to collect upstream of donors.
$option{"5dn"}     = "10";   # Number of bases to collect downstream of donors.
$option{"donord"}  = "2";    # Order for donor model.
$option{"dontype"} = "bic";  # Selection criteria for donor model.
$option{"3up"}     = "20";   # Number of bases to collect upstream of acceptors.
$option{"3dn"}     = "3";    # Number of bases to collect downstream of acceptors.
$option{"accord"}  = "2";    # Order for acceptor model.
$option{"acctype"} = "bic";  # Selection criteria for acceptor model.
$option{"cdsOrd"}  = "5";    # Order of coding exon model.
$option{"inOrd"}   = "2";    # Order of intron model.
$option{"igOrd"}   = "2";    # Order of intergenic model.
if (!(&GetOptions(\%option,"k","v","ends","f=s","o=s",
  "5up=s","5dn=s","donord=s","dontype=s",
  "3up=s","3dn=s","accord=s","acctype=s",
  "cdsOrd=s","inOrd=s","igOrd=s"))) {
  &usage();
  die("Error: problem processing command-line options\n");
}


##
## Names for the output data files.
##
$outFile{"cdsFreq"} = "exon.freq.dat";
$outFile{"inFreq"}  = "intron.freq.dat";
$outFile{"igFreq"}  = "intergene.freq.dat";
$outFile{"don"}     = "donor.dat";
$outFile{"acc"}     = "acceptor.dat";
$outFile{"cdsSing"} = "exon.len.sing.dat";
$outFile{"cdsInit"} = "exon.len.init.dat";
$outFile{"cdsIntr"} = "exon.len.intl.dat";
$outFile{"cdsTerm"} = "exon.len.term.dat";
$outFile{"stop"}    = "stopprobs.dat";
$outFile{"inLen"}   = "intron.len.dat";
$outFile{"igLen"}   = "intergene.len.dat";
$outFile{"trans"}   = "transitions.dat";
$outFile{"init"}    = "initprobs.dat";


##
## Extensions for the files seqExtract.pl creates, so they
## can be deleted later.
##
@seqExtFile = (
  "acc",
  "cds",
  "cds.init",
  "cds.intr",
  "cds.sing",
  "cds.term",
  "don",
  "intergene",
  "intron");


##
## Make sure model types for acceptors and donors are valid.
##
if(! ($option{"dontype"} =~ /^(mm|aic|bic)$/)) {
  print STDERR "Error: Invalid type specified with -dontype option: ",$option{"dontype"},"\n";
} elsif(! ($option{"acctype"} =~ /^(mm|aic|bic)$/)) {
  print STDERR "Error: Invalid type specified with -acctype option: ",$option{"acctype"},"\n";
} elsif($option{"donord"} < 0) {
  print STDERR "Error: -donord must specify a nonnegative integer\n";
} elsif($option{"accord"} < 0) {
  print STDERR "Error: -accord must specify a nonnegative integer\n";
}

##
## Make sure output directory is specified, and create it.
##
if($option{"o"} eq "") {
  &usage;
  print STDERR "Must specify directory for writing output files with -o option\n";
  die;
}
$outDir = $option{"o"};
$outDir .= "/" if(! ($outDir =~ /\/$/));
@pathParts = split('\/',$outDir);
$base = pop(@pathParts);
die "Can't create $outDir\n" if ((! -e $outDir) && system("mkdir $outDir"));


##
## Get the list of sequences to process.
##
if($option{"f"} ne "") {
  ## read entries from specified file
  if(scalar(@ARGV) > 0) {
    &usage;
    print STDERR "Too many command line arguments, ignored the extras\n";
  }
  $inFile = $option{"f"};
  open(SEQLIST,$inFile) || die "$0: can't open $inFile for read\n";

  @seq = ();
  while(<SEQLIST>) {
    push(@seq,$1) if(/^\s*(\S+)\s*$/);
  }
} else {
  ## read entries from command line
  @seq = @ARGV;
}
if(scalar(@seq) == 0) {
  &usage();
  die "Error: no input sequences specified\n"
}


##
## Run seqExtract.pl to parse the genbank files and extract the 
## relevant sequences.
##
$extractCommand  = $exe{"seqExtr"} . " -o $base" . " ";
$extractCommand .= " -5up " . $option{"5up"} . " -5dn " . $option{"5dn"} . " ";
$extractCommand .= " -3up " . $option{"3up"} . " -3dn " . $option{"3dn"} . " ";
$extractCommand .= " -v " if($option{"v"});
$extractCommand .= " -ends " if($option{"ends"});
$extractCommand .= join(" ",@seq);
die "Problem extracting sequences.  Call was:\n $extractCommand\n" if system($extractCommand);



##
## Build data files for the stationary Markov models.
##
&seq2freq("$base.cds",      $option{"cdsOrd"}, 3, $outDir.$outFile{"cdsFreq"});
&seq2freq("$base.intron",   $option{"inOrd"},  1, $outDir.$outFile{"inFreq"});
&seq2freq("$base.intergene",$option{"igOrd"},  1, $outDir.$outFile{"igFreq"});

##
## Build data files for the non-stationary Markov models.
##
&buildSiteModel("$base.don",$outDir.$outFile{"don"},$option{"5up"},$option{"5dn"},$exe{"fitmm"},$option{"donord"},$option{"dontype"},$option{"v"});
&buildSiteModel("$base.acc",$outDir.$outFile{"acc"},$option{"3up"},$option{"3dn"},$exe{"fitmm"},$option{"accord"},$option{"acctype"},$option{"v"});

##
## Build data files for generalized state length distributions
##
&buildLengthModel("$base.cds.sing",$outDir.$outFile{"cdsSing"},$exe{"estLen"},$option{"v"});
&buildLengthModel("$base.cds.init",$outDir.$outFile{"cdsInit"},$exe{"estLen"},$option{"v"});
&buildLengthModel("$base.cds.intr",$outDir.$outFile{"cdsIntr"},$exe{"estLen"},$option{"v"});
&buildLengthModel("$base.cds.term",$outDir.$outFile{"cdsTerm"},$exe{"estLen"},$option{"v"});

##
## The stop-codon model is a special case.
##
&buildStopcodonModel("$base.cds",$outDir.$outFile{"stop"},$option{"v"});

##
## Estimate parameters of the geometric I-state distributions.
##
&buildIstateModel("$base.intron",,$outDir.$outFile{"inLen"},$option{"v"});
&buildIstateModel("$base.intergene",$outDir.$outFile{"igLen"},$option{"v"});

##
## Estimate transition parameters based on the 
##
&buildTransMat($outDir.$outFile{"trans"},$outDir.$outFile{"init"},$option{"v"},@seq);


##
## Delete the temporary files seqExtract.pl created, unless
## the user requested not to.
##
if(!$option{"k"}) {
  foreach $file (@seqExtFile) {
    unlink "$base.$file";
  }
}

exit(0);






#
# Usage function - will print out commented lines at the top of this 
# file until it encounters a non-commented line.  (Skips the first.)
#
sub usage {
  open(MYSELF,$0) || die "$0: can't open $0 to print usage!\n";
  while(<MYSELF>) {
    if(/^\#!/) {
      next;
    } elsif (/^\s*\#(.*)$/) {
      eval "print STDERR \"$1\n\"";
    } else {
      last;
    }
  }
  close MYSELF;
}
