#!/usr/local/bin/perl -w

########################################################################
########################################################################
##                                                                    ##
## Phat - a system for predicting gene structures in DNA.             ##
## Copyright (C) 2000 Simon Cawley and Anthony Wirth                  ##
##                                                                    ##
## This program is free software; you can redistribute it and/or      ##
## modify it under the terms of the GNU General Public License        ##
## as published by the Free Software Foundation; either version 2     ##
## of the License, or (at your option) any later version.             ##
##                                                                    ##
## This program is distributed in the hope that it will be useful,    ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of     ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ##
## GNU General Public License for more details.                       ##
##                                                                    ##
## You should have received a copy of the GNU General Public License  ##
## along with this program; if not, write to the Free Software        ##
## Foundation, Inc.,                                                  ##
## 59 Temple Place - Suite 330,                                       ##
## Boston, MA  02111-1307, USA.                                       ##
##                                                                    ##
########################################################################
########################################################################

use Bio::Seq;
use Bio::SeqIO;

##
## Makes the transition matrix based on the Genbank files specified in @seq.
##
sub buildTransMat {
  my($transFile) = shift;
  my($initFile) = shift;
  my($verbose) = shift;
  my(@seq)     = @_;
  my(@genes,$nSeqFiles,$seqFileIndex,$seqFile,$seq,$seqIndex,$seqobj,$seqId);
  my($feat,$geneName,@subFeatures,@exonLens,$subFeat,$thisGene,$nCDS,$nIG,$nIN);
  my($endLastExon,$seqLen,$thisIG,$thisCDS,$thisIN,$count);

  ##
  ## Read CDS entries from @seq, store exon lengths in @genes.
  ##
  @genes = ();
  $nSeqFiles = scalar(@seq);
  $seqFileIndex = 0;
  $nCDS = $nIG = $nIN = 0;
  foreach $seqFile (@seq) {
    $seqFileIndex++;
    print "Processing file $seqFileIndex of $nSeqFiles: $seqFile\n" if($verbose);

    $seq = Bio::SeqIO->new(-file => $seqFile, -format => 'genbank');
    $seqIndex = 0;
    while($seqobj = $seq->next_seq()) {
      $seqIndex++;
      # make sure there is sequence to analyze
      if(! defined($seqobj->seq)) {
        print STDERR "  Sequence entry $seqIndex is empty in $seqFile\n";
        next;
      }
      $seqLen = length($seqobj->seq);
      $thisIG = $thisCDS = $thisIN = 0;
      $seqId  = (defined($seqobj->display_id())) ? $seqobj->display_id() : "Unknown";
      ($seqId) = ($1) if($seqId =~ /^(\S+)\s+/);
     
      foreach $feat ( $seqobj->top_SeqFeatures() ) {
        if( $feat->primary_tag =~ /(CDS)|(cds)/ ) {
          $geneName = defined($feat->source_tag) ? $feat->source_tag : "Unknown";
          @subFeatures = $feat->sub_SeqFeature;
          @subFeatures = ($feat) if(scalar(@subFeatures) == 0);
          @exonLens = ();
          $endLastExon = -1;
          foreach $subFeat (@subFeatures) {
            push(@exonLens,length($subFeat->seq->seq));
            $thisCDS += length($subFeat->seq->seq);
            if($endLastExon > -1) {
              $thisIN += $subFeat->start - $endLastExon - 1;
            }
            $endLastExon = $subFeat->end;
          }
          @exonLens = reverse(@exonLens) if($subFeatures[0]->strand == -1);
          $thisGene = {
            "seqId"   => $seqId,
            "source"  => $geneName,
            "exonLen" => [@exonLens],
          };
          push(@genes,$thisGene);
        }
      }
      $thisIG = $seqLen - $thisCDS - $thisIN;
      # print "Sequence $seqId: $seqLen = $thisIG + $thisCDS + $thisIN (IG + CDS + IN)\n";
    }
    $nIG  += $thisIG;
    $nCDS += $thisCDS;
    $nIN  += $thisIN;
  }

  ## Some code for debugging - prints out all exon lengths.
  if($verbose) {
    foreach $gene (@genes) {
      printf ("%15s: %s\n", $gene->{"seqId"},join(" ",@{$gene->{"exonLen"}}));
    }
  }

  open(TRANSFILE,">$transFile") || die "Can't open $transFile for write.\n";
  $count = &writeTransitions(\*TRANSFILE,@genes);
  close(TRANSFILE);

  open(INITFILE,">$initFile") || die "Can't open $initFile for write.\n";
  &writeInit(\*INITFILE,$count,$nIG,$nIN);
  close(INITFILE);

  return;
}


##
## Writes out and returns the transition matrix for the HMM.
##
sub writeTransitions {
  my($fh) = shift;
  my(@genes) = @_;
  my(@states,%count,$state1,$state2,$gene,@exons,$next,$prev,$nExons);

  # initialization of counts
  @states = ("N","I0","I1","I2");
  foreach $state1 (@states) {
    foreach $state2 (@states) {
      $count{$state1}->{$state2} = 0;
    }
  }

  foreach $gene (@genes) {
    @exons   = @{$gene->{"exonLen"}};
    if(scalar(@exons) == 1) {
      $count{$states[0]}->{$states[0]} += 1;
    } else {
      $next = $exons[0] % 3;
      $count{$states[0]}->{"I$next"} += 1;
      $prev = $next;
      $nExons = scalar(@exons);
      for($i=1; $i < ($nExons-1); $i++) {
        $next = ($prev + $exons[$i]) % 3;
        $count{"I$prev"}->{"I$next"} += 1;
        $prev = $next;
      }
      $count{"I$prev"}->{$states[0]} += 1;
    }
  }

  ##
  ## print out the counts
  ##
  # line 1
  printf $fh ("%7s %7s %7s %7s %7s %7s %7s %7s\n",
    "","IG","I0F","I1F","I2F","I0B","I1B","I2B");
  # line 2
  printf $fh ("%7s %7d %7s %7s %7s %7s %7s %7s\n",
    "IG",
    $count{$states[0]}->{$states[0]},
    $count{$states[0]}->{$states[1]}/2,
    $count{$states[0]}->{$states[2]}/2,
    $count{$states[0]}->{$states[3]}/2,
    $count{$states[0]}->{$states[1]}/2,
    $count{$states[0]}->{$states[2]}/2,
    $count{$states[0]}->{$states[3]}/2);
  # the next three lines
  foreach $state1 (@states[1..3]) {
    printf $fh ("%7s ",$state1 . "F");
    foreach $state2 (@states) {
      printf $fh ("%7d ",$count{$state1}->{$state2});
    }
    printf $fh ("%7d %7d %7d\n",0,0,0);
  }
  # the last three lines
  foreach $state1 (@states[1..3]) {
    printf $fh ("%7s ",$state1 . "B");
    printf $fh ("%7d ",$count{$state1}->{$states[0]});
    printf $fh ("%7d %7d %7d ",0,0,0);
    foreach $state2 (@states[1..3]) {
      printf $fh ("%7d ",$count{$state1}->{$state2});
    }
    printf $fh "\n";
  }
  
  return(\%count);
}


##
## Writes out and returns the transition matrix for the HMM.
##
sub writeInit {
  my($fh)     = shift;
  my($count) = shift;
  my($nIG)    = shift;
  my($nIN)    = shift;
  my($tot,$n,@states,$state,%weight,$totWeight,$state1,$state2);

  $n = $nIG + $nIN;

  @states = ("N","I0","I1","I2");
  foreach $state (@states[1..3]) {
    $weight{$state} = 0;
  }
  $totWeight = 0;
  foreach $state1 (@states[1..3]) {
    foreach $state2 (@states) {
      $weight{$state1} += $count->{$state1}->{$state2};
    }
    $totWeight += $weight{$state1};
  }

  # Set initial probability of intergene to frequency of intergene...
  printf $fh ("%6f intergene\n",$nIG/$n);
  foreach $state (@states[1..3]) {
    printf $fh ("%6f %s\n", ($weight{$state}/$totWeight) * 0.5 * ($nIN/$n), $state . "F");
  }
  foreach $state (@states[1..3]) {
    printf $fh ("%6f %s\n", ($weight{$state}/$totWeight) * 0.5 * ($nIN/$n), $state . "B");
  }

  return;
}

1
