#!/usr/local/bin/perl -w

########################################################################
########################################################################
##                                                                    ##
## Phat - a system for predicting gene structures in DNA.             ##
## Copyright (C) 2000 Simon Cawley and Anthony Wirth                  ##
##                                                                    ##
## This program is free software; you can redistribute it and/or      ##
## modify it under the terms of the GNU General Public License        ##
## as published by the Free Software Foundation; either version 2     ##
## of the License, or (at your option) any later version.             ##
##                                                                    ##
## This program is distributed in the hope that it will be useful,    ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of     ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ##
## GNU General Public License for more details.                       ##
##                                                                    ##
## You should have received a copy of the GNU General Public License  ##
## along with this program; if not, write to the Free Software        ##
## Foundation, Inc.,                                                  ##
## 59 Temple Place - Suite 330,                                       ##
## Boston, MA  02111-1307, USA.                                       ##
##                                                                    ##
########################################################################
########################################################################

use Bio::Seq;
use Bio::SeqIO;

#
# Write out base frequencies of a general sequence.
#
sub seq2freq {
  my($inFile)  = shift;
  my($order)   = shift;
  my($period)  = shift;
  my($outFile) = shift;

  my($nContext,$p,$c,$b,$freq,$dna,$n,@dna,$i,$context);
  my($done,$j,$base,$thisBit,$ok,$unused,$inSeq,$seqObj);

  ##
  ## Initialize the frequencies.
  ##
  $nContext = 4 ** $order;
  for ($p=0; $p < $period; $p++) {
    for ($c=0; $c < $nContext; $c++) {
      for ($b=0; $b < 4; $b++) {
        $freq[$p][$c][$b] = 0;
      }
    }
  }


  ##
  ## Run through all sequences, incrementing the counts.
  ##
  $unused = 0;
  $inSeq = Bio::SeqIO->new(-file => $inFile);
  while($seqObj = $inSeq->next_seq()) {
    $dna = $seqObj->seq;
    $n = length($dna);
    $dna =~ tr/A/0/;
    $dna =~ tr/C/1/;
    $dna =~ tr/G/2/;
    $dna =~ tr/T/3/;
    @dna = split("",$dna);

    # Get the initial context, possibly scanning ahead to find 
    # a stretch with only ACGT.
    $i = $done = 0;
    while( (!$done) && (($i+$order) < $n) ) {
      $thisBit = substr($dna,$i,$order+1);
      $ok = ($thisBit =~ /^[0123]+$/);
      if($ok) {
        for($j=0, $context=0; $j<$order; $i++,$j++) {
          $context *= 4;
          $context += $dna[$i];
        }
	$done = 1;
      } else {
        $i++;
      }
    }
    $p = $order % $period if($period > 0);

    #print "Initial context is $context, next base is ",$dna[$i],"\n";
    for( ; $i < $n; $p++,$i++) {
      $thisBit = substr($dna,$i-$order,$order+1);
      $ok = ($thisBit =~ /^[0123]+$/);
      $unused++ if (!$ok);

      $p %= $period if($period > 0);
      $base = $dna[$i];
      $freq[$p][$context][$base] += 1 if($ok);
      $context *= 4;
      $context += $base if($ok);
      $context %= $nContext;
    }
  }


  ##
  ## Write out frequencies
  ##
  open(OUTFILE,">$outFile") || die("seq2freq: can't open $outFile for write\n");
  for ($p=0; $p < $period; $p++) {
    printf OUTFILE ("Order %d, Period %d\n",$order,$p+1);
    for ($c=0; $c < $nContext; $c++) {
      for ($b=0; $b < 4; $b++) {
        printf OUTFILE "%7d ",$freq[$p][$c][$b];
      }
      printf OUTFILE "\n";
    }
  }
  close(OUTFILE);
  
  return($unused);
}

1
