#!/usr/local/bin/perl -w

########################################################################
########################################################################
##                                                                    ##
## Phat - a system for predicting gene structures in DNA.             ##
## Copyright (C) 2000 Simon Cawley and Anthony Wirth                  ##
##                                                                    ##
## This program is free software; you can redistribute it and/or      ##
## modify it under the terms of the GNU General Public License        ##
## as published by the Free Software Foundation; either version 2     ##
## of the License, or (at your option) any later version.             ##
##                                                                    ##
## This program is distributed in the hope that it will be useful,    ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of     ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ##
## GNU General Public License for more details.                       ##
##                                                                    ##
## You should have received a copy of the GNU General Public License  ##
## along with this program; if not, write to the Free Software        ##
## Foundation, Inc.,                                                  ##
## 59 Temple Place - Suite 330,                                       ##
## Boston, MA  02111-1307, USA.                                       ##
##                                                                    ##
########################################################################
########################################################################

##
## seqExtract.pl
## 
## Usage
##  seqExtract.pl [options] <-o basename> <sequence files>
##
## Option  Argument    Effect
## -v      none        Run in verbose mode.
## -q      none        Run in quiet mode.
## -f      filename    Read names of sequence files to analyze from filename.
## -o      basename    Use basename as base for any output files.
## -ends   none        Keep intergenic regions at the ends (otherwise dumped)
## -5up    number      # bases to collect upstream of donors      (default 3)
## -5dn    number      # bases to collect downstream of donors    (default 10)
## -3up    number      # bases to collect upstream of acceptors   (default 20)
## -3dn    number      # bases to collect downstream of acceptors (default 3)
##

use Getopt::Long;
use Bio::Seq;
use Bio::SeqIO;

##
## Command line options
##
$option{"v"}    = 0;     # verbose
$option{"q"}    = 0;     # verbose
$option{"f"}    = "";    # File from which to read entries
$option{"o"}    = "";    # Base name for output file
$option{"ends"} = 0;     # Should we keep intergenic regions at the ends?
$option{"5up"}  = "3";   # Number of bases to collect upstream of donors
$option{"5dn"}  = "10";  # Number of bases to collect downstream of donors
$option{"3up"}  = "20";  # Number of bases to collect upstream of acceptors
$option{"3dn"}  = "3";   # Number of bases to collect downstream of acceptors
if (!(&GetOptions(\%option,"v","q","f=s","ends","o=s","5up=s","5dn=s","3up=s","3dn=s"))) {
  &usage();
  die("Error: problem processing command-line options\n");
}

##
## Make sure output file is specified.
##
if($option{"o"} eq "") {
  &usage;
  print STDERR "Must define base name for output files with -o option\n";
  die;
}
$outFile = $option{"o"};

##
## Get the list of sequences to process.
##
if($option{"f"} ne "") {
  ## read entries from specified file
  if(scalar(@ARGV) > 0) {
    &usage;
    print STDERR "Too many command line arguments, ignored the extras\n";
  }
  $inFile = $option{"f"};
  open(SEQLIST,$inFile) || die "$0: can't open $inFile for read\n";

  @seq = ();
  while(<SEQLIST>) {
    push(@seq,$1) if(/^\s*(\S+)\s*$/);
  }
} else {
  ## read entries from command line
  @seq = @ARGV;
}
die "Error: no input sequences specified\n" if(scalar(@seq) == 0);



$cdsSingOut   = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.cds.sing");
$cdsInitOut   = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.cds.init");
$cdsIntrOut   = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.cds.intr");
$cdsTermOut   = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.cds.term");
$cdsOut       = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.cds");
$intronOut    = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.intron");
$donOut       = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.don");
$accOut       = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.acc");
$intergeneOut = Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$outFile.intergene");

#
# Read all CDS entries from all the specified sequences.
# Store them all in @genes
#
$nSeqFiles = scalar(@seq);
$seqFileIndex = 0;
$nBasesCdsTot = $nBasesIntronTot = $nBasesIGTot = 0;
foreach $seqFile (@seq) {
  $seqFileIndex++;
  print "Processing file $seqFileIndex of $nSeqFiles: $seqFile\n" if($option{"v"});

  $seq = Bio::SeqIO->new(-file => $seqFile, -format => 'genbank');
  $seqIndex = 0;
  $nCds = $nCdsSing = $nCdsInit = $nCdsIntr = $nCdsTerm = $nIntron = $nDon = $nAcc = $nIG = 0;
  while($seqobj = $seq->next_seq()) {
    $igLo = 1;
    @ig = ();
    $seqIndex++;
    # make sure there is sequence to analyze
    if(! defined($seqobj->seq)) {
      print STDERR "  Sequence entry $seqIndex is empty in $seqFile\n";
      next;
    }
    $seqLen = length($seqobj->seq);
    $seqId  = (defined($seqobj->display_id())) ? $seqobj->display_id() : "Unknown";
    ($seqId) = ($seqId =~ /^(\S+)\s*/);
     
    $nBasesCds = $nBasesIntron = $nBasesIG = 0;
    $cdsIgnored = 0;
    foreach $feat ( $seqobj->top_SeqFeatures() ) {
      if( $feat->primary_tag =~ /(CDS)|(cds)/ ) {
        @subFeatures = $feat->sub_SeqFeature;
        @subFeatures = ($feat) if(scalar(@subFeatures) == 0);
        $revStrand = ($subFeatures[0]->strand == -1) ? 1 : 0;
        $n           = scalar(@subFeatures);
        if($feat->has_tag('gene')) {
          ($cdsName)   = $feat->each_tag_value('gene');
        } elsif($feat->has_tag('product')) {
          ($cdsName)   = $feat->each_tag_value('product');
        } else {
          $cdsName = "Unknown";
        }

        ##
        ## add on the intergenic region just completed
        ##
        $igHi = $subFeatures[0]->start-1;
        if(($igLo > 0) && ($igLo <= $igHi) && ($igHi <= $seqLen)) {
          $thisIG = Bio::PrimarySeq->new();
          $thisIG->seq($seqobj->subseq($igLo,$igHi));
          $thisIG->display_id($seqId);
          push(@ig,$thisIG);
        }
        $igLo = $subFeatures[$n-1]->end+1;
        
        @introns  = ();
        @don      = ();
        @acc      = ();
        @subFeatures = reverse(@subFeatures) if($revStrand);
        $nextCds  = $subFeatures[0]->seq;
        $nextCds->display_id($cdsName);
        @cds      = ($nextCds);
        @cdsCoord = { "start" => $subFeatures[0]->start, "end" => $subFeatures[0]->end,};
        $len      = length($cds[0]->seq);
        for($i=1; $i < $n; $i++) {
          ##
          ## Add on cds and cds coordinates
          ##
          $nextCds  = $subFeatures[$i]->seq;
          $nextCds->display_id($cdsName);
          push(@cds,$nextCds);
          push(@cdsCoord,{ "start" => $subFeatures[$i]->start, "end" => $subFeatures[$i]->end,});

          ##
          ## Add on an intron 
          ##
          $thisIntron = Bio::PrimarySeq->new();
          if($revStrand) {
            $intronHi = $cdsCoord[$i-1]->{"start"} - 1;
            $intronLo = $cdsCoord[$i]->{"end"} + 1;
          } else {
            $intronLo = $cdsCoord[$i-1]->{"end"} + 1;
            $intronHi = $cdsCoord[$i]->{"start"} - 1;
          }
          $thisIntron->seq($seqobj->subseq($intronLo,$intronHi));
          $thisIntron = $thisIntron->revcom if($revStrand);
          $thisIntron->display_id($cdsName);
          push(@introns,$thisIntron);

          ##
          ## Add on a donor
          ##
          if($revStrand) {
            $donLo = $intronHi - $option{"5dn"} + 1;
            $donHi = $intronHi + $option{"5up"};
          } else {
            $donLo = $intronLo - $option{"5up"};
            $donHi = $intronLo + $option{"5dn"} - 1;
          }
          if( ($donLo > 0) && ($donHi <= $seqLen)) {
            $thisDon = Bio::PrimarySeq->new();
            $thisDon->seq($seqobj->subseq($donLo,$donHi));
            $thisDon = $thisDon->revcom if($revStrand);
            $thisDon->display_id($cdsName);
            push(@don,$thisDon);
          }

          ##
          ## Add on an acceptor
          ##
          if($revStrand) {
            $accLo = $intronLo - $option{"3dn"};
            $accHi = $intronLo + $option{"3up"} - 1;
          } else {
            $accLo = $intronHi - $option{"3up"} + 1;
            $accHi = $intronHi + $option{"3dn"};
          }
          if( ($accLo > 0) && ($accHi <= $seqLen)) {
            $thisAcc = Bio::PrimarySeq->new();
            $thisAcc->seq($seqobj->subseq($accLo,$accHi));
            $thisAcc = $thisAcc->revcom if($revStrand);
            $thisAcc->display_id($cdsName);
            push(@acc,$thisAcc);
          }

          $len += length($cds[$i]->seq);
        }


        ##
        ## Only write out the CDS-related info if it isn't
        ## labelled "IGNORE"
        ##
        if(! ($feat->primary_tag =~ /IGNORE/ )) {
          ##
          ## Write out coding sequences to the appropriate cds files.
          ##
          if($n == 1) {
            $cdsSingOut->write_seq($cds[0]);
            $cdsOut->write_seq($cds[0]);
            $nCdsSing++;
          } else {
            $thisCDS = Bio::PrimarySeq->new();
            $thisCDS->display_id($cdsName);
            $cdsInitOut->write_seq($cds[0]);
            $thisCDS->seq($cds[0]->seq);
            for($i=1; $i < ($n-1); $i++) {
              $cdsIntrOut->write_seq($cds[$i]);
              $thisCDS->seq($thisCDS->seq . $cds[$i]->seq);
            }
            $cdsTermOut->write_seq($cds[$n-1]);
            $thisCDS->seq($thisCDS->seq . $cds[$n-1]->seq);
            $cdsOut->write_seq($thisCDS);
            $nCdsInit++;
            $nCdsIntr += $n-2;
            $nCdsTerm++;
          }
          $nCds += scalar(@cds);

          ## Write out any intron sequences
          foreach $intron (@introns) {
            $intronOut->write_seq($intron);
          }
          $nIntron += scalar(@introns);
  
          ## Write out any donor sequences
          foreach $don (@don) {
            $donOut->write_seq($don);
          }
          $nDon += scalar(@don);
  
          ## Write out any acceptor sequences
          foreach $acc (@acc) {
            $accOut->write_seq($acc);
          }
          $nAcc += scalar(@acc);
  
          ## Sum the number of coding and intron bases
          foreach $cds (@cds) {
            $nBasesCds += length($cds->seq);
            # print "Adding " . length($cds->seq) . " to cds\n";
          }
          foreach $intron (@introns) {
            $nBasesIntron += length($intron->seq);
            # print "Adding " . length($intron->seq) . " to intron\n";
          }
        } else {
          $cdsIgnored = 1;
        }

      }
    }

    ##
    ## add on the final intergenic region
    ##
    $igHi = $seqLen;
    if(($igLo > 0) && ($igLo <= $igHi) && ($igHi <= $seqLen)) {
      $thisIG = Bio::PrimarySeq->new();
      $thisIG->seq($seqobj->subseq($igLo,$igHi));
      $thisIG->display_id($seqId);
      push(@ig,$thisIG);
    }

    ##
    ## Shift and pop off the end intergenic sequences, unless option overrides.
    ##
    if(! $option{"ends"}) {
      shift(@ig);
      pop(@ig);
    }

    ##
    ## Write out all intergenic sequences
    ##
    foreach $ig (@ig) {
      $intergeneOut->write_seq($ig);
      $nBasesIG += length($ig->seq);
      #print "Added " . length($ig->seq) . " to IG\n";
    }
    $nIG += scalar(@ig);

    ##
    ## If it was parsed correctly then the condition below should hold,
    ## unless some CDS's were left out because they were labelled to
    ## be ignored, or the ingtergenic regions at the end were ignored.
    ##
    $seqLen2 = $nBasesCds + $nBasesIntron + $nBasesIG;
    if((!$cdsIgnored) && ($option{"ends"}) && ($seqLen != $seqLen2)) {
      print STDERR "Warning: length of sequence $seqIndex in $seqFile is $seqLen, but\n";
      print STDERR "  it was parsed as $seqLen2 bases (maybe due to overlapping annotations?)\n";
    }

    $nBasesCdsTot    += $nBasesCds;
    $nBasesIntronTot += $nBasesIntron;
    $nBasesIGTot     += $nBasesIG;
  }

  if($option{"v"}) {
    print "Processed $seqFile\n";
    printf("  %5d Sequences\n",$seqIndex);
    printf("  %5d CDS entries\n",$nCds);
    printf("  %5d   single  \n",$nCdsSing);
    printf("  %5d   initial \n",$nCdsInit);
    printf("  %5d   internal\n",$nCdsIntr);
    printf("  %5d   terminal\n",$nCdsTerm);
    printf("  %5d Introns\n",$nIntron);
    printf("  %5d Donors\n",$nDon);
    printf("  %5d Acceptors\n",$nAcc);
    printf("  %5d Intergenic regions\n",$nIG);
    print "\n";
  }
}

if(!$option{"q"}) {
  print "Base count of all data\n";
  printf("  %7d coding\n",$nBasesCdsTot);
  printf("  %7d intron\n",$nBasesIntronTot);
  printf("  %7d intergene\n",$nBasesIGTot);
  printf("\nNote that some CDS's were left out as they were labelled \"IGNORE\"\n") if($cdsIgnored);
}
exit(0);


#
# Usage function - will print out commented lines at the top of this 
# file until it encounters a non-commented line.  (Skips the first.)
#
sub usage {
  open(MYSELF,$0) || die "$0: can't open $0 to print usage!\n";
  while(<MYSELF>) {
    if(/^\#!/) {
      next;
    } elsif (/^\s*\#(.*)$/) {
      eval "print STDERR \"$1\n\"";
    } else {
      last;
    }
  }
  close MYSELF;
}
