/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#ifndef HEADER_COMMON
#define HEADER_COMMON 1


/*
** Global functions
*/
void *myAlloc(size_t nBytes);
void error(char *fmt, ...);


/*
** Global variables
*/

/* splice site lengths */
EXTERN int acc_upstr;
EXTERN int acc_dnstr;
EXTERN int don_upstr;
EXTERN int don_dnstr;


/*
** Global constants
*/

#define N_BASES   4   /* number of DNA bases = 4 (ACGT) */
#define INTRORDER 2   /* the order of the MM used in intron/intergene */
#define INTRPRED 16   /* 4^INTRORDER, number of possible intron/intergene contexts */
#define ORDER 5       /* the order of the MM used in exons */
#define EXTRA 5       /* number of bases to paste on before start of seq (ie ORDER) */
#define EXONTYPES 4   /* 4 types: single, initial, internal, terminal */
#define NSTOPS 3      /* the number of possible stop codons (3: TAA, TAG, TGA) */
#define STOPLEN 3     /* length of stop signal (always 3 bases) */
#define NUMSTATES 7   /* number of I-states in the HMM */


/*
** Global enums
*/
enum types      {TYPE_SNGL,TYPE_INIT,TYPE_INTL,TYPE_TERM};
enum states     {INTERGENE,I0F,I1F,I2F,I0B,I1B,I2B};
enum seqFormats {FASTA,GENBANK,UNKNOWN_FORMAT};
enum exType     {NADA,SNGL_F,INIT_F,INTL_F,TERM_F,SNGL_B,INIT_B,INTL_B,TERM_B};
enum directions {FORWARD, BACKWARD};

#define DI_NUC 2        /* number of bases in a di-nucleotide */
#define N_DI_NUC 16     /* number of possible di-nucleotides */
#define TRI_NUC 3       /* number of bases in a tri-nucleotide */
#define N_TRI_NUC 64    /* number of possible tri-nucleotides */

#define START_SIGNAL_LEN 3  /* length of the start signal, ATG */
#define STOP_SIGNAL_LEN  3  /* length of the stop signal, TAA/TAG/TGA */
#define PHASES 3            /* 3 possible phases in which an exon can be started (0/1/2) */

#define MAX_PATH_LEN     512   /* max # of characters in a file's path */
#define MAX_FILENAME_LEN 256   /* max # of characters in a filename (not including path) */
#define MAX_LINE_LEN     1024  /* max # of characters in a line read from input files */
#define MAX_WORD_LEN     256   /* max # of characters in a line read from input files */
#define ENV_PAR_PATH     "PHATPARS"  /* environment var, location of data directory */

/* macros for max/min */
#define MAX(x,y) ( (x) > (y) ) ? (x) : (y)
#define MIN(x,y) ( (x) < (y) ) ? (x) : (y)

/* macros for trinucleotides */
#define isfstart(x) (x == 14)   /* forward start codon */
#define isbstart(x) (x == 19)   /* reverse start codon */
#define isfstop(x) ((x == 48) || (x == 50) || (x == 56))  /* forward stop codon */
#define isbstop(x) ((x == 28) || (x == 52) || (x == 60))  /* reverse stop codon */

/* macros for dinucleotides */
#define isfacc(y) (y == 2)   /* candidate acceptor on direct strand */
#define isfdon(y) (y == 11)  /* candidate donor on direct strand */
#define isbacc(y) (y == 7)   /* candidate acceptor on reverse strand */
#define isbdon(y) (y == 1)   /* candidate acceptor on reverse strand */


/*
** Default extensions for output files
*/
#define SPLICE_FILE_EXT  ".spl"
#define MRNA_FILE_EXT    ".rna"
#define PROT_FILE_EXT    ".pep"
#define GFF_FILE_EXT     ".gff"
#define PROB_FILE_EXT    ".prb"

/*
** Default names for input parameter files
*/
#define ACCEPTOR_FILE        "acceptor.dat"
#define DONOR_FILE           "donor.dat"
#define EXONFREQ_FILE        "exon.freq.dat"
#define EXON_LEN_SING_FILE   "exon.len.sing.dat"
#define EXON_LEN_INIT_FILE   "exon.len.init.dat"
#define EXON_LEN_INTL_FILE   "exon.len.intl.dat"
#define EXON_LEN_TERM_FILE   "exon.len.term.dat"
#define INTRONFREQ_FILE      "intron.freq.dat"
#define INTRON_LEN_FILE      "intron.len.dat"
#define INTERGENE_FREQ_FILE  "intergene.freq.dat"
#define INTERGENE_LEN_FILE   "intergene.len.dat"
#define TRANSITIONS_FILE     "transitions.dat"
#define INITPROBS_FILE       "initprobs.dat"
#define STOP_PROBS_FILE      "stopprobs.dat"

/*
** Stuff for storing gene locations 
*/
#define MAX_EXONS 250
#define MAX_GENES 700

#define NOMEM "Insufficient memory: aborting\n"

/*
** Macro for adding log probabilities
*/
#define FALSE 0
#define myadd(x,y) \
do{\
	(y) += log(1 + exp((x)-(y)));\
}while(FALSE)


/*
** The option structure
*/
typedef struct {
  /* output options */
  int  noprob;     /* if true, skip writing probability file */
  int  norna;      /* if true, skip writing predicted rna file */
  int  nopep;      /* if true, skip writing predicted peptide file */
  int  noextras;   /* if true, will only write to screen and gff file */
  int  okstops;    /* if true, in-frame stop codons are allowed */
  int  oldprb;     /* if true, use old (big!) format for posterior probability file */

  int  noext;      /* don't add extensions to seq names in output files */
  int  verbose;    /* run in verbose mode (quite a bit of output) */
  int  debug;      /* run in debug mode (lots of output) */
  int  geometric;  /* use geometric distributions for E states */
  int  offset;     /* to be added to coordinates when writing output */
  char baseFileName[MAX_FILENAME_LEN]; /* base name (incl path) for output files */
  char outSeqName[MAX_FILENAME_LEN];   /* name for output sequences */

  /* the input file */
  char inFile[MAX_PATH_LEN + MAX_FILENAME_LEN];

  /* output files */
  char outSpliceFileName[MAX_FILENAME_LEN];
  char outMrnaFileName[MAX_FILENAME_LEN];
  char outProtFileName[MAX_FILENAME_LEN];
  char outGffFileName[MAX_FILENAME_LEN];
  char outProbFileName[MAX_FILENAME_LEN];

  /* parameter files */
  char acceptorFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char donorFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenSingFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenInitFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenIntlFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenTermFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intronFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intronLenFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intergeneFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intergeneLenFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char transitionsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char initProbsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char stopProbsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
} Option;


/* 
** The gene structure
*/
typedef struct {
  int nExons;
  int start[MAX_EXONS];
  int stop[MAX_EXONS];
  int frame[MAX_EXONS];
  int phase[MAX_EXONS];
  int type[MAX_EXONS];
  double prob[MAX_EXONS];
} Gene;


#endif
