/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#define EXTERN extern
#include "full_common.h"
#include "full_seq.h"
#undef EXTERN

#define EXTERN
#include "full_output.h"
#undef EXTERN

/*
** Functions servicing the globally-available
** output-writing functions
*/
int printFastaSeq(int start, int stop, char *thisSeq, int linePos, int charsPerLine, int addNewLine, int revComp, int mRna, FILE *outFile);
void copyFastaSeq(int start, int stop, char *thisSeq, char *destination, int dir);
void translate(char *cDNA, int n, FILE *outFile, int rev, int phase, int charsPerLine, char *name);
char translateCodon(int f, int s, int t);
int  num2base(int c);
int  num2mrna(int c);


/*
** Screen output
*/
void writeScreenOutput(Gene *gene, int nGenes, int nBases, Option *option) {
  char geneName[10];
  int i,j;
  int nExons,exonType,exLeft,exRight,exPhase,exFrame,exLen;
  char exDir;
  double exProb;
  char exName[9][5] = {
    "nada",  /* (this one shouldn't ever be needed) */
    "sing",
    "init",
    "intl",
    "term",
    "sing",
    "init",
    "intl",
    "term"};

  if(option->verbose)
    printf("\nWriting output to screen\n");

  printf("\n####################################################\n");
  if(nGenes ==1)
    printf("#         %4d gene    %7d base pairs          #\n",nGenes,nBases);
  else
    printf("#         %4d genes    %7d base pairs         #\n",nGenes,nBases);
  printf("####################################################\n");
  printf("\n");
  printf("Gene Exon Dir Type   Start    Stop Length Fr Ph   Pr\n");
  for(i=nGenes-1; i>=0; i--) {
    sprintf(geneName,"%d",nGenes-i);
    nExons = gene[i].nExons;
    for(j=nExons-1; j>=0; j--) {
      exonType = gene[i].type[j];
      exLeft   = gene[i].start[j];
      exRight  = gene[i].stop[j];
      exPhase  = gene[i].phase[j];
      exFrame  = gene[i].frame[j]; 
      exProb   = gene[i].prob[j];
      exLen      = exRight - exLeft + 1;
      exDir = (exonType <= TERM_F) ? '+' : '-';

      printf("%4s %4d %3c %4s %7d %7d %6d %2d %2d %1.2f\n",
        geneName,nExons-j,exDir,exName[exonType],
        exLeft+1,exRight+1,exLen,exFrame,exPhase,exProb);
    }
  }

  return;
}


/*
** write GFF file
*/
void writeGffFile(Gene *gene, int nGenes, Option *option) {
  FILE *outGffFile = NULL;
  char geneName[10];
  int i,j;
  int nExons,exonType,exLeft,exRight,exPhase,exFrame;
  double exProb;
  char exDir;

  /* open gff file for write */
  if((outGffFile = fopen(option->outGffFileName,"w")) == NULL) {
    fprintf(stderr,"writeGffFile: can't open %s for write\n",option->outGffFileName);
    exit(1);
  }
  
  /* write the gff output */
  for(i=nGenes-1; i>=0; i--) {
    sprintf(geneName,"Phat%d",nGenes-i);
    nExons = gene[i].nExons;
    for(j=nExons-1; j>=0; j--) {
      exonType = gene[i].type[j];
      exLeft   = gene[i].start[j];
      exRight  = gene[i].stop[j];
      exPhase  = gene[i].phase[j];
      exFrame  = gene[i].frame[j]; 
      exProb   = gene[i].prob[j];
      exDir = (exonType <= TERM_F) ? '+' : '-';
      fprintf(outGffFile,"%s\tFullPhat\tCDS\t%d\t%d\t%1.2f\t%c\t%d\tgene \"%s\" frame\"%d\"\n",
        option->outSeqName,exLeft+1,exRight+1,exProb,exDir,exPhase,geneName,exFrame);
    }
  }
  fclose(outGffFile);

  if(option->verbose) printf("  Wrote gff output to %s\n",option->outGffFileName);
  return;
}


/*
** write mRNA file
*/
void writeMrnaFile(Gene *gene, int nGenes, Option *option) {
  FILE *outMrnaFile = NULL;
  char geneName[MAX_PATH_LEN+MAX_FILENAME_LEN];
  int i,j,isMrna;
  int exLeft,exRight,exLen;
  int nExons;
  int dir;
  char *mRNA = NULL;
  char *mRNAtemp;
  static char *genomicSeq;
  int mRNAlen,mRNAmem;

  if(option->verbose)
    printf("\nWriting predicted mRNA to %s\n",option->outMrnaFileName);

  /* open mrna file for write */
  if((outMrnaFile = fopen(option->outMrnaFileName,"w")) == NULL) {
    fprintf(stderr,"writeMrnaFile: can't open %s for write\n",option->outMrnaFileName);
    exit(1);
  }
  
  /* write the mrna output */
  mRNAmem = 0;
  for(i=nGenes-1; i>=0; i--) {
    if(option->noext)
      sprintf(geneName,"%s",option->outSeqName);
    else
      sprintf(geneName,"%s.%d",option->outSeqName,nGenes-i);
    if(option->verbose)
      printf("  Writing rna for %s\n",geneName);
    nExons = gene[i].nExons;
    dir = (gene[i].type[0] <= TERM_F) ? FORWARD : BACKWARD;

    /* compute mRNA length and declare mem if necessary */
    for(mRNAlen=0, j=nExons-1; j >=0; j--)
      mRNAlen += gene[i].stop[j] - gene[i].start[j] + 1;
    if(mRNAlen > mRNAmem) {
      if(mRNA != NULL)
        free(mRNA);
      mRNA = (char *) malloc (mRNAlen * sizeof(char));
      if(mRNA == NULL) {
        fprintf(stderr,"writeMrnaFile: couldn't get mem for mRNA of length %d\n",mRNAlen);
        exit(1);
      }
      mRNAmem = mRNAlen;
    }
    if(option->verbose)
      printf("    length = %d\n",mRNAlen);

    genomicSeq = getGenomicSeqPointer();
    /* copy the mRNA sequence over */
    mRNAtemp = mRNA;
    if(dir == FORWARD) {
      for(j=nExons-1; j>=0; j--) {
        exLeft   = gene[i].start[j] - option->offset;
        exRight  = gene[i].stop[j] - option->offset;
        exLen    = exRight - exLeft + 1;
        copyFastaSeq(exLeft,exRight,genomicSeq,mRNAtemp,dir);
        mRNAtemp += exLen;
      }
    } else {
      for(j=0; j<nExons; j++) {
        exLeft   = gene[i].start[j] - option->offset;
        exRight  = gene[i].stop[j] - option->offset;
        exLen    = exRight - exLeft + 1;
        copyFastaSeq(exLeft,exRight,genomicSeq,mRNAtemp,dir);
        mRNAtemp += exLen;
      }
    }

    /* print out the full-length mRNA */
    isMrna = 1;
    if(option->verbose)
      printf("    %s, %d exons, %d bp\n",geneName,nExons,mRNAlen);
    fprintf(outMrnaFile,">%s, %d exons, %d bp\n",geneName,nExons,mRNAlen);
    printFastaSeq(0,mRNAlen-1,mRNA,0,60,1,0,isMrna,outMrnaFile);
    fprintf(outMrnaFile,"\n");
  }
  fclose(outMrnaFile);

  if(mRNA != NULL)
    free(mRNA);

  if(option->verbose) printf("  Wrote predicted mrna to %s\n",option->outMrnaFileName);

  return;
}


/*
** write predicted peptide file
*/
void writeProtFile(Gene *gene, int nGenes, Option *option) {
  FILE *outProtFile = NULL;
  char geneName[MAX_PATH_LEN+MAX_FILENAME_LEN];
  int i,j;
  int exLeft,exRight,exLen;
  int nExons;
  int dir;
  char *mRNA = NULL;
  char *mRNAtemp;
  static char *genomicSeq;
  int mRNAlen,mRNAmem,mRNAphase;

  /* open peptide file for write */
  if((outProtFile = fopen(option->outProtFileName,"w")) == NULL) {
    fprintf(stderr,"writeProtFile: can't open %s for write\n",option->outProtFileName);
    exit(1);
  }
  
  /* write the mrna to mRNA */
  mRNAmem = 0;
  for(i=nGenes-1; i>=0; i--) {
    if(option->noext)
      sprintf(geneName,"%s",option->outSeqName);
    else
      sprintf(geneName,"%s.%d",option->outSeqName,nGenes-i);
    nExons = gene[i].nExons;
    dir = (gene[i].type[0] <= TERM_F) ? FORWARD : BACKWARD;

    /* compute mRNA length and declare mem if necessary */
    for(mRNAlen=0, j=nExons-1; j >=0; j--)
      mRNAlen += gene[i].stop[j] - gene[i].start[j] + 1;
    if(mRNAlen > mRNAmem) {
      if(mRNA != NULL)
        free(mRNA);
      mRNA = (char *) malloc (mRNAlen * sizeof(char));
      if(mRNA == NULL) {
        fprintf(stderr,"writeProtFile: couldn't get mem for mRNA of length %d\n",mRNAlen);
        exit(1);
      }
      mRNAmem = mRNAlen;
    }

    genomicSeq = getGenomicSeqPointer();
    /* copy the mRNA sequence over */
    mRNAtemp = mRNA;
    if(dir == FORWARD) {
      for(j=nExons-1; j>=0; j--) {
        exLeft   = gene[i].start[j] - option->offset;
        exRight  = gene[i].stop[j] - option->offset;
        exLen    = exRight - exLeft + 1;
        copyFastaSeq(exLeft,exRight,genomicSeq,mRNAtemp,dir);
        mRNAtemp += exLen;
      }
      mRNAphase = gene[i].phase[nExons-1];
    } else {
      for(j=0; j<nExons; j++) {
        exLeft   = gene[i].start[j] - option->offset;
        exRight  = gene[i].stop[j] - option->offset;
        exLen    = exRight - exLeft + 1;
        copyFastaSeq(exLeft,exRight,genomicSeq,mRNAtemp,dir);
        mRNAtemp += exLen;
      }
      mRNAphase = gene[i].phase[0];
    }

    /* translate and print the mRNA */
    translate(mRNA,mRNAlen,outProtFile,0,mRNAphase,50,geneName);
  }
  fclose(outProtFile);

  if(mRNA != NULL)
    free(mRNA);

  if(option->verbose) printf("  Wrote predicted peptides to %s\n",option->outProtFileName);

  return;
}



/*
** Use the standard genetic code to translate the sequence in the mRNA array.
*/
void translate(char *mRNA, int n, FILE *outFile, int rev, int phase, int charsPerLine, char *name) {
  int i,charCounter,start,step,limit,baseIndex,f,s,t,needNewline,nAA;
  char aa;

  if(n-phase < 3)
    return;

  nAA = (n-phase)/3;

  if(!rev) {
    start = 0;
    step = 1;
    limit = n-1;
    rev = -1;
  } else {
    start = n-1;
    step = -1;
    limit = 0;
    rev = 1;
  }

  /* printf("gene %s: phase is %d\n",name,phase); */
  baseIndex = start;
  /* deal with nonzero phase */
  for(i=0; i<phase; i++)
    baseIndex += step;

  charCounter=0;
  needNewline = 0;
  fprintf(outFile,">%s, %d aa\n",name,nAA);
  while(rev*(baseIndex+2*step-limit) >= 0) {
    f = mRNA[baseIndex];
    s = mRNA[baseIndex+step];
    t = mRNA[baseIndex+2*step];
    /* printf("  translating %d%d%d\n",f,s,t); */
    /* printf("  rev is %d\n",rev); */
    /* complement if reverse direction */
    if(rev == 1) {
      f = N_BASES - 1 - f;
      s = N_BASES - 1 - s;
      t = N_BASES - 1 - t;
    }
    /* printf("  translating %d%d%d\n",f,s,t); */
    aa = translateCodon(f,s,t);
    /* printf("    aa is %c\n",aa); */
/* exit(1); */
    ++charCounter;
    /* printf("    baseIndex is %d, charCounter is %d\n",baseIndex,charCounter); */
    if(aa == 'J') {
      fprintf(stderr,"phat: translate: couldn't translate codon %c%c%c\n",
        num2base(f),num2base(s),num2base(t));
      exit(1);
    }
    if(needNewline) {
      needNewline = 0;
      fprintf(outFile,"\n");
    }
    fprintf(outFile,"%c",aa);
    if(charCounter == charsPerLine) {
      needNewline = 1;
      charCounter = 0;
    } else if ( (charCounter % 10) == 0) {
      fprintf(outFile," ");
    }
    baseIndex += 3*step;
  }
  fprintf(outFile,"\n\n");
  return;
}



/*
** translateCodon(f,s,t) returns the amino acid implied by the DNA, according to the 
** standard genetic code.
*/
char translateCodon (int f, int s, int t) {
  if (f == 0) {
    if (s == 0) {
      if ((t == 0) ||   (t == 2) )
        return ('K');
        else
        return ('N');
      }
  
    if (s == 1)
      return ('T');
  
    if (s == 2) {
      if ((t == 0) ||   (t == 2) )
        return ( 'R');
      else
        return ( 'S');
    }
  
    if (s == 3) {
      if (t == 2) 
        return ( 'M');
      else 
        return ( 'I');
    }
  }
    
  if (f == 1) {
    if (s == 0) {
      if ((t == 0) ||   (t == 2) )
        return ( 'Q');
      else
        return ( 'H');
    }
  
    if (s == 1)
      return ( 'P');
  
    if (s == 2)
      return ( 'R');
  
    if (s == 3)
      return ( 'L');
  }
  
  if (f == 2) {
    if (s == 0) {
      if ((t == 0) ||   (t == 2) )
        return ( 'E');
      else
        return ('D');
    }
  
    if (s == 1)
      return ( 'A');
  
    if (s == 2)
      return ( 'G');
  
    if (s == 3)
      return ( 'V');
  }
  
  if (f == 3) {
    if (s == 0) {
      if ((t == 0) ||   (t == 2))
        return ( '*');
      else
        return ( 'Y');
    }
  
    if (s == 1)
      return ( 'S');
  
    if (s == 3) {
      if ((t == 0) ||   (t == 2) )
        return ( 'L');
      else
        return ( 'F');
    }
  
    if (s == 2) {
      if (t == 0)
        return ( '*');
  
      if (t == 2) 
        return ( 'W');
  
      if ((t == 1) ||   (t == 3) )
        return ('C');
    }
  }

  /* use this symbol as alternative, should never be used*/
  return ('J');  
}


/*
 * Prints out the bases in the range thisSeq[left,right] inclusive.
 * left should be less than of right.
 * charsPerLine determines the number of characters per line.
 * If addNewLine is true then the final line will be terminated
 * with a newline, otherwise it won't.
 * If revComp is true the reverse complement of the sequence will printed.
 */
int printFastaSeq(int left, int right, char *thisSeq, int charCounter, int charsPerLine, int addNewLine, int revComp, int mRna, FILE *outFile) {
  int i,needNewline,start,stop,step,len,count;
  char base;

  /* make sure thisSeq is non NULL */
  if(thisSeq == NULL) {
    fprintf(stderr,"printFastaSeq: pointer thisSeq points to NULL, exiting...\n");
    exit(EXIT_FAILURE);
  }

  needNewline = 0;
  if(left <= right) {
    /* printf("start,stop are %d,%d\n",start,stop); */
    if(revComp) {
      start = right;
      stop  = left;
      step  = -1;
    } else {
      start = left;
      stop  = right;
      step  = 1;
    }
    len = right-left+1;
    for(i=start,count=0; count < len; i+=step,count++) {
      if(needNewline) {
        fprintf(outFile,"\n");
        needNewline = 0;
      }
      if(mRna) {
        base = revComp ? num2mrna(N_BASES -1 -thisSeq[i]) : num2mrna(thisSeq[i]);
      } else {
        base = revComp ? num2base(N_BASES -1 -thisSeq[i]) : num2base(thisSeq[i]);
      }
      fprintf(outFile,"%c",base);
      if(++charCounter == charsPerLine) {
        needNewline = 1;
        charCounter = 0;
      }
    }
    if(addNewLine)
      fprintf(outFile,"\n");
  } else {
    fprintf(stderr,"phat: printFastaSeq: bad input, left is greater than right\n");
    exit(1);
  }

  return(charCounter);
}


/*
** Copies the bases from thisSeq[left,right] into the supplied char array.
** If revComp is true then the copied sequence will be reverse complemented.
*/
void copyFastaSeq(int left, int right, char *thisSeq, char *destination, int dir) {
  int i,start,stop,step,count,len,revComp;

  /* determine if reverse complementing is necessary */
  revComp = (dir == FORWARD) ? 0 : 1;

  /* make sure thisSeq is non NULL */
  if(thisSeq == NULL) {
    fprintf(stderr,"copyFastaSeq: pointer thisSeq points to NULL, exiting...\n");
    exit(EXIT_FAILURE);
  }

  if(left <= right) {
    /* printf("left,right are %d,%d\n",start,stop); */
    if(revComp) {
      start = right;
      stop  = left;
      step  = -1;
    } else {
      start = left;
      stop  = right;
      step  = 1;
    }
    len = right-left+1;

    for(i=start,count=0; count < len; i+=step,count++)
      *destination++ = revComp ? (N_BASES -1 -thisSeq[i]) : thisSeq[i];

  } else {
    fprintf(stderr,"phat: copyFastaSeq: left greater than right!\n");
    exit(1);
  }
  return;
}


/* 
** returns
** A if c is 0
** C if c is 1
** G if c is 2
** T if c is 3
** N otherwise
*/
int num2base(int c) {
  switch(c) {
    case 0:
      return 'A';
    break;

    case 1:
      return 'C';
    break;

    case 2:
      return 'G';
    break;

    case 3:
      return 'T';
    break;

    default:
      return 'N';
    break;
  }
}


/* 
** returns
** A if c is 0
** C if c is 1
** G if c is 2
** U if c is 3
** N otherwise
*/
int num2mrna(int c) {
  switch(c) {
    case 0:
      return 'A';
    break;

    case 1:
      return 'C';
    break;

    case 2:
      return 'G';
    break;

    case 3:
      return 'U';
    break;

    default:
      return 'N';
    break;
  }
}
