/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>

#define EXTERN extern
#include "full_common.h"
#include "full_splice.h"
#undef EXTERN

#define EXTERN
#include "full_probs.h"
#undef EXTERN


/*
** Exon probability data
** 
** fExonFreq/bExonFreq are the frequencies determined from the 
**   parameter file, just a 3x1024x5 array of frequencies.
**   The last dimension is 5 because there is an extra entry for
**   the sum of the frequencies for each phase and context.
** fLogExonFreq/bLogExonFreq are logs of the frequencies in the
**   fExonFreq/bExonFreq arrays, though they will be zero if they
**   haven't been required yet - avoids unnecessary logging.
**   Array size is 3x1024x4.
** fExonProb/bExonProb are potentially big: 3 x slen+1.
**     fExonProb[j][t] = log(Pr(Y_0^t | forward coding exon in frame j))
**     fExonProb[j][t] = log(Pr(Y_0^t | backward coding exon in frame j))
**   where Y_0^t is the DNA sequence from [0..t] and the frame is
**   computed relative to the start (left end) of the sequence, 
**   irrespective of whether the direction is forward or backward.
** fStopCount/bStopCount are also potentially big: 3 x slen+1.
**      fStopCount[j][t] = # forward stop codons up to and incl t in frame j.
**      bStopCount[j][t] = # backward stop codons up to and incl t in frame j.
** logStop is an array of size 3, just log probabilities of stop codons.
*/
static int    exonOrder;
static double ***fLogExonFreq;
static double ***bLogExonFreq;
static double *fExonProb[PHASES];
static double *bExonProb[PHASES];
static int    *fStopCount[PHASES];
static int    *bStopCount[PHASES];
static double logStop[NSTOPS];  /* probabilities of stop codons */

/*
** Intergene probability data
** 
** intergenicFreq holds the frequencies determined from the 
**   parameter file, a 16x5 array of frequencies.
**   The last dimension is 5 because there is an extra entry for
**   the sum of the frequencies for each context.
** logIntergenicFreq holds logs of the frequencies in the
**   intergenicFreq array, though they will be zero if they
**   haven't been required yet - avoids unnecessary logging.
**   Array size is 16x4.
** intergenicProbs is a potentially big array of size slen.
**     intergenicProbs[t] = log(Pr(Y_t | intergene))
**   where Y_t is the DNA base in position t (zero-based)
*/
static int    intergenicFreq[INTRPRED][N_BASES];
static double logIntergenicFreq[INTRPRED][N_BASES];
static double *intergenicProbs;

/*
** Intron probability data
** 
** fIntronFreq/bIntronFreq hold the frequencies determined from the 
**   parameter file, a 16x5 array of frequencies.
**   The last dimension is 5 because there is an extra entry for
**   the sum of the frequencies for each context.
** fLogIntronFreq/bLogIntronFreq holds logs of the frequencies 
**   in the fIntronFreq/bIntronFreq arrays, though they will be zero 
**   if they haven't been required yet - avoids unnecessary logging.
**   Array size is 16x4.
** fIntronProbs/bIntronProbs are potentially big arrays of size slen.
**     fIntronProbs[t] = log(Pr(Y_t | forward intron))
**     bIntronProbs[t] = log(Pr(Y_t | backward intron))
**   where Y_t is the DNA base in position t (zero-based)
*/
static int    fIntronFreq[INTRPRED][N_BASES+1];
static int    bIntronFreq[INTRPRED][N_BASES+1];
static double fLogIntronFreq[INTRPRED][N_BASES];
static double bLogIntronFreq[INTRPRED][N_BASES];
static double *fIntronProbs;
static double *bIntronProbs;

/*
** Exon length probability objects
**  numExons is an int array of length 4 holding the max length
**    of an exon of each possible type (single,initial,internal,terminal)
**  exonLenDensity is a ragged array of size 4, each entry is an
**    array holding the length density of that type of exon.  Note that
**    this will be normalized such that the density sums to one for each
**    length modulo 3, so the total sum will actually be 3.  This is beacuse
**    of the frame constraints in the model which restrict the length of each
**    exon to some value modulo 3.
**  exonQ is an array of self-transition probabilities that can be used
**    if a geometric length density is chosen instead of using general lengths.
**  exonP = log(1-exp(exonQ))
*/
static int numExons[EXONTYPES];
static double *exonLenDensity[EXONTYPES];
static double exonP[EXONTYPES];
static double exonQ[EXONTYPES];

/*
** Intron/Intergene length density parameters:
**
** Intron/Intergene lengths are geometric.  The notation
** for these is the same as the (rarely used) exonP/exonQ variables.
** logQ is the log probability of a self transition, logP is log(1-exp(logQ))
*/
static double logP[NUMSTATES];
static double logQ[NUMSTATES];



/* HMM intial and transition probabilities */
static double logInit[NUMSTATES];
static double transitions[NUMSTATES][NUMSTATES];


/* functions internal to full_probs.c */
double ***declare3dArray(int dim1,int dim2,int dim3);
void free3dArray( double ***array, int dim1, int dim2);


/*
** Reads in all the model parameters.
*/
void readStats(Option *option) {

  readAcceptorData(option->acceptorFile,option);
  readDonorData(option->donorFile,option);

  readIntronLen(option->intronLenFile); 
  readIntronFreq(option->intronFreqFile);

  readExonLen(option);
  readExonFreq(option->exonFreqFile);

  readIntergeneLen(option->intergeneLenFile);
  readIntergeneFreq(option->intergeneFreqFile);

  readInitProbs(option->initProbsFile);
  readTransitions(option->transitionsFile, option);
  readStopProbs(option->stopProbsFile, option);

  if(option->geometric)
    geometricFix(option);

  return;
}

/*
** Reads intron frequencies.  Zero counts are replaced by 1.
*/
void readIntronFreq(char *inFile) {
  FILE *fp;
  char line[MAX_LINE_LEN];
  int  context,base;
  int  revContext,revBase;
  int  tempContext,i;
  int  sum;
  int  off1,off2,freq,lineNo;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntronFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
    fprintf(stderr,"readIntronFreq: can't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(context=0,lineNo=3; context<INTRPRED; context++) {
    if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
      fprintf(stderr,"readIntronFreq: bad format in line %d of %s\n",lineNo+3,inFile);
      exit(EXIT_FAILURE);
    }
    off1 = 0;
    for(base=0; base<N_BASES; base++) {
      sscanf(line+off1,"%d%n",&freq,&off2);
      if(freq == 0)
        freq = 1;
      off1 += off2;

      /* figure out revContext and revBase */
      tempContext = context;
      revContext = (N_BASES-1) - base;
      for(i=0; i<INTRORDER-1; i++) {
        revContext *= N_BASES;
        revContext += (N_BASES-1) - (tempContext % N_BASES);
        tempContext /= N_BASES;
      }
      revBase = (N_BASES-1) - tempContext;

      fIntronFreq[context][base]       = freq;
      bIntronFreq[revContext][revBase] = freq;
    }
  }

  /* now do totals */
  for(context=0; context<INTRPRED; context++) {
    for(sum=0, base=0; base<N_BASES; base++)
      sum += fIntronFreq[context][base];
    fIntronFreq[context][N_BASES] = sum;

    for(sum=0, base=0; base<N_BASES; base++)
      sum += bIntronFreq[context][base];
    bIntronFreq[context][N_BASES] = sum;
  }

  fclose(fp);
  return;
}


/*
** Just reads the mean intron length, which determines the 
** geometric parameter used for intron lengths.
*/
void readIntronLen(char *inFile) {
  FILE   *fp;
  char   line[MAX_LINE_LEN];
  double mean,p;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntronLen: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAX_LINE_LEN,fp) == NULL) { fprintf(stderr,"readIntronLen: file %s ended too early\n",inFile);
    exit(EXIT_FAILURE);
  }
  if(1 != sscanf(line,"%lf",&mean)) {
    fprintf(stderr,"readIntronLen: bad format\n");
    exit(EXIT_FAILURE);
  }
  p = (mean-1)/mean;

  logQ[I0F] = logQ[I1F] = logQ[I2F] = logQ[I0B] = logQ[I1B] = logQ[I2B] = log(p);
  logP[I0F] = logP[I1F] = logP[I2F] = logP[I0B] = logP[I1B] = logP[I2B] = log(1-p);

  fclose(fp);
  return;
}

/*
** Reads the intergenic frequencies.  Zero counts will be replaced by 1.
*/
void readIntergeneFreq(char *inFile) {
  FILE   *fp;
  char   line[MAX_LINE_LEN];
  int    i,j,sum;
  int    off1,off2,freq;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntergeneFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
    fprintf(stderr,"readIntergeneFreq: can't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0; i<INTRPRED; i++) {
    if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
      fprintf(stderr,"readIntergeneFreq: bad format in line %d of %s\n",i+3,inFile);
      exit(EXIT_FAILURE);
    }
    off1 = 0;
    for(j=0,sum=0; j<N_BASES; j++) {
      sscanf(line+off1,"%d%n",&freq,&off2);
      if(freq == 0)
        freq = 1;
      intergenicFreq[i][j]=freq;
      sum += freq;
      off1 += off2;
    }
    for(j=0; j<N_BASES; j++)
      logIntergenicFreq[i][j] = log((double) intergenicFreq[i][j] / (double) sum);
  }

  fclose(fp);
  return;
}


/*
** Juset reads in the mean intergenic length, which will
** be used to estimate a geometric parameter.
*/
void readIntergeneLen(char *inFile) {
  FILE   *fp;
  char   line[MAX_LINE_LEN];
  double mean,p;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntergeneLen: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
    fprintf(stderr,"readIntergeneLen: file %s ended too early\n",inFile);
    exit(EXIT_FAILURE);
  }
  if(1 != sscanf(line,"%lf",&mean)) {
    fprintf(stderr,"readIntergeneLen: bad format in line 1\n");
    exit(EXIT_FAILURE);
  }

  p = (mean-1) / mean;
  logP[INTERGENE] = log(1-p);
  logQ[INTERGENE] = log(p);

  fclose(fp);
  return;
}


/* 
** Reads in exon frequencies.  Zero counts will be set to 1.
** The sum for each will be computed, but the conversion to
** log probabilites is done elsewhere, in the lookup table.
*/
void readExonFreq(char *inFile) {
  FILE *fp;
  int phase,context,base,lineNo,revPhase,revContext,revBase;
  int tempContext,i,freq,off1,off2,nContexts;
  char line[MAX_LINE_LEN];
  double fFreqSum,bFreqSum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readExonFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  /* Pick up the order of the Markov model from the first line */ 
  if(fgets(line,MAX_LINE_LEN,fp) == NULL)
    error((char *)"readExonFreq: couldn't read first line\n");
  if(1 != sscanf(line,"Order %d",&exonOrder))
    error((char *)"readExonFreq: bad format in first line of %s\n",inFile);
  nContexts = (int) pow((double) N_BASES, (double) exonOrder);

  /* free old mem (if any) and declare the new mem */
  if(fLogExonFreq != NULL)
    free3dArray(fLogExonFreq,PHASES,N_BASES);
  fLogExonFreq = declare3dArray(PHASES,N_BASES,nContexts);
  if(bLogExonFreq != NULL)
    free3dArray(bLogExonFreq,PHASES,N_BASES);
  bLogExonFreq = declare3dArray(PHASES,N_BASES,nContexts);

  for(lineNo=2,phase=0; phase<PHASES; phase++) {
    /* 
    ** Phase on the reverse direction:
    ** In general, for a p-periodic k^th order MM, 
    **   phase i forwardstrand <-> phase ( k + (p-1)-i ) % p
    */
    revPhase = (exonOrder + (PHASES-1)-phase) % PHASES;
    for(context=0; context<nContexts; context++,lineNo++) {
      if(fgets(line,MAX_LINE_LEN,fp) == NULL)
        error((char *)"readExonFreq: bad format in line %d\n",lineNo);
      off1 = 0;
      for(base=0; base<N_BASES; base++) {
        sscanf(line+off1,"%d%n",&freq,&off2);
        off1 += off2;
        if(freq == 0)
          freq = 1;

        /* figure out revContext and revBase */
        revContext = (N_BASES-1) - base;
        tempContext = context;
        for(i=0; i < (exonOrder-1); i++) {
          revContext *= N_BASES;
          revContext += (N_BASES-1) - (tempContext % N_BASES);
          tempContext /= N_BASES;
        }
        revBase = (N_BASES-1) - tempContext;

        /* write the frequencies */
        fLogExonFreq[phase][base][context] = freq;
        bLogExonFreq[revPhase][revBase][revContext] = freq;
      }
    }

    /* normalize the frequencies and take logs */
    for(context=0; context<nContexts; context++) {
      /* get the sum for the current context */
      for(fFreqSum=0,bFreqSum=0,base=0; base<N_BASES; base++) {
        fFreqSum += fLogExonFreq[phase][base][context];
        bFreqSum += bLogExonFreq[revPhase][base][context];
      }
      /* normalize & take logs context */
      for(base=0; base<N_BASES; base++) {
        if(fLogExonFreq[phase][base][context] > 0)
          fLogExonFreq[phase][base][context] = log(fLogExonFreq[phase][base][context]/fFreqSum);
        else
          fLogExonFreq[phase][base][context] = -FLT_MAX;

        if(bLogExonFreq[revPhase][base][context] > 0)
          bLogExonFreq[revPhase][base][context] = log(bLogExonFreq[revPhase][base][context]/bFreqSum);
        else 
          bLogExonFreq[revPhase][base][context] = -FLT_MAX;
      }
    }

    /* skip the first line of the next section (it just tells the phase) */
    if((phase < PHASES-1) && (fgets(line,MAX_LINE_LEN,fp) == NULL))
      error((char *)"readExonFreq: file ended too early\n");
    lineNo++;
  }

  fclose(fp);
  return;
}


/* 
** Reads the four exon length densities into the exonLenDensity[][] 
** array.  If option->geometric is 1 then it will compute geometric
** parameters from these.
*/
void readExonLen(Option *option) {

  /* free possibly existing memory */
  if(exonLenDensity[TYPE_SNGL] != NULL)  free(exonLenDensity[TYPE_SNGL]);
  if(exonLenDensity[TYPE_INIT] != NULL)  free(exonLenDensity[TYPE_INIT]);
  if(exonLenDensity[TYPE_INTL] != NULL)  free(exonLenDensity[TYPE_INTL]);
  if(exonLenDensity[TYPE_TERM] != NULL)  free(exonLenDensity[TYPE_TERM]);

  /* read in the length distributions */
  exonLenDensity[TYPE_SNGL] = readLengthDist(numExons+TYPE_SNGL,option->exonLenSingFile,1,option->verbose);
  exonLenDensity[TYPE_INIT] = readLengthDist(numExons+TYPE_INIT,option->exonLenInitFile,1,option->verbose);
  exonLenDensity[TYPE_INTL] = readLengthDist(numExons+TYPE_INTL,option->exonLenIntlFile,1,option->verbose);
  exonLenDensity[TYPE_TERM] = readLengthDist(numExons+TYPE_TERM,option->exonLenTermFile,1,option->verbose);
  if(option->verbose) printf("\n");

  return;
}


/*
** Adjusts the parameters if we're using geometric exon lengths.
** Mean lengths are estimated from the supplied densities, and
** the transition probabilites are adjusted to account for the fact
** that the lengths are restricted modulo 3.
**
** NOTE: exonLenDensity[][] must have been initialized before this!
*/
void geometricFix(Option *option) {
  int i,j,phase;
  double mean[PHASES];
  double avgMean,p,sum;
  double p0,p1,p2;

  /* estimate geometric probabilities from exonLenDensity[][] */
  for(i=0;i<EXONTYPES;i++) {
    for(j=0; j<PHASES; j++)
      mean[j]=0;
    for(j=0, phase=0; j<numExons[i]; j++,phase++) {
      phase = phase % PHASES;
      mean[phase] += (j+1) * exp(exonLenDensity[i][j]);
    }
    for(j=0,avgMean=0; j<PHASES; j++)
      avgMean += mean[j];
    avgMean = (double) avgMean / (double) PHASES;

    p = (avgMean-1) / avgMean;
    exonQ[i] = log(p);
    exonP[i] = log(1-p);
  }

  /*
  ** Adjust transition probailities, which corrects for 
  ** the fact that lenths are restricted modulo 3.
  */

  /* single exons - no adjustment necessary */
  
  /* initial exons */
  p  = exp(exonQ[TYPE_INIT]);
  p0 = log(  1/(1 + p + p*p));
  p1 = log(  p/(1 + p + p*p));
  p2 = log(p*p/(1 + p + p*p));
  transitions[INTERGENE][I0F] -= p0;
  transitions[INTERGENE][I1F] -= p1;
  transitions[INTERGENE][I2F] -= p2;
  transitions[I0B][INTERGENE] -= p0;
  transitions[I1B][INTERGENE] -= p1;
  transitions[I2B][INTERGENE] -= p2;

  /* internal exons */
  p  = exp(exonQ[TYPE_INTL]);
  p0 = log(  1/(1 + p + p*p));
  p1 = log(  p/(1 + p + p*p));
  p2 = log(p*p/(1 + p + p*p));
  transitions[I0F][I0F] -= p0;
  transitions[I1F][I1F] -= p0;
  transitions[I2F][I2F] -= p0;
  transitions[I0B][I0B] -= p0;
  transitions[I1B][I1B] -= p0;
  transitions[I2B][I2B] -= p0;

  transitions[I0F][I1F] -= p1;
  transitions[I1F][I2F] -= p1;
  transitions[I2F][I0F] -= p1;
  transitions[I0B][I2B] -= p1;
  transitions[I1B][I0B] -= p1;
  transitions[I2B][I1B] -= p1;

  transitions[I0F][I2F] -= p2;
  transitions[I1F][I0F] -= p2;
  transitions[I2F][I1F] -= p2;
  transitions[I0B][I1B] -= p2;
  transitions[I1B][I2B] -= p2;
  transitions[I2B][I0B] -= p2;

  /* terminal exons */
  p  = exp(exonQ[TYPE_TERM]);
  p0 = log(  1/(1 + p + p*p));
  p1 = log(  p/(1 + p + p*p));
  p2 = log(p*p/(1 + p + p*p));
  transitions[I0F][INTERGENE] -= p0;
  transitions[I1F][INTERGENE] -= p2;
  transitions[I2F][INTERGENE] -= p1;
  transitions[INTERGENE][I0B] -= p0;
  transitions[INTERGENE][I1B] -= p2;
  transitions[INTERGENE][I2B] -= p1;

  if(option->verbose) {
    printf("Mean exon lengths (using geometric distribution)\n");
    for(i=0; i<EXONTYPES; i++)
      printf("  type %d = %f\n", i, 1/(1-exp(exonQ[i])) );
    printf("\n");
   
    printf("Transitions (adjusted for geometric lengths)\n");
    printf("      N      I0F     I1F    I2F      I0B     I1B    I2B\n  ");
    for(i=0; i<NUMSTATES; i++) {
      for(j=0,sum=0; j<NUMSTATES; j++) {
        printf("%2.5f ",exp(transitions[i][j]));
        sum += exp(transitions[i][j]);
      }
      printf("   (sum = %g)\n  ",sum);
    }
    printf("\n");
  }

  return;
}


/*
** Read in the transition probabilities.  Note that there is only one way to go from each
** I-state to another I-state, this is why we can store the transitions this way.  There is
** a single exception - there are two ways (which we treat as equiprobable) to jump from
** intergene to intergene.  In the transitions matrix we first ignore this, but after 
** possibly printint out the matrix we halve that particular transition probability.
*/
void readTransitions(char *inFile, Option *option) {
  FILE *fp;
  char line[MAX_LINE_LEN];
  char str[MAX_LINE_LEN];
  double freq,sum;
  int i,j,off1,off2;


  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readTransitions: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  /* skip the first line */
  if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
    fprintf(stderr,"readTransitions: couldn't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0; i<NUMSTATES; i++) {
    /* read in the next line */
    if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
      fprintf(stderr,"readTransitions: couldn't read line %d of %s\n",i+1,inFile);
      exit(EXIT_FAILURE);
    }

    /* skip the initial string */
    if(1 != sscanf(line,"%s%n",str,&off1)) {
      fprintf(stderr,"readTransitions: bad start to line %d of %s\n",i+1,inFile);
      exit(EXIT_FAILURE);
    }

    /* read the transitions */
    for(j=0,sum=0; j<NUMSTATES; j++) {
      if(1 != sscanf(line+off1,"%lf%n",&freq,&off2)) {
        fprintf(stderr,"readTransitions: bad format, line %d of %s\n",i+1,inFile);
        fprintf(stderr,"  Here's the offending line...\n%s\n",line);
        exit(EXIT_FAILURE);
      }
      if(freq < 0) {
        fprintf(stderr,"readTransitions: all entries must be >= 0 in %s\n",inFile);
        fprintf(stderr,"  (entry[%d][%d] isn't)\n",i+1,j+1);
        exit(EXIT_FAILURE);
      } else if (freq == 0) {
        freq++;
      }
      transitions[i][j]=freq;
      off1 += off2;
      sum += freq;
    }

    /* normalize the transitions */
    for(j=0; j<NUMSTATES; j++)
      transitions[i][j] = log(transitions[i][j] / sum);
  }

  /*
  ** print out all transitions if verbose
  */
  if(option->verbose) {
    printf("Transitions\n");
    printf("      N      I0F     I1F     I2F     I0B     I1B    I2B\n  ");
    for(i=0; i<NUMSTATES; i++) {
      for(j=0,sum=0; j<NUMSTATES; j++) {
        printf("%2.5f ",exp(transitions[i][j]));
        sum += exp(transitions[i][j]);
      }
      printf("   (sum = %g)\n  ",sum);
    }
    printf("\n");
  }

  /*
  ** halve transitions[INTERGENE][INTERGENE]
  */
  transitions[INTERGENE][INTERGENE] -= log(2.0);

  return;
}


/*
** Read in the intial probabilities of the HMM.
*/
void readInitProbs(char *inFile) {
  FILE   *fp;
  char   line[MAX_LINE_LEN];
  int    i;
  double sum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readInitProbs: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0,sum=0; i<NUMSTATES; i++) {
    if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
      fprintf(stderr,"readInitProbs: file %s ended too early\n",inFile);
      exit(EXIT_FAILURE);
    }
    if(1 != sscanf(line,"%lf",logInit+i)) {
      fprintf(stderr,"readInitProbs: bad format in line %d\n",i+1);
      exit(EXIT_FAILURE);
    }
    if(logInit[i] < 0) {
      fprintf(stderr,"readInitProbs: entries in %s must be >= 0\n",inFile);
      exit(EXIT_FAILURE);
    } else if (logInit[i] == 0) {
      logInit[i] = 1;  /* give a pseudocount of 1 if the entry is zero */
    }
    sum += logInit[i];
  }

  /* normalize */
  for(i=0; i<NUMSTATES; i++)
    logInit[i] = log(logInit[i]/sum);

  fclose(fp);
  return;
}

/*
** read the probabilities of the 3 possible stop codons.
*/
void readStopProbs(char *inFile, Option *option) {
  FILE   *fp;
  char   line[MAX_LINE_LEN];
  int    i;
  double sum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readStopProbs: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0,sum=0; i<NSTOPS; i++) {
    if(fgets(line,MAX_LINE_LEN,fp) == NULL) {
      fprintf(stderr,"readStopProbs: file %s ended too early\n",inFile);
      exit(EXIT_FAILURE);
    }
    if(1 != sscanf(line,"%lf",logStop+i)) {
      fprintf(stderr,"readStopProbs: bad format in line %d\n",i+1);
      exit(EXIT_FAILURE);
    }
    if(logStop[i] <= 0) {
      fprintf(stderr,"readStopProbs: entries in %s must be positive\n",inFile);
      exit(EXIT_FAILURE);
    }
    sum += logStop[i];
  }

  for(i=0; i<NSTOPS; i++)
    logStop[i] = log(logStop[i]/sum);

  if(option->verbose) {
    printf("Stop codon probabilities (TAA/TAG/TGA)\n");
    for(i=0; i<NSTOPS; i++)
      printf("  stop[%d] = %f\n",i,exp(logStop[i]));
    printf("\n");
  }

  fclose(fp);
  return;
}


/*
** Returns initial probability of state j
*/
double pinit(int j) {
  return logInit[j];
}

/*
** Returns self-transition probability for state j
*/
double pextra(int j) {
  return logQ[j];
}

/*
** Returns non-self-transition probability for state j
*/
double pnotextra(int j) {
  return logP[j];
}


/*
** Returns the probability of a transition from I-state x to I-state y, given that x doesn't 
** have a self transition.  There is a small issue that this can happen in two ways for
** a transition from intergene to intergene.  This was handled when the transition matrix was
** read in by treating those as equiprobable and setting that transition to half of its 
** original value.  Just in case you expected the transition matrix rows to sum to 1.
*/
double ptrans(int x, int y) {
  return transitions[x][y];
}


/*
** Returns log(b_state(seq[pos])) for state = intron or intergene.
** b_state(seq[pos]) is defined as in Rabiner.
**
** The value is returned from the lookup table initialized in lookupIntronIntergene.
** 
*/
double bIntronIntergene(int pos, int state, int slen) {
  double p;

  if((pos < 0) || (pos >= slen)) {
    fprintf(stderr,"bIntronIntergene: pos=%d, should be in range [0,%d]\n",pos,slen-1);
    exit(EXIT_FAILURE);
  }

  switch(state) {
    case INTERGENE:
      p = intergenicProbs[pos];
    break;

    case I0F:
    case I1F:
    case I2F:
      p = fIntronProbs[pos];
    break;

    case I0B:
    case I1B:
    case I2B:
      p = bIntronProbs[pos];
    break;

    default:
      fprintf(stderr,"bIntronIntergene: Invalid state: %d\n",state);
      p=-FLT_MAX;
    break;
  }

  return(p);
}


/*
** Scores a candidate exon on the forward strand.  The arguments are...
**
** start            Left end of the region (might be a splice site or exon).
** stop             Right end of the region (might be a splice site or exon).
** prevExtraBases   Number of bases of the first codon that 
**                    were generated in the last exon.
** type             Type of exon (single/initial/internal/terminal).
** str              Pointer to the DNA sequence.
** option           The usual option pointer.
*/
double getFwdExonProb(int start, int stop, int prevExtraBases, int type, char *str, Option *option) {
  double prob;
  double donProb=0;
  double accProb=0;
  double exProb=0;
  double lenProb=0;
  int frame,nStops;
  int exonLeft,exonRight,exonLen,innerLeft,innerRight;

  switch(type) {
    case TYPE_SNGL:
      /* skip the ATG (no score needed, log(prob) = 0) */
      exonLeft  = start;
      innerLeft = start + 3;

      /* cut out and score the stop signal */
      exonRight  = stop;
      innerRight = stop - 3;

      if(((exonRight-exonLeft+1) % PHASES) != 0)
        fprintf(stderr,"getFwdExonProb: Warning, passed a single exon of length %d\n",exonLen);

      prob = getStopProb(str,innerRight+1,FORWARD);
      nStops = 1;
    break;

    case TYPE_INIT:
      /* skip the ATG (no score needed, log(prob) = 0) */
      exonLeft  = start;
      innerLeft = start + 3;

      /* skip and score donor */
      exonRight  = stop - don_dnstr;
      innerRight = stop - don_dnstr - don_upstr;    
      donProb = scoreDonor(str+(innerRight+1),innerRight+1,FORWARD);

      prob = donProb;
      nStops = 0;
    break;

    case TYPE_INTL:
      /* cut out and score acceptor */
      exonLeft  = start + acc_upstr;
      innerLeft = start + acc_upstr + acc_dnstr;
      accProb = scoreAcceptor(str+start,start,FORWARD);

      /* cut out and score donor */
      exonRight  = stop - don_dnstr;
      innerRight = stop - don_dnstr - don_upstr;
      donProb = scoreDonor(str+(innerRight+1),innerRight+1,FORWARD);

      prob = accProb + donProb;
      nStops = 0;
    break;

    case TYPE_TERM:
      /* cut out and score acceptor */
      exonLeft = start + acc_upstr;
      innerLeft = start + acc_upstr + acc_dnstr;
      accProb = scoreAcceptor(str+start,start,FORWARD);

      /* cut out and score the stop signal */
      exonRight  = stop;
      innerRight = stop - 3;

      prob = accProb + getStopProb(str,innerRight+1,FORWARD);
      nStops = 1;
    break;

    default:
      fprintf(stderr,"getFwdExonProb: bad exon-type passed in\n");
      exit(1);
    break;
  }

  /*
  ** At this point, exon boundary scores (ie for startcodon, stopcodon, 
  ** donor or acceptor) have already been added in to prob.
  */

  /* just a warning if innerLeft > innerRight, but this shouldn't happen */
  if(innerLeft > innerRight) {
    fprintf(stderr,"getFwdExonProb: innerLeft > innerRight!  Should not have happened\n");
    printf("  (start,stop) = (%d,%d)\n",start,stop);
    printf("  (innerLeft,innerRight) = (%d,%d)\n",innerLeft,innerRight);
    printf("  (prevExtraBases,type) = (%d,%d)\n",prevExtraBases,type);
    exit(1);
  }

  /* check for in-frame stop codons, unless option->okstops is true */
  frame = (exonLeft - prevExtraBases + PHASES) % PHASES;
  if(!option->okstops)
    if (fStopCount[frame][exonRight] > (nStops + fStopCount[frame][exonLeft-1]))
      return(-FLT_MAX);

  /* score the coding potential of the exon part */
  exProb = fExonProb[frame][innerRight] - fExonProb[frame][innerLeft-1];
  prob += exProb;

  /* score the length */
  exonLen = exonRight - exonLeft + 1;
  if(option->geometric) {
    /* use geometric length distribution */
    lenProb = exonP[type] + (exonQ[type] * (exonLen));
  } else {
    /* use a general length distribution */
    lenProb = exonLenDensity[type][exonLen-1];
  }
  prob += lenProb;
  

  /* debugging code */
  if(option->debug) {
    printf("[%7d,%7d], fwd, f = %d, t = %d, peb = %d, s = %3.10f\n",innerLeft,innerRight,frame,type,prevExtraBases,prob);
    printf("  len[%d] scores %2.15f\n",exonLen,lenProb);
    if(donProb != 0)
      printf("  don[%d,%d] scores %2.15f\n",innerRight+1,innerRight+don_upstr+don_dnstr,donProb);
    if(accProb != 0)
      printf("  acc[%d,%d] scores %2.15f\n",start,start+acc_upstr+acc_dnstr-1,accProb);
    printf("\n");
    /*
    counter++;
    if(counter == 100)
      exit(1);
    */
  }

  return prob;
}


/*
** Scores a candidate exon on the reverse strand.  The arguments are...
**
** left             Left end of the region (might be splice site or exon).
** right            Right end of the region (might be splice site or exon).
** prevExtraBases   Number of bases of the first codon that 
**                    were generated in the last exon.
** type             Type of exon (single/initial/internal/terminal).
** str              Pointer to the DNA sequence.
** option           The usual option pointer.
*/
double getBakExonProb(int left, int right, int prevExtraBases, int type, char *str, Option *option) {
  double prob;
  double donProb=0;
  double accProb=0;
  double exProb=0;
  double lenProb=0;
  int frame,nStops;
  int exonLeft,exonRight,exonLen,innerLeft,innerRight;

  switch(type) {
    case TYPE_SNGL:
      /* cut out and score the stop signal */
      exonLeft  = left;
      innerLeft = left + 3;

      /* skip the ATG (no score needed, log(prob) = 0) */
      exonRight  = right;
      innerRight = right - 3;

      if(((exonRight-exonLeft+1) % PHASES) != 0)
        fprintf(stderr,"getBakExonProb: Warning, passed a single exon of length not a multiple of 3\n");

      prob = getStopProb(str,exonLeft,BACKWARD);
      nStops = 1;
    break;

    case TYPE_INIT:
      /* skip and score donor */
      exonLeft  = left + don_dnstr;
      innerLeft = left + don_dnstr+don_upstr;    
      donProb   = scoreDonor(str+left,left,BACKWARD);

      /* skip the ATG (no score needed, log(prob) = 0) */
      exonRight  = right;
      innerRight = right - 3;

      prob = donProb;
      nStops = 0;
    break;

    case TYPE_INTL:
      /* cut out and score donor */
      exonLeft  = left + don_dnstr;
      innerLeft = left + don_dnstr + don_upstr;
      donProb = scoreDonor(str+left,left,BACKWARD);

      /* cut out and score acceptor */
      exonRight  = right - acc_upstr;
      innerRight = right - acc_upstr - acc_dnstr;
      accProb = scoreAcceptor(str+(innerRight+1),innerRight+1,BACKWARD);

      prob = donProb + accProb;
      nStops = 0;
    break;

    case TYPE_TERM:
      /* cut out and score the stop signal */
      exonLeft = left;
      innerLeft = left + 3;

      /* cut out and score acceptor */
      exonRight  = right - acc_upstr;
      innerRight = right - acc_upstr - acc_dnstr;
      accProb = scoreAcceptor(str+(innerRight+1),innerRight+1,BACKWARD);

      prob = getStopProb(str,left,BACKWARD) + accProb;
      nStops = 1;
    break;

    default:
      fprintf(stderr,"getFwdExonProb: bad exon-type passed in\n");
      exit(1);
    break;
  }

  /*
  ** At this point, exon boundary scores (ie for startcodon, stopcodon, 
  ** donor or acceptor) have already been added in to prob.
  */

  /* just a warning if innerLeft > innerRight, but this shouldn't happen */
  if(innerLeft > innerRight) {
    fprintf(stderr,"getBakExonProb: innerLeft > innerRight!  Should not have happened\n");
    printf("  (left,right) = (%d,%d)\n",left,right);
    printf("  (innerLeft,innerRight) = (%d,%d)\n",innerLeft,innerRight);
    printf("  (prevExtraBases,type) = (%d,%d)\n",prevExtraBases,type);
    exit(1);
  }

  /* check for in-frame stop codons, if option->okstops is not set */
  frame = (exonRight + prevExtraBases + 1) % PHASES;
  if(!option->okstops)
    if (bStopCount[frame][exonRight] > (nStops + bStopCount[frame][exonLeft-1]))
      return(-FLT_MAX);

  /* score the coding potential of the exon part */
  exProb = bExonProb[frame][exonRight] - bExonProb[frame][exonLeft-1];
  prob += exProb;

  /* score the length */
  exonLen = exonRight - exonLeft + 1;
  if(option->geometric) {
    /* use geometric length distribution */
    lenProb = exonP[type] + (exonQ[type] * exonLen);
  } else {
    /* use a general length distribution */
    lenProb = exonLenDensity[type][exonLen-1];
  }
  prob += lenProb;
  

  /* debugging code */
/*
  if(option->debug) {
    printf("[%7d,%7d], bak, f = %d, t = %d, peb = %d, s = %3.10lf\n",exonLeft,exonRight,frame,type,prevExtraBases,exProb);
    printf("  len[%d] scores %2.15lf\n",exonLen,lenProb);
    if(donProb != 0)
      printf("  don[%d,%d] scores %2.15lf\n",left,left+don_upstr+don_dnstr-1,donProb);
    if(accProb != 0)
      printf("  acc[%d,%d] scores %2.15lf\n",exonRight+1,exonRight+acc_upstr+acc_dnstr,accProb);
    printf("\n");
  }
*/

  return prob;
}


/*
** Scans through whole sequence to set the following two arrays...
**   P_intergene (Y_t|Y_0^t-1) = intergeneProbs[t]
**   P_fIntron   (Y_t|Y_0^t-1) = fIntronProbs[t]
**   P_bIntron   (Y_t|Y_0^t-1) = bIntronProbs[t]
** (indices are zero-based)
** 
** This takes up an extra chunk of memory, but avoids repeated calculation of contexts 
** (such as would otherwise happen in calculation of the backward variables, for example).
**
*/
void lookupIntronIntergene(char *str, int slen) {
  int i,context,c;
  double p,q;

  /* 
  ** Allocate new memory
  */
  fIntronProbs = (double *) malloc (slen * sizeof(double));
  if(fIntronProbs == NULL) {
    fprintf(stderr, "lookupIntronIntergene: can't allocate %d doubles for fIntronProbs\n",slen);
    exit(1);
  }

  bIntronProbs = (double *) malloc (slen * sizeof(double));
  if(bIntronProbs == NULL) {
    fprintf(stderr, "lookupIntronIntergene: can't allocate %d doubles for bIntronProbs\n",slen);
    exit(1);
  }

  intergenicProbs = (double *) malloc (slen * sizeof(double));
  if(intergenicProbs == NULL) {
    fprintf(stderr, "lookupIntronIntergene: can't allocate %d doubles for intergenicProbs\n",slen);
    exit(1);
  }

  /* set up the initial context */
  for(context=0, i=-INTRORDER; i<0; i++) {
    context *= N_BASES;
    context += str[i];
  }

  for(i=0; i<slen; i++) {
    c = str[i];

    /* 
    ** Intergene: only take the log if the word actually occurs
    */
    p = logIntergenicFreq[context][c];
    if(p == 0) {
      q = (double) intergenicFreq[context][c] /
        (double) intergenicFreq[context][N_BASES];
      p = log(q);
      logIntergenicFreq[context][c] = p;
    }
    intergenicProbs[i] = logIntergenicFreq[context][c];

    /* 
    ** Intron: only take the log if the word actually occurs
    */
    p = fLogIntronFreq[context][c];
    if(p == 0) {
      q = (double) fIntronFreq[context][c] / (double) fIntronFreq[context][N_BASES];
      p = log(q);
      fLogIntronFreq[context][c] = p;

      q = (double) bIntronFreq[context][c] / (double) bIntronFreq[context][N_BASES];
      p = log(q);
      bLogIntronFreq[context][c] = p;
    }
    fIntronProbs[i] = fLogIntronFreq[context][c];
    bIntronProbs[i] = bLogIntronFreq[context][c];

    /* advance the context */
    context *= N_BASES;
    context %= INTRPRED;
    context += c;
  }

  return;
}



/* 
** In fExonProb and bExonProb are both arrays of length slen+1, though
** fExonProb will point to the second element as the first will just have
** value 0 (ie log(1)) to make the looping work.  bExonProb just points at
** the first element of the array but its last value will be 0, corresponding
** to position slen+1 (again, just to make the looping work).
** The same applies to fStopCount and bStopCount.
**
** The end result is that we can treat the indices as zero-based, as usual.
**   fExonProb[j][t] will give P(seq[0..t] | forward coding in frame j)
**   bExonProb[j][t] will give P(seq[0..t] | backward coding in frame j)
**   fStopCount[j][t] = # forward stop codons up to and including t, frame j.
**   bStopCount[j][t] = # backward stop codons up to and including t, frame j.
** 
** Frame is defined relative to the start (ie left end) of the sequence, 
** whether we're looking at the forward or reverse direction.
** 
** Note that the term "phase" below can be used somewhat interchangeably with
** "frame".
*/
void lookupExon(char *str,int slen, Option *option) {
  int i,frame,context,c,nContexts;
  int nStopF,nStopB,tri_nuc;
  int phase[PHASES];
  double *tempPointer;
  int *intPointer;

  nContexts = (int) pow((double) N_BASES, (double) exonOrder);

  /* 
  ** Allocate memory for each frame
  */
  for(i=0;i<PHASES;i++) {
    /* forward */
    tempPointer = (double *) malloc ((slen+1) * sizeof(double));
    if(tempPointer == NULL) {
      fprintf(stderr, "lookupExon: can't allocate %d doubles for fExonProb[%d]\n",slen,i);
      exit(1);
    }
    fExonProb[i] = tempPointer+1;

    intPointer = (int *) malloc ((slen+1) * sizeof(int));
    if(intPointer == NULL) {
      fprintf(stderr, "lookupExon: can't allocate %d ints for fStopCount[%d]\n",slen+1,i);
      exit(1);
    }
    fStopCount[i] = intPointer+1;

    /* backward */
    tempPointer = (double *) malloc ((slen+1) * sizeof(double));
    if(tempPointer == NULL) {
      fprintf(stderr, "lookupExon: can't allocate %d doubles for bExonProb[%d]\n",slen,i);
      exit(1);
    }
    bExonProb[i] = tempPointer+1;

    intPointer = (int *) malloc ((slen+1) * sizeof(int));
    if(intPointer == NULL) {
      fprintf(stderr, "lookupExon: can't allocate %d ints for bStopCount[%d]\n",slen+1,i);
      exit(1);
    }
    bStopCount[i] = intPointer+1;
  }


  /*
  ** Score coding sequence
  */

  /* 
  ** Initialize the phases for each frame.
  **   for frame 0 the first base is in phase 0
  **   for frame 1 the first base is in phase 2
  **   for frame 2 the first base is in phase 1
  ** Therefore the phase is given by (PHASES-frame) % PHASES
  */
  for(frame=0; frame<PHASES; frame++) {
    phase[frame] = (PHASES-frame) % PHASES;
    fExonProb[frame][-1]  = 0;
    bExonProb[frame][-1]  = 0;
    fStopCount[frame][-1] = 0;
    bStopCount[frame][-1] = 0;
  }

  /* get initial context */
  for(context=0, i=-ORDER; i<0; i++) {
    context *= N_BASES;
    context += str[i];
  }

  /* score the sequence */
  nStopF = nStopB = 0;
  for(i=0,tri_nuc=0; i<slen; i++) {
    c = str[i];

    /* get new probs */
    for(frame=0; frame<PHASES; frame++) {
      fExonProb[frame][i] = fExonProb[frame][i-1] + fLogExonFreq[phase[frame]][c][context];
      bExonProb[frame][i] = bExonProb[frame][i-1] + bLogExonFreq[phase[frame]][c][context];
      fStopCount[frame][i] = fStopCount[frame][i-1];
      bStopCount[frame][i] = bStopCount[frame][i-1];
    }

    /* check for new stop codons */
    tri_nuc *= N_BASES;
    tri_nuc %= N_TRI_NUC;
    tri_nuc += c;
    if(isfstop(tri_nuc)) {
      fStopCount[(i+1) % PHASES][i] += 1;
      nStopF++;
    }
    if(isbstop(tri_nuc)) {
      bStopCount[(i+1) % PHASES][i] += 1;
      nStopB++;
    }

    for(frame=0; frame<PHASES; frame++) {
      phase[frame]++;
      phase[frame] %= PHASES;
    }
    context *= N_BASES;
    context %= nContexts;
    context += c;
  }
  if(option->verbose)
    printf("In lookupExon: found %d forward and %d reverse stops\n",nStopF,nStopB);

  return;
}


/*
** Read a length density from filename.
** Returns a poitner to newly allocated memory holding the log of the density.
** If tertiaryNormalise is 1 then it will be normalized such that the sums
** over length modulo 3 will be 1.  Otherwise it is normalized in the usual manner.
** The max len of the density is written into *n.
*/
double *readLengthDist(int *n, char *fileName, int tertiaryNormalise, int verbose) {
  FILE *dataFile;
  char line[MAX_LINE_LEN];
  int i,j;
  double sum[PHASES];
  double *lenDist,*lenDistPointer;

  if(verbose) printf("Reading length density from %s\n",fileName);
  
  if ((dataFile = fopen(fileName,"r")) == NULL) {
    fprintf(stderr,"readLengthDist: can't open %s for read, exiting\n",fileName);
    exit(1);
  }

  /* find out how many entries there are to be read */
  fgets(line,MAX_LINE_LEN,dataFile);
  sscanf(line,"%d ",n);
  if(verbose) printf("  (max length is %d)\n",*n);

  /* declare the memory for lenDist */
  if(NULL == (lenDist = (double *) malloc ((*n) * sizeof(double))) ) {
    fprintf(stderr,"readLengthDist: unable to get mem for %d doubles, exiting\n",*n);
    exit(EXIT_FAILURE);
  }

  /* read in the entries */
  for(i=0,lenDistPointer=lenDist;  i<*n;  i++,lenDistPointer++) {
    if(! fgets(line,MAX_LINE_LEN,dataFile)) {
      fprintf(stderr,"readLengthDist: file %s shorter than expected\n  (On line %d, expected %d)\n",fileName,i+2,*n+1);
      exit(EXIT_FAILURE);
    }
    if(1 != (sscanf(line,"%lf ",lenDistPointer)) ) {
      fprintf(stderr,"readLengthDist: couln't read a double from line %d of %s\n",i+2,fileName);
      exit(EXIT_FAILURE);
    }
    /* *lenDistPointer = exp(*lenDistPointer); */
  }
  fclose(dataFile);

  if(tertiaryNormalise) {
    /* Normalize the distribution such that  sum_{i mod PHASES = c} len_i = 1  for c=0,1,2 */
    /* This is for exon length distributions, as their length mod PHASES is fixed */

    /* initialize the sums to zero */
    for(j=0; j<PHASES; j++)
      sum[j] = 0;

    /* calculate the sums */
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++) {
      sum[j++] += *lenDistPointer++;
      j = j % PHASES;
    }

    /*
    for(j=0; j<PHASES; j++)
      printf("  sum[%d] = %g\n",j,sum[j]);
    */

    /* normalize the densities */
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++,lenDistPointer++) {
      *lenDistPointer /= sum[j++];
      *lenDistPointer = (*lenDistPointer > 0) ? log(*lenDistPointer) : -FLT_MAX;
      j = j % PHASES;
    }

    /* A check to see if everything was normalized OK
    for(j=0; j<PHASES; j++)
      sum[j] = 0;
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++) {
      sum[j++] += exp(*lenDistPointer++);
      j = j % PHASES;
    }
    for(j=0; j<PHASES; j++)
      printf("  sum[%d] = %g\n",j,sum[j]);
    */

  } else {
    /* normalize the distribution in the usual manner */

    /* calculate the sum */
    sum[0] = 0;
    lenDistPointer=lenDist;
    for(i=0; i<*n; i++)
      sum[0] += *lenDistPointer++;
    /* printf("  sum = %g\n",sum[0]); */

    /* normalize the density */
    lenDistPointer=lenDist;
    for(i=0; i<*n; i++,lenDistPointer++) {
      *lenDistPointer /= sum[0];
      *lenDistPointer = (*lenDistPointer > 0) ? log(*lenDistPointer) : -FLT_MAX;
    }
  }

  return(lenDist);
}


/*
** return the max possible length of an exon of the given type.
*/
int maxExonLen(int type) {
  switch(type) {
    case TYPE_SNGL:
      return(numExons[TYPE_SNGL]);
    break;

    case TYPE_INIT:
      return(numExons[TYPE_INIT]);
    break;

    case TYPE_INTL:
      return(numExons[TYPE_INTL]);
    break;

    case TYPE_TERM:
      return(numExons[TYPE_TERM]);
    break;

    default:
      fprintf(stderr,"maxExonLen: bad type passed in: %d\n",type);
      exit(EXIT_FAILURE);
    break;
  }
}

double getStopProb(char *str, int loc, int dir) {
  int i,j,context;
  double ans;

  if(dir == FORWARD) {
    for(i=0,j=loc,context=0; i<STOPLEN; i++,j++) {
      context *= N_BASES;
      context += str[j];
    }
  } else if (dir == BACKWARD) {
    for(i=0,j=loc+STOPLEN-1,context=0; i<STOPLEN; i++,j--) {
      context *= N_BASES;
      context += (N_BASES-1) - str[j];
    }
  } else {
    fprintf(stderr,"getStopProb: bad direction passed in\n");
    exit(EXIT_FAILURE);
  }

  switch(context) {
    case 48:            /* TAA */
      ans = logStop[0];  
    break;
    
    case 50:            /* TAG */
      ans = logStop[1];
    break;
    
    case 56:            /* TGA */
      ans = logStop[2];
    break;
    
    default:            /* none of the above */
      fprintf(stderr,"getStopProb: can't score a non-stop codon\n");
      fprintf(stderr,"  (This shouldn't have happened)\n");
      fprintf(stderr,"  loc is %d, context = %d, dir = %d\n",loc,context,dir);
      fprintf(stderr,"  bases are %d, %d, %d\n",str[loc],str[loc+1],str[loc+2]);
      exit(EXIT_FAILURE);
    break;
  }
    
  return(ans);
}


void free3dArray( double ***array, int dim1, int dim2) {
  int i,j;
  if(array != NULL) {
    for(i=0; i<dim1; i++) {
      if(array[i] != NULL) {
        for(j=0; j<dim2; j++) {
          if(array[i][j] != NULL) {
            free(array[i][j]);
          }
        }
      }
      free(array[i]);
    }
    free(array);
  }
  return;
}


double ***declare3dArray(int dim1,int dim2,int dim3) {
  int i,j;
  double ***x;

  x = (double ***) myAlloc(dim1 * sizeof(double **));
  for(i=0; i<dim1; i++) {
    x[i] = (double **) myAlloc(dim2 * sizeof(double *));
    for(j=0; j<dim2; j++) {
      x[i][j] = (double *) myAlloc(dim3 * sizeof(double));
    }
  }

  return(x);
}

