/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#define EXTERN extern
#include "full_common.h"
#include "full_splice.h"
#include "full_probs.h"
#undef EXTERN

#define EXTERN
#include "full_seq.h"
#undef EXTERN

#define PROB_SIG_FIG 2

/* 
** An object for backtracking in the Viterbi algorithm
*/
typedef struct {
  int state;
  int loc;
  short type;
} backtrackStructure;

/*
** Sequence data:
** *sequence is used only once, it points to the start of the memory
** dynamically allocated for the DNA.  *seq is the array used everywhere
** else, it points 6 bases in to sequence, allowing for putting an extra
** 5 bases at the start (which are arbitrarily duplicated from the end of
** the sequence).
**
** slen is the length of the DNA sequence being analyzed.
*/
static char *sequence;
static char *seq;
static int slen;

/*
** Exon boundary coordinate arrays, hold locations of potential exon boundaries.
** See function locateBounds (in this file) for exact definitions.
*/
static int *fstart;
static int *facc;
static int *fdon;
static int *fstop;
static int *bstart;
static int *bacc;
static int *bdon;
static int *bstop;

/*
** The sizes of the above dynamically-allocated exon boundary coordinate arrays.
** See function locateBounds (in this file) for exact definitions.
*/
static int nFstart;
static int nFacc;
static int nFdon;
static int nFstop;
static int nBstart;
static int nBacc;
static int nBdon;
static int nBstop;

/*
** Global (to this file) pointers to the exon boundary coordinate arrays.
** See function locateBounds (in this file) for exact definitions.
*/
static int curFstart;
static int curFacc;
static int curFdon;
static int curFstop;
static int curBstart;
static int curBacc;
static int curBdon;
static int curBstop;


/*
** forward/backward vars and Viterbi stuff - will document properly shortly
*/
static backtrackStructure *backtrack[NUMSTATES];
static double *fb[NUMSTATES];
static double *delta[NUMSTATES];
static double *al[NUMSTATES];
static double *alpha[NUMSTATES];
static double *be[NUMSTATES];
static double *beta[NUMSTATES];
static double likelihood;




int readSequence(Option *option) {
  FILE *seqFile=NULL;
  /* int y,i,done,c,n,cNum; */
  int base,i,cNum;
  int format;
  
  /*
  ** Allocate memory for sequence and set up the pointers.
  ** sequence will point to the start of memory, seq will point
  ** to the start of the sequence, 5 bases further on to allow for the
  ** cycling round of the last 5 bases to the start.
  */
  slen = getSeqLen(option->inFile);
  sequence = (char *) malloc((slen+EXTRA) * sizeof(char));
  if(sequence == NULL) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  seq = sequence + EXTRA;


  /*
  ** determine format of sequence file
  */
  if( (format = getSeqFormat(option->inFile)) == UNKNOWN_FORMAT) {
    fprintf(stderr,"getSeqLen: bad format in %s",option->inFile);
    exit(EXIT_FAILURE);
  }

  /*
  ** read in the sequence 
  */
  openSeq(&seqFile,option->inFile,format);
  for( i=0, base = getNextBase(&seqFile,format); 
       base != EOF; 
       i++, base = getNextBase(&seqFile,format) ) {
    cNum = base2num(base);
    if(cNum > 3) {
      fprintf(stderr,"Error: readSequence: can't handle non-ACGT bases in %s\n",option->inFile);
      fprintf(stderr,"  (read character %c, treating it as \"A\")\n",base);
      cNum = 0;
    }
    seq[i] = cNum;
    if(i > slen) {
      fprintf(stderr,"readSequence: input sequence longer than I thought it was!\n");
      exit(EXIT_FAILURE);
    }
  }
  fclose(seqFile);

  for(i=-EXTRA;i<0;i++)
    seq[i] = seq[slen+i];

  /*
  ** Allocate the memory for the gamma (used in Viterbi), alpha (forward)
  ** and beta (backward) variables.
  */
  for(i=0;i<NUMSTATES;i++) {
    fb[i] = (double *) malloc((slen+2)*sizeof(double));
    backtrack[i] = (backtrackStructure *) malloc((slen+1)*
        (sizeof(backtrackStructure)));
    if(fb[i] == NULL || backtrack[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    delta[i] = fb[i]+1;
    al[i] = (double *) malloc((slen+2)*sizeof(double));
    if(al[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    alpha[i] = al[i]+1;
    be[i] = (double *) malloc((slen+2)*sizeof(double));
    if(be[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    beta[i] = be[i]+1;
  }

  if(option->verbose) printf("Builing exon probability lookup table...\n");
  lookupExon(seq,slen,option);
  if(option->verbose) printf("Builing intron/intergene probability lookup table...\n");
  lookupIntronIntergene(seq,slen);

  return slen;
}

/*
** returns the length of the sequence to be analyzed
*/
int getSeqLen(char *inFile) {
  FILE *seqFile;
  int format,seqLen;
  int base;

  if( (format = getSeqFormat(inFile)) == UNKNOWN_FORMAT) {
    fprintf(stderr,"getSeqLen: bad format in %s",inFile);
    exit(EXIT_FAILURE);
  }

  openSeq(&seqFile,inFile,format);
  base = getNextBase(&seqFile,format);
  seqLen = 0;
  while (base != EOF) {
    seqLen++;
    base = getNextBase(&seqFile,format);
  }
  fclose(seqFile);

  return(seqLen);
}


/*
** opens inFile (whose format is specified by the int format) and
** returns the first base.  FILE *seqFile will now point to the next
** character in the file
*/
void openSeq(FILE **seqFile, char *inFile, int format) {
  char word[MAX_WORD_LEN];
  char line[MAX_LINE_LEN];

  if (((*seqFile) = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"openSeq: can't open %s for read, exiting\n",inFile);
    exit(1);
  }


  /* 
  ** figure out the format, skip over header to the start of sequence
  */
  if(NULL == fgets(line,MAX_LINE_LEN,*seqFile)) {
    fprintf(stderr,"Error: openSeq: problem reading first line of %s \n",inFile);
    exit(1);
  }

  if(1 != sscanf(line,"%s",word)) {
    fprintf(stderr,"Error: openSeq: first line of %s is blank\n",inFile);
    exit(1);
  }

  if(format==FASTA) {
    /* already read first line, nothing more to do */
   return;
  } else if (format==GENBANK) {
    /* read everything up to the line starting with "ORIGIN" */
    for(;;) {
      /* read ahead to skip over the header */
      if(NULL == fgets(line,MAX_LINE_LEN,*seqFile)) {
        fprintf(stderr,"openSeq: incorrect genbank format in %s\n",inFile);
        fprintf(stderr,"  (never found keyword \"ORIGIN\")\n");
        exit(1);
      }
      sscanf(line,"%s",word);
      if(0 == strcmp(word,"ORIGIN"))
        break;
    }
    return;
  } else {
    fprintf(stderr,"openSeq: don't know how to process format %d\n",format);
    exit(EXIT_FAILURE);
  }
}


/*
** return the next character that is supposed to be a DNA base in 
** the file.  This could still be an N or a * or something, checking 
** isn't done here.  EOF returned when EOF or the end of the sequence
** is reached.
*/
char getNextBase(FILE **seqFile, int format) {
  int c;

  if(format == FASTA) {
    c = fgetc(*seqFile);
    for( ; ; c = fgetc(*seqFile) ) {
      if(c == EOF) {
        break;
      } else if(isspace(c)) {
        continue;
      } else if(c == '>') {
        /* ignore second sequence, treat as EOF */
        c = EOF;
        break;
      } else {
        break;
      }
    }
  } else if (format == GENBANK) {
    c = fgetc(*seqFile);
    for( ; ; c = fgetc(*seqFile) ) {
      if( (c == EOF) || (c == '/') ) {
        c = EOF;
        break;
      } else if( isspace(c) || isdigit(c) ) {
        continue;
      } else {
        break;
      }
    }
  }

  return(c);
}


/*
** opens inFile and determines whether the format is FASTA, GENBANK
** or UNKNOWN_FORMAT
*/
int getSeqFormat(char *inFile) {
  FILE *seqFile;
  char word[MAX_WORD_LEN];
  char line[MAX_LINE_LEN];
  int fileFormat;

  if ((seqFile = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"getSeqFormat: can't open %s for read, exiting\n",inFile);
    exit(1);
  }

  /* 
  ** figure out format, skip over the header to the start of sequence
  */
  if(NULL == fgets(line,MAX_LINE_LEN,seqFile)) {
    fprintf(stderr,"Error: getSeqFormat: problem reading first line of %s\n",inFile);
    exit(1);
  }

  if(1 != sscanf(line,"%s",word)) {
    fprintf(stderr,"Error: getSeqFormat: first line of %s is blank\n",inFile);
    exit(1);
  }

  if(word[0] == '>') {
    /* file is assumed to be in fasta format */
    fileFormat = FASTA;
  } else if (0 == strcmp(word,"LOCUS")) {
    /* file is assumed to be in genbank format */
    fileFormat = GENBANK;
  } else {
    fileFormat = UNKNOWN_FORMAT;
  }

  /* printf("fileFormat = %d\n",fileFormat); */
  fclose(seqFile);
  return(fileFormat);
}



/*
** locates all the potential start codons, stop codons, donor sites 
** and acceptor sites in the sequence, and records their numbers 
** and locations.
**
** The logic (if any) behind the coordinate system is explained below.
** The convention used is that "left" and "right" ends of an exon are
** the lower and higher coordinates respectively in the original 
** sequence, irrespective of which strand the exon is on.
**
** For signals occuring at the left end of an exon, we store the first
** base of the signal.  For signals occuring at the right end of an
** exon, we store the base AFTER the last base of the signal.  This is 
** tabulated below:
**
** fstart[] = first base of start codon
** fdon[]   = first base after donor
** facc[]   = first base of acceptor
** fstop[]  = first base after stop codon
** bstop[]  = first (ie leftmost) base of stop codon 
** bacc[]   = first (ie leftmost) base after acceptor
** bdon[]   = first (ie leftmost) base of donor
** bstart[] = first (ie leftmost) base after start codon
*/
void locateBounds(Option *option) {
  int i,leftSide;
  int dinuc,trinuc;

  nFstart = 0;
  nFacc   = 0;
  nFdon   = 0;
  nFstop  = 0;
  nBstart = 0;
  nBacc   = 0;
  nBdon   = 0;
  nBstop  = 0;

  /*
  ** Firstly compute the number of minimal signals of each type,
  ** so we can declare just enough memory for them.  Note, the way
  ** the nubmer of minimal signals is computed MUST match the way
  ** they are initialized just afterwards.  So if this code is altered
  ** at all, so must the later bit, and in the exact same manner.
  */ 
  for(trinuc=0, i=0; i < DI_NUC; i++) {
    trinuc *= N_BASES;
    trinuc += seq[i];
  }
  /* (not checking for a splice site in the first 2 bases) */
  for( ; i < slen; i++) {
    trinuc *= N_BASES;
    trinuc %= N_TRI_NUC;
    trinuc += seq[i];
    dinuc = trinuc % N_DI_NUC;

    /* forward strand */
    if(isfstart(trinuc))
      nFstart++;
    if(isfstop(trinuc))
      nFstop++;
    if( (isfacc(dinuc)) && (i-acc_upstr+1 > -1) && (i+acc_dnstr < slen) )
      nFacc++;
    if( (isfdon(dinuc)) && (i-DI_NUC-don_upstr+1 > -1) && (i-DI_NUC+don_dnstr < slen) )
      nFdon++;

    /* backward strand */
    if(isbstart(trinuc))
      nBstart++;
    if(isbstop(trinuc))
      nBstop++;
    if( (isbacc(dinuc)) && (i-DI_NUC-acc_dnstr+1 > -1) && (i-DI_NUC+acc_upstr < slen) )
      nBacc++;
    if( (isbdon(dinuc)) && (i-don_dnstr+1 > -1) && (i+don_upstr < slen) )
      nBdon++;
  }


  /*
  ** Allocate the memory for the boundaries
  */
  if((fstart = (int *) malloc(nFstart * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((fdon = (int *) malloc(nFdon * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((facc = (int *) malloc(nFacc * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((fstop = (int *) malloc(nFstop * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }

  if((bstop = (int *) malloc(nBstop * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((bacc = (int *) malloc(nBacc * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((bdon = (int *) malloc(nBdon * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  if((bstart = (int *) malloc(nBstart * sizeof(int))) == NULL ) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }

  /*
  ** Allocate memory for splice site scores
  */
  donLookupInit(slen,nFdon,nBdon,option);
  accLookupInit(slen,nFacc,nBacc,option);

  /*
  ** Go through the sequence again, this time store the locations
  ** and score the splice sites.
  */
  curFstart = 0;
  curFacc   = 0;
  curFdon   = 0;
  curFstop  = 0;
  curBstart = 0;
  curBacc   = 0;
  curBdon   = 0;
  curBstop  = 0;

  for(trinuc=0, i=0; i < DI_NUC; i++) {
    trinuc *= N_BASES;
    trinuc += seq[i];
  }
  /* (not checking for a splice site in the first 2 bases) */
  for( ; i < slen; i++) {
    trinuc *= N_BASES;
    trinuc %= N_TRI_NUC;
    trinuc += seq[i];
    dinuc = trinuc % N_DI_NUC;

    /* forward strand */
    if(isfstart(trinuc))
      fstart[curFstart++] = i-TRI_NUC+1;      /* first base of start signal */
    if(isfstop(trinuc))
      fstop[curFstop++] = i+1;                /* first base after stop signal */ 
    if( (isfacc(dinuc)) && (i-acc_upstr+1 > -1) && (i+acc_dnstr < slen) ) {
      leftSide = i-acc_upstr+1;        /* first base of acceptor */
      facc[curFacc++] = leftSide;
      scoreAcceptor(seq+leftSide,leftSide,FORWARD);
    }
    if( (isfdon(dinuc)) && (i-DI_NUC-don_upstr+1 > -1) && (i-DI_NUC+don_dnstr < slen) ) {
      fdon[curFdon++] = i-DI_NUC+don_dnstr+1;   /* first base after donor */
      leftSide = i-DI_NUC-don_upstr+1;          /* first base of donor */
      scoreDonor(seq+leftSide,leftSide,FORWARD);
    }

    /* backward strand */
    if(isbstart(trinuc))
      bstart[curBstart++] = i+1;  /* first base after start signal (after in the sense of left to right) */
    if(isbstop(trinuc))
      bstop[curBstop++] = i-TRI_NUC+1;  /* first base of stop signal */
    if( (isbacc(dinuc)) && (i-DI_NUC-acc_dnstr+1 > -1) && (i-DI_NUC+acc_upstr < slen) ) {
      bacc[curBacc++] = i-DI_NUC+acc_upstr+1; /* first base after acceptor signal */
      leftSide = i-DI_NUC-acc_dnstr+1;        /* first base of acceptor signal */
      scoreAcceptor(seq+leftSide,leftSide,BACKWARD);
    }
    if( (isbdon(dinuc)) && (i-don_dnstr+1 > -1) && (i+don_upstr < slen) ) {
      leftSide = i-don_dnstr+1;     /* first base of donor signal */
      bdon[curBdon++] = leftSide;
      scoreDonor(seq+leftSide,leftSide,BACKWARD);
    }
  }

  /* left ends of exons */
  curFstart = nFstart-1;
  curFacc   = nFacc-1;
  curBstop  = nBstop-1;
  curBdon   = nBdon-1;

  /* right ends of exons */
  curFdon   = 0;
  curFstop  = 0;
  curBstart = 0;
  curBacc   = 0;

  if (option->verbose) {
    printf("Candidate exon boundaries:\n");
    printf("%7d potential forward start codons\n",nFstart);
    printf("%7d potential forward donors\n",nFdon);
    printf("%7d potential forward acceptors\n",nFacc);
    printf("%7d potential forward stop codons\n",nFstop);
    printf("%7d potential backward stop codons\n",nBstop);
    printf("%7d potential backward acceptors\n",nBacc);
    printf("%7d potential backward donors\n",nBdon);
    printf("%7d potential backward start codons\n",nBstart);
    printf("\n");
  }

  return;
}


/*
** Implements the Viterbi algorithm and computes the forward variables.
** See the phat writeup for details, the notation used is fairly consistent.
*/
int viterbi(Option *option) {
  int i;             /* index ranging over the I-states of the model */
  int t;             /* index ranging over the DNA sequence */
  double aii;        /* a_{i,i} */
  double bsi;        /* b_{s_i} (Y_t | Y_1^{t-1}) */
  double deltaMaxA;  /* first term in the maximization in Viterbi recursion */
  double deltaMaxB;  /* second term in the maximization in Viterbi recursion */
  double deltaMax;   /* overall max value of delta[i][t] */
  int last_state;    /* the final I-state attaining deltaMax */

  /* initialization */
  for(i=0; i<NUMSTATES; i++) {
    delta[i][-1]   = pinit(i);
    alpha[i][-1]   = pinit(i);
    backtrack[i][-1].state = i;
    backtrack[i][-1].loc = -1;
    backtrack[i][-1].type = NADA;
  }

  /* iteration */
  for(t=0; t<slen; t++) {
    for(i=0; i<NUMSTATES; i++) {
      bsi = bIntronIntergene(t,i,slen); /* g factors out in forward/Viterbi, add in later */
      aii = pextra(i);                  /* probability of self-transition in the I-state  */

      deltaMaxA   = delta[i][t-1] + aii;
      alpha[i][t] = alpha[i][t-1] + aii;

      deltaMaxB = -FLT_MAX;
      switch(i) {
        case INTERGENE:
        deltaMaxB = endforward(t,deltaMaxB,option);
        deltaMaxB = startbackward(t,deltaMaxB,option);
        break;

        case I0F:
        case I1F:
        case I2F:
        deltaMaxB = donorforward(t,i,deltaMaxB,option);
        break;

        case I0B:
        case I1B:
        case I2B:
        deltaMaxB = acceptorbackward(t,i,deltaMaxB,option);
        break;

        default:
        fprintf(stderr,"Error in viterbi: bad state\n");
        break;
      }
      if(deltaMaxA > deltaMaxB) {
        backtrack[i][t].state = i;
        backtrack[i][t].loc = t-1;
        backtrack[i][t].type = NADA;
        delta[i][t] = deltaMaxA + bsi;
      } else {
        delta[i][t] = deltaMaxB + bsi;
      }
      alpha[i][t] += bsi;
    }
  }

  /* termination */
  deltaMax = -FLT_MAX;
  for(i=0; i<NUMSTATES; i++) {
    aii = pextra(i);
    /* final state is silent - no bsi term here */

    deltaMaxA      = delta[i][slen-1] + aii;
    alpha[i][slen] = alpha[i][slen-1] + aii;

    deltaMaxB = -FLT_MAX;
    switch(i) {
      case INTERGENE:
      deltaMaxB = endforward(slen,deltaMaxB,option);
      deltaMaxB = startbackward(slen,deltaMaxB,option);
      break;

      case I0F:
      case I1F:
      case I2F:
      deltaMaxB = donorforward(slen,i,deltaMaxB,option);
      break;

      case I0B:
      case I1B:
      case I2B:
      deltaMaxB = acceptorbackward(slen,i,deltaMaxB,option);
      break;

      default:
      fprintf(stderr,"Error in viterbi (termination loop) bad state\n");
      break;
    }
    if(deltaMaxA > deltaMaxB) {
      backtrack[i][slen].state = i;
      backtrack[i][slen].loc = slen-1;
      backtrack[i][slen].type = NADA;
      delta[i][slen] = deltaMaxA;
    } else {
      delta[i][slen] = deltaMaxB;
    }
    if(delta[i][slen] > deltaMax) {
      deltaMax = delta[i][slen];
      last_state = i;
    }
  }
  
  /*
  ** Compute the likelihood.
  */
  for(likelihood=alpha[0][slen], i=1; i<NUMSTATES; i++)
    myadd(alpha[i][slen],likelihood);

  return last_state;
}


/*
** Calculate the backward variables, beta.
** See the phat writeup for details, the notation used is fairly consistent.
*/
void backward(Option *option) {
  int t,i;
  double test1,test2;

  /* initialization */
  for(i=0; i<NUMSTATES; i++) {
    beta[i][slen] = 0;
    beta[i][slen-1] = pextra(i);
  }

  /* iteration */
  for(t=slen-2; t >= -1; t--) {
    for(i=0; i<NUMSTATES; i++) {
      /* probability if continuing state  */
      beta[i][t] = pextra(i) + beta[i][t+1] + bIntronIntergene(t+1,i,slen);

      switch(i) {
        case INTERGENE:
	startforward(t+1,option);
        endbackward(t+1,option);
        break;

        case I0F:
        case I1F:
        case I2F:
	acceptorforward(t+1,i,option);
        break;

        case I0B:
        case I1B:
        case I2B:
	donorbackward(t+1,i,option);
        break;

        default:
        fprintf(stderr,"Error in backward: bad state\n");
        break;
      }
    }
  }

  /* 
  ** A check: the following 2 quantities should be equal...
  ** sum_i exp(beta[i][-1] + pinit(i))
  ** sum_i exp(alpha[i][slen])
  */
  if(option->verbose) {
    test1 = alpha[0][slen];
    test2 = beta[0][-1] + pinit(0);
    for(i=1; i<NUMSTATES; i++) {
      myadd(alpha[i][slen],test1);
      myadd(beta[i][-1] + pinit(i),test2);
    }
    printf("likelihood from alphas = %2.15f\n",test1);
    printf("likelihood from betas  = %2.15f\n",test2);
    printf("\n");
  }

  return;
}



/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible terminal and single exons in the forward direction.
*/
double endforward(int t, double deltaMaxB, Option *option) {
  int i,j,seqLen,exonLen,minLen;
  int prevExtraBases,lastState;
  double prob;
  double tempProb;

  if(t != fstop[curFstop] || curFstop >= nFstop)
    return(deltaMaxB);
  else
    curFstop++;
  /* t points to the first base _after_ the stop codon */

  /* Sum over all possible single exons ending at t-1 */
  minLen = START_SIGNAL_LEN + STOP_SIGNAL_LEN + 1;
  for(i=0; i<nFstart; i++) {
    j = fstart[i];
    seqLen = t-j;

    /* skip to next if exon too long */
    if( seqLen > maxExonLen(TYPE_SNGL)) continue;

    /* stop if the single exon is too short (as all the rest will be too) */
    if( seqLen < minLen) break;

    /* skip if length not a multiple of 3 */
    if((seqLen % PHASES) != 0) continue;

    prevExtraBases = 0;
    prob = getFwdExonProb(j,t-1,prevExtraBases,TYPE_SNGL,seq,option);
    if(prob > -FLT_MAX) {
      /* the pnotextra below accounts for not having another self-transition in intergene */
      prob += ptrans(INTERGENE,INTERGENE) + pnotextra(INTERGENE);
  
      tempProb = prob + alpha[INTERGENE][j-1];
      myadd(tempProb,alpha[INTERGENE][t]);
  
      prob += delta[INTERGENE][j-1];
      if(prob > deltaMaxB) {
        deltaMaxB = prob;
        (backtrack[INTERGENE][t]).state=INTERGENE;
        (backtrack[INTERGENE][t]).loc=j-1;
        (backtrack[INTERGENE][t]).type=SNGL_F;
      }
    }
  }

  /* Sum over all possible terminal exons ending at t-1 */
  minLen = acc_upstr + acc_dnstr + STOP_SIGNAL_LEN + 1;
  for(i=0; i<nFacc; i++) {
    j = facc[i];
    seqLen = t-j;

    /* skip if the terminal exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM) ) continue;

    /* stop if the terminal exon would be too short */
    if( seqLen < minLen) break;

    exonLen = seqLen-acc_upstr;
    prevExtraBases = (PHASES - (exonLen % PHASES) ) % PHASES;
    lastState = I0F+prevExtraBases;

    prob = getFwdExonProb(j,t-1,prevExtraBases,TYPE_TERM,seq,option);
    if(prob > -FLT_MAX) {
      prob += ptrans(lastState,INTERGENE) + pnotextra(lastState);

      tempProb = prob + alpha[lastState][j-1];
      myadd(tempProb,alpha[INTERGENE][t]);

      prob += delta[lastState][j-1];
      if(prob > deltaMaxB) {
        deltaMaxB = prob;
        backtrack[INTERGENE][t].state=lastState;
        backtrack[INTERGENE][t].loc=j-1;
        backtrack[INTERGENE][t].type=TERM_F;
      }
    }
  }

  return(deltaMaxB);
}


/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible initial and single exons in the backward direction.
*/
double startbackward(int t, double deltaMaxB, Option *option) {
  int i,j,seqLen,exonLen,minLen;
  int prevExtraBases,thisExtraBases,leftState;
  double prob;
  double tempProb;

  if(t != bstart[curBstart] || curBstart >= nBstart)
    return(deltaMaxB);
  else
    curBstart++;
  /* t points to the first base to the right of the backward stop codon */

  /* Sum over all possible backward single exons whose right coordinate is t-1 */
  minLen = START_SIGNAL_LEN + STOP_SIGNAL_LEN + 1;
  for(i=0; i<nBstop; i++) {
    j = bstop[i];
    seqLen = t-j;

    /* skip to next if exon too long */
    if( seqLen > maxExonLen(TYPE_SNGL)) continue;

    /* stop if the single exon is too short (as all the rest will be too) */
    if( seqLen < minLen) break;

    /* skip if length not a multiple of 3 */
    if((seqLen % PHASES) != 0) continue;

    prevExtraBases = 0;
    prob = getBakExonProb(j,t-1,prevExtraBases,TYPE_SNGL,seq,option);

    /* the pnotextra below accounts for not having another self-transition in intergene */
    prob += ptrans(INTERGENE,INTERGENE) + pnotextra(INTERGENE);

    tempProb = prob + alpha[INTERGENE][j-1];
    myadd(tempProb,alpha[INTERGENE][t]);

    prob += delta[INTERGENE][j-1];
    if(prob > deltaMaxB) {
      deltaMaxB = prob;
      (backtrack[INTERGENE][t]).state=INTERGENE;
      (backtrack[INTERGENE][t]).loc=j-1;
      (backtrack[INTERGENE][t]).type=SNGL_B;
    }
  }

  /* Sum over all possible backward initial exons starting at t-1 */
  minLen = START_SIGNAL_LEN + don_upstr + don_dnstr + 1;
  for(i=0; i<nBdon; i++) {
    j = bdon[i];
    seqLen = t-j;

    /* skip if the exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT) ) continue;

    /* stop if the exon would be too short */
    if( seqLen < minLen) break;

    exonLen = seqLen-don_dnstr;
    prevExtraBases = 0;
    thisExtraBases = exonLen % PHASES;
    leftState = I0B + thisExtraBases;

    prob = getBakExonProb(j,t-1,prevExtraBases,TYPE_INIT,seq,option);
    prob += ptrans(leftState,INTERGENE) + pnotextra(leftState);

    tempProb = prob + alpha[leftState][j-1];
    myadd(tempProb,alpha[INTERGENE][t]);

    prob += delta[leftState][j-1];
    if(prob > deltaMaxB) {
      deltaMaxB = prob;
      backtrack[INTERGENE][t].state=leftState;
      backtrack[INTERGENE][t].loc=j-1;
      backtrack[INTERGENE][t].type=INIT_B;
    }
  }

  return(deltaMaxB);
}


/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible initial and internal exons in forward direction.
*/
double donorforward(int t, int state, double deltaMaxB, Option *option) {
  int i,j,seqLen,exonLen,minLen;
  int prevExtraBases,lastState;
  double prob;
  double tempProb;

  if( (t != fdon[curFdon]) || (curFdon >= nFdon) )
    return(deltaMaxB);
  if(state == I2F)
    curFdon++;
  /* t points to the first base _after_ the end of the donor signal */

  /* Sum over all possible initial exons */
  minLen = don_upstr+don_dnstr+START_SIGNAL_LEN+1;
  for(i=0;i<nFstart;i++) {
    j = fstart[i];
    seqLen = t-j;

    /* skip if the initial exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT) ) continue;

    /* stop if the initial exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen - don_dnstr; 

    /* skip if this exon not the right length */
    if( (exonLen % PHASES) != (state - I0F) ) continue;

    prevExtraBases = 0;
    prob = getFwdExonProb(j,t-1,prevExtraBases,TYPE_INIT,seq,option);
    if(prob > -FLT_MAX) {
      prob += ptrans(INTERGENE,state) + pnotextra(INTERGENE);

      tempProb = prob + alpha[INTERGENE][j-1];
      myadd(tempProb,alpha[state][t]);

      prob += delta[INTERGENE][j-1];
      if(prob > deltaMaxB) {
        deltaMaxB = prob;
        backtrack[state][t].state=INTERGENE;
        backtrack[state][t].loc=j-1;
        backtrack[state][t].type=INIT_F;
      }
    }
  }

  /* Sum over all possible internal exons */
  minLen = acc_upstr+acc_dnstr+don_upstr+don_dnstr+1;
  for(i=0; i<nFacc; i++) {
    j = facc[i];
    seqLen = t-j;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the internal exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    prevExtraBases = ( (state-I0F) + PHASES - (exonLen % PHASES)) % PHASES;
    lastState = I0F+prevExtraBases;

    prob = getFwdExonProb(j,t-1,prevExtraBases,TYPE_INTL,seq,option);
    if(prob > -FLT_MAX) {
      prob += ptrans(lastState,state) + pnotextra(lastState);

      tempProb = prob + alpha[lastState][j-1];
      myadd(tempProb,alpha[state][t]);

      prob += delta[lastState][j-1];
      if(prob > deltaMaxB) {
        deltaMaxB = prob;
        backtrack[state][t].state=lastState;
        backtrack[state][t].loc=j-1;
        backtrack[state][t].type=INTL_F;
      }
    }
  }

  return (deltaMaxB);
}


/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible internal and terminal exons in the backward direction.
*/
double acceptorbackward(int t, int state, double deltaMaxB, Option *option) {
  int i,j,seqLen,exonLen,minLen;
  int prevExtraBases,thisExtraBases,leftState;
  double prob;
  double tempProb;

  if( (t != bacc[curBacc]) || (curBacc >= nBacc) )
    return(deltaMaxB);
  if(state == I2B)
    curBacc++;
  /* t points to the first base to the left of the acceptor signal */

  /* Sum over all possible backward terminal exons whose right coordinate is t */
  minLen = STOP_SIGNAL_LEN + acc_upstr + acc_dnstr + 1;
  for(i=0; i<nBstop; i++) {
    j = bstop[i];
    seqLen = t-j;

    /* skip if the initial exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM) ) continue;

    /* stop if the initial exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen - acc_upstr; 

    /* skip if this exon not the right length */
    prevExtraBases = (state-I0B);
    if(  ((exonLen+prevExtraBases) % PHASES)  !=  0  ) continue;

    prob = getBakExonProb(j,t-1,prevExtraBases,TYPE_TERM,seq,option);
    prob += ptrans(INTERGENE,state) + pnotextra(INTERGENE);

    tempProb = prob + alpha[INTERGENE][j-1];
    myadd(tempProb,alpha[state][t]);

    prob += delta[INTERGENE][j-1];
    if(prob > deltaMaxB) {
      deltaMaxB = prob;
      backtrack[state][t].state=INTERGENE;
      backtrack[state][t].loc=j-1;
      backtrack[state][t].type=TERM_B;
    }
  }

  /* Sum over all possible backward internal exons */
  minLen = acc_upstr+acc_dnstr+don_upstr+don_dnstr+1;
  for(i=0; i<nBdon; i++) {
    j = bdon[i];
    seqLen = t-j;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the internal exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    prevExtraBases = (state-I0B);
    thisExtraBases = (exonLen + prevExtraBases) % PHASES;
    leftState = I0B + thisExtraBases;

    prob = getBakExonProb(j,t-1,prevExtraBases,TYPE_INTL,seq,option);
    prob += ptrans(leftState,state) + pnotextra(leftState);

    tempProb = prob + alpha[leftState][j-1];
    myadd(tempProb,alpha[state][t]);

    prob += delta[leftState][j-1];
    if(prob > deltaMaxB) {
      deltaMaxB = prob;
      backtrack[state][t].state=leftState;
      backtrack[state][t].loc=j-1;
      backtrack[state][t].type=INTL_B;
    }
  }

  return (deltaMaxB);
}


/*
** Required for calculation of backward variables.
** Sums over possible initial and single exons on the 
** forward strand, starting in position t.
*/
void startforward(int t, Option *option) {
  int seqLen,exonLen,minLen;
  int rightState,prevExtraBases;
  int i,j;
  double prob;

  if( (t != fstart[curFstart]) || (curFstart < 0) )
    return;
  else
    curFstart--;

  /* sum over all possible single exons starting at t */
  minLen = START_SIGNAL_LEN+STOP_SIGNAL_LEN+1;
  for(i=nFstop-1; i>=0; i--) {
    j = fstop[i];   /* j points to the first base _after_ the stop codon */
    exonLen = j-t;

    /* skip if single exon is too long */
    if( exonLen > maxExonLen(TYPE_SNGL) ) continue;

    /* stop if the single exon is too short (all the rest will be too) */
    if( exonLen < minLen ) break;

    /* skip if length not a multiple of 3 */
    if((exonLen % PHASES) != 0) continue;

    prevExtraBases = 0;
    prob  = getFwdExonProb(t,j-1,prevExtraBases,TYPE_SNGL,seq,option);
    if(prob > -FLT_MAX) {
      prob += pnotextra(INTERGENE) + ptrans(INTERGENE,INTERGENE);
      prob += beta[INTERGENE][j];
      if(j < slen)
        prob += bIntronIntergene(j,INTERGENE,slen);

      myadd(prob,beta[INTERGENE][t-1]);
    }
  }

  /* sum over all possible initial exons */
  minLen = START_SIGNAL_LEN + don_upstr + don_dnstr + 1;
  for(i=nFdon-1; i>=0; i--) {
    j = fdon[i];   /* j points to first base _after_ end of donor */
    seqLen = j-t;

    /* skip if the initial exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT) ) continue;

    /* stop if the initial exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-don_dnstr;

    prevExtraBases = 0;
    rightState = I0F + (exonLen % PHASES);
    prob  = getFwdExonProb(t,j-1,prevExtraBases,TYPE_INIT,seq,option);
    if(prob > -FLT_MAX) {
      prob += pnotextra(INTERGENE) + ptrans(INTERGENE,rightState);
      prob += beta[rightState][j];
      if(j < slen)
        prob += bIntronIntergene(j,rightState,slen);
  
      myadd(prob,beta[INTERGENE][t-1]);
    }
  }

  return;
}


/*
** Required for calculation of backward variables.
** Sums over possible terminal and single exons on the 
** backward strand, ending in position t.
*/
void endbackward(int t, Option *option) {
  int seqLen,exonLen,minLen;
  int rightState,prevExtraBases;
  int i,j;
  double prob;

  if( (t != bstop[curBstop]) || (curBstop < 0) )
    return;
  else
    curBstop--;

  /* sum over all possible single exons starting at t */
  minLen = START_SIGNAL_LEN+STOP_SIGNAL_LEN+1;
  for(i=nBstart-1; i>=0; i--) {
    j = bstart[i];   /* j points to the first base after the start codon */
    exonLen = j-t;

    /* skip if single exon is too long */
    if( exonLen > maxExonLen(TYPE_SNGL) ) continue;

    /* stop if the single exon is too short (all the rest will be too) */
    if( exonLen < minLen ) break;

    /* skip if length not a multiple of 3 */
    if((exonLen % PHASES) != 0) continue;

    prevExtraBases = 0;
    prob  = getBakExonProb(t,j-1,prevExtraBases,TYPE_SNGL,seq,option);
    prob += pnotextra(INTERGENE) + ptrans(INTERGENE,INTERGENE);
    prob += beta[INTERGENE][j];
    if(j < slen)
      prob += bIntronIntergene(j,INTERGENE,slen);

    myadd(prob,beta[INTERGENE][t-1]);
  }

  /* sum over all possible terminal exons */
  minLen = STOP_SIGNAL_LEN + acc_upstr + acc_dnstr + 1;
  for(i=nBacc-1; i>=0; i--) {
    j = bacc[i];   /* j points to first base after end of acceptor */
    seqLen = j-t;

    /* skip if the exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM) ) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr;

    prevExtraBases = (PHASES - (exonLen % PHASES)) % PHASES;
    rightState = I0B + prevExtraBases;

    prob  = getBakExonProb(t,j-1,prevExtraBases,TYPE_TERM,seq,option);
    prob += pnotextra(INTERGENE) + ptrans(INTERGENE,rightState);
    prob += beta[rightState][j];
    if(j < slen)
      prob += bIntronIntergene(j,rightState,slen);

    myadd(prob,beta[INTERGENE][t-1]);
  }

  return;
}


/*
** Required for calculation of backward variables.
** Sums over possible internal and terminal exons in the forward 
** direction, starting at t.
*/
void acceptorforward(int t, int state, Option *option) {
  int exonLen,seqLen,minLen;
  int rightState,prevExtraBases;
  int i,j;
  double prob;

  if( (t != facc[curFacc]) || (curFacc < 0) )
    return;
  else if (state == I2F)
    curFacc--;

  prevExtraBases = (state-I0F);  /* the phase in which we'll begin */

  /* sum over all possible internal exons */
  minLen = don_upstr+don_dnstr+acc_upstr+acc_dnstr+1;
  for(i=nFdon-1; i>=0; i--) {
    j = fdon[i];   /* j points to the first base _after_ the stop codon */
    seqLen = j-t;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    rightState = I0F + ((prevExtraBases + exonLen) % PHASES);

    prob = getFwdExonProb(t,j-1,prevExtraBases,TYPE_INTL,seq,option);
    if(prob > -FLT_MAX) {
      prob += pnotextra(state) + ptrans(state,rightState);
      prob += beta[rightState][j];
      if(j < slen)
        prob += bIntronIntergene(j,rightState,slen);
  
      myadd(prob,beta[state][t-1]);
    }
  }

  /* sum over all possible terminal exons */
  minLen = acc_upstr+acc_dnstr+STOP_SIGNAL_LEN+1;
  for(i=nFstop-1; i>=0; i--) {
    j = fstop[i];   /* j points to the first base _after_ the stop codon */
    seqLen = j-t;

    /* skip if terminal exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM)) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr;
    if( 0 != ((prevExtraBases+exonLen) % PHASES)) continue;

    prob  = getFwdExonProb(t,j-1,prevExtraBases,TYPE_TERM,seq,option);
    if(prob > -FLT_MAX) {
      prob += pnotextra(state) + ptrans(state,INTERGENE);
      prob += beta[INTERGENE][j];
      if(j < slen)
        prob += bIntronIntergene(j,INTERGENE,slen);
      myadd(prob,beta[state][t-1]);
    }
  }

  return;
}


/*
** Required for calculation of backward variables.
** Sums over possible internal and initial exons in the backward
** direction, starting at t.
*/
void donorbackward(int t, int state, Option *option) {
  int exonLen,seqLen,minLen;
  int rightState,thisExtraBases,prevExtraBases;
  int i,j;
  double prob;

  if( (t != bdon[curBdon]) || (curBdon < 0) )
    return;
  else if (state == I2B)
    curBdon--;

  thisExtraBases = state-I0B;

  /* sum over all possible internal exons */
  minLen = don_upstr+don_dnstr+acc_upstr+acc_dnstr+1;
  for(i=nBacc-1; i>=0; i--) {
    j = bacc[i];   /* j points to the first base after the backward acceptor */
    seqLen = j-t;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    prevExtraBases = (thisExtraBases - (exonLen % PHASES) + PHASES) % PHASES;
    rightState = I0B + prevExtraBases;

    prob = getBakExonProb(t,j-1,prevExtraBases,TYPE_INTL,seq,option);
    prob += pnotextra(state) + ptrans(state,rightState);
    prob += beta[rightState][j];
    if(j < slen)
      prob += bIntronIntergene(j,rightState,slen);

    myadd(prob,beta[state][t-1]);
  }

  /* sum over all possible initial exons */
  minLen = don_dnstr+don_upstr+START_SIGNAL_LEN+1;
  for(i=nBstart-1; i>=0; i--) {
    j = bstart[i];   /* j points to the first base afte_ the start codon */
    seqLen = j-t;

    /* skip if exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT)) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-don_dnstr;
    prevExtraBases = 0;
    if( thisExtraBases != (exonLen % PHASES) ) continue;


    prob  = getBakExonProb(t,j-1,prevExtraBases,TYPE_INIT,seq,option);
    prob += pnotextra(state) + ptrans(state,INTERGENE);
    prob += beta[INTERGENE][j];
    if(j < slen)
      prob += bIntronIntergene(j,INTERGENE,slen);
    myadd(prob,beta[state][t-1]);
  }

  return;
}


/* 
** Backtracking part of the viterbi algorithm.  The backtrack will start 
** from state startstate at position tStart.  If startpos is <= 0 then
** it will start from the last base.
*/
void doBackTrack(int startstate, int tStart, Gene *gene, int *nGenes, Option *option) {
  short type;
  int t;
  int state;
  int oldstate;
  int tOld;

  /* start from last state if tStart <= 0 */
  if(tStart <= 0)
    tStart = slen;

  for(state=startstate,t=tStart,tOld=tStart; ; ) {
    /*
    ** check if we've reached the start of the sequence yet, and if
    ** so then break out of the for loop.
    */
    if(t == -1) {
      if(tOld != -1) {
        fprintf(stdout, "%d\n",t+2);
      }
      break;
    }

    /* skip over intron, ie skip states of duration 1 base */
    oldstate = state;
    tOld     = t;
    state = backtrack[oldstate][tOld].state;
    t     = backtrack[oldstate][tOld].loc;
    type  = backtrack[oldstate][tOld].type;
    while(tOld == t+1) {
      oldstate = state;
      tOld     = t;
      state = backtrack[oldstate][tOld].state;
      t     = backtrack[oldstate][tOld].loc;
      type  = backtrack[oldstate][tOld].type;
    }
    if(tOld != -1) {
      /* haven't reached the start of the sequence yet */
      if((*nGenes) < MAX_GENES) {
        gene = storeGene(gene,oldstate,tOld,nGenes,option);
      } else {
        fprintf(stderr,"Warning: ran out of space for storing genes\n");
        fprintf(stderr,"  Have space for %d in total\n",MAX_GENES);
        fprintf(stderr,"  (Maybe recompile with higher MAX_GENES value)\n");
        return;
      }
    }
  }
  /* must increment the gene count if we finished with part of a gene */
  if( (*gene).nExons != 0)
    (*nGenes)++;

  /* printf("about to exit doBacktrack\n"); */
  return;
}


/* 
** print out the details of the current exon
*/
void printBestExon(int i,int t) {
  int start,stop,prev,cur,type,phase,frame;
  char name[30];

  /* compute everything zero-based indices first */
  start = backtrack[i][t].loc + 1;
  stop  = t-1;
  cur   = i;
  prev  = backtrack[i][t].state;
  type  = backtrack[i][t].type;

  if(type == NADA)
    return;

  if(type == SNGL_F) {
    strcpy(name,"Single   ");
    phase = 0;
  } else if(type == INIT_F) { 
    strcpy(name,"Initial  ");
    phase = 0;
    stop -= don_dnstr;
  } else if (type == INTL_F) {
    strcpy(name,"Internal ");
    phase = 0;
    stop -= don_dnstr;
    start += acc_upstr;
  } else {
    strcpy(name,"Terminal ");
    phase = 0;
    start += acc_upstr;
  }
  frame = ((start + phase) % PHASES);

  printf("%s %6d -> %6d,  %4d bp   ph %d  fr %d  (st %d to %d)\n",
    name,start+1,stop+1,stop-start+1,phase,frame,prev,cur);
}


/*
** Records the exon in a Gene structure.  geneCount will be incremented by one
** and gene will be set to point to the next element in the Gene array if we
** add in a single or an initial exon.
*/
Gene *storeGene(Gene *gene, int i, int t, int *geneCount, Option *option) {
  int exonType;              /* NADA,SNGL_F,INIT_F,INTL_F,TERM_F,SNGL_B,INIT_B,INTL_B,TERM_B */
  int exonType2;             /* TYPE_SNGL,TYPE_INIT,TYPE_INTL,TYPE_TERM */ 
  int leftState,rightState;  /* I-states to the left and right of the exon */
  int left,right;            /* exon coordinates (including splice sites) */
  int exonLeft,exonRight;    /* exon coordinates (exon only, no splice sites) */
  double thisExonProb;       /* exon probability */
  int peb;                   /* prevExtraBases, # bases previously generated for first codon */
  int dir;                   /* exon direction: FORWARD or BACKWARD */
  int phase,frame;           /* phase and frame of index (see below for definition) */
                             /* exon to be skipped before reaching first base of next codon */
  int thisIndex;             /* just an index */
  double part1,part2,part3,part4,allparts;

  /* exon details */
  leftState  = backtrack[i][t].state;
  rightState = i;
  left       = backtrack[i][t].loc + 1;
  right      = t-1;
  exonType   = backtrack[i][t].type;


  if(exonType == SNGL_F) {
    exonLeft  = left;
    exonRight = right;
    dir       = FORWARD;
    exonType2 = TYPE_SNGL;
    peb       = 0;
  } else if (exonType == INIT_F) {
    exonLeft  = left;
    exonRight = right - don_dnstr;
    dir       = FORWARD;
    exonType2 = TYPE_INIT;
    peb       = 0;
  } else if (exonType == INTL_F) {
    exonLeft  = left + acc_upstr;
    exonRight = right - don_dnstr;
    dir       = FORWARD;
    exonType2 = TYPE_INTL;
    peb       = leftState-I0F;
  } else if (exonType == TERM_F) {
    exonLeft  = left + acc_upstr;
    exonRight = right;
    dir       = FORWARD;
    exonType2 = TYPE_TERM;
    peb       = leftState-I0F;
  } else if (exonType == SNGL_B) {
    exonLeft  = left;
    exonRight = right;
    dir       = BACKWARD;
    exonType2 = TYPE_SNGL;
    peb       = 0;
  } else if (exonType == INIT_B) {
    exonLeft  = left + don_dnstr;
    exonRight = right;
    dir       = BACKWARD;
    exonType2 = TYPE_INIT;
    peb       = 0;
  } else if (exonType == INTL_B) {
    exonLeft  = left + don_dnstr;
    exonRight = right - acc_upstr;
    dir       = BACKWARD;
    exonType2 = TYPE_INTL;
    peb       = rightState-I0B;
  } else if (exonType == TERM_B) {
    exonLeft  = left;
    exonRight = right - acc_upstr;
    dir       = BACKWARD;
    exonType2 = TYPE_TERM;
    peb       = rightState-I0B;
  } else {
    fprintf(stderr,"storeGene: bad exon type - shouldn't have happened!\n");
    exit(1);
  }

  /* get the probability this is a real exon */
  part1 = alpha[leftState][left-1];
  part2 = pnotextra(leftState) + ptrans(leftState,rightState);
  if(dir == FORWARD)
    part3 = getFwdExonProb(left,right,peb,exonType2,seq,option);
  else
    part3 = getBakExonProb(left,right,peb,exonType2,seq,option);
  part4 = beta[rightState][right];
  allparts  = part1+part2+part3+part4;
  thisExonProb  = exp(allparts-likelihood);

  if( option->verbose && ((thisExonProb < 0) || (thisExonProb > 1)) ) {
    fprintf(stderr,"Warning: computed an invalid exon probability!\n");
    fprintf(stderr,"  Exon [%d,%d], type %d\n",exonLeft,exonRight,exonType);
    fprintf(stderr,"  leftState  =  %d\n",leftState);
    fprintf(stderr,"  rightState =  %d\n",rightState);
    fprintf(stderr,"  prob     = %f\n",thisExonProb);
    fprintf(stderr,"  part1    = %f\n",part1);
    fprintf(stderr,"  part2    = %f\n",part2);
    fprintf(stderr,"  part3    = %f\n",part3);
    fprintf(stderr,"  part4    = %f\n",part4);
    fprintf(stderr,"  allparts = %f\n",allparts);
    fprintf(stderr,"  likelihood = %f\n",likelihood);
    thisExonProb = (thisExonProb > 1) ? 1 : 0;
  }

  /* 
  ** Compute phase and frame.
  ** Phase=i if ith base (going 5' -> 3') is start of a codon.
  ** Frame=i if codon boundaries lie between coordinates k-1 and k 
  **   where (k mod 3) = i (k is a zero-based index)
  */
  phase = (PHASES - peb) % PHASES;
  frame = (dir==FORWARD) ? exonLeft + phase : exonRight - phase + 1;
  frame = (option->offset + frame) % PHASES;

  if((*gene).nExons >= MAX_EXONS) {
    fprintf(stderr,"Warning: storeGene: ran out of space for exon storage\n");
    fprintf(stderr,"  Can store at most %d\n",MAX_EXONS);
    fprintf(stderr,"  (Maybe recompile with higher value of MAX_EXONS)\n");
    return gene;
  } else {
    thisIndex = (*gene).nExons;
    (*gene).start[thisIndex] = option->offset + exonLeft;
    (*gene).stop[thisIndex]  = option->offset + exonRight;
    (*gene).frame[thisIndex] = frame;
    (*gene).phase[thisIndex] = phase;
    (*gene).type[thisIndex]  = exonType;
    (*gene).prob[thisIndex]  = thisExonProb;
    (*gene).nExons++;
  }

  if( (exonType==INIT_F) || (exonType==SNGL_F) || (exonType==TERM_B) || (exonType==SNGL_B) ) {
    (*geneCount)++;
    return gene+1;
  } else {
    return gene;
  }
}

/* 
** returns
** 0 if c is a or A
** 1 if c is c or C
** 2 if c is g or G
** 3 if c is t or T
** 4 otherwise
*/
int base2num(int c) {

  switch(c) {
    case 'a':
    case 'A':
      return 0;
    break;

    case 'c':
    case 'C':
      return 1;
    break;

    case 'g':
    case 'G':
      return 2;
    break;

    case 't':
    case 'T':
      return 3;
    break;

    default:
      return 4;
    break;
  }

}

char *getGenomicSeqPointer(void) {
  return(seq);
}


/* 
** Write out the marginal probabilities of each base being some kind of exon
** for the forward pass.
*/
void writeProbFile(int nBases, Option *opt) {
  FILE *outProbFile = NULL;
  int t,i;
  double numerator,prob,lastProb,diff;

  if((outProbFile = fopen(opt->outProbFileName,"w")) == NULL) {
    fprintf(stderr,"writeProbFile: can't open %s for write\n",opt->outProbFileName);
    exit(1);
  }
  lastProb = -1;
  diff = pow(10.0, (double)(-(PROB_SIG_FIG+1)));

  for(t=0; t < nBases; t++) {
    numerator = alpha[0][t] + beta[0][t];
    for(i=1; i < NUMSTATES; i++)
      myadd(alpha[i][t] + beta[i][t], numerator);
    prob = round(1-exp(numerator-likelihood),PROB_SIG_FIG);

    /* this should be impossible, but just in case... */
    if(prob < 0)
      prob = 0;
    else if(prob > 1)
      prob = 1;

    if(opt->oldprb) {
      fprintf(outProbFile,"%g\n",prob);
    } else {
      if( (fabs(prob-lastProb) > diff) || (t == (nBases-1)) ) {
        fprintf(outProbFile,"%d %g\n",t+1,prob);
        lastProb = prob;
      }
    }
  }

  if(opt->verbose) printf("  Wrote exon probabilities to %s\n",opt->outProbFileName);

  return;
}


void testFunction(void) {
  int t,i,j;
  double sum,prob;
  FILE *testFile;
  FILE *alphaFile;
  FILE *betaFile;

  testFile = fopen("test.dat","w");
  if(testFile == NULL) {
    fprintf(stderr,"testFunction: can't open output file\n");
    exit(EXIT_FAILURE);
  }
  for(t=0; t<slen; t++) {
    /* sum = alpha[0][t] + beta[0][t] - bIntronIntergene(t,0,slen); */
    sum = alpha[0][t] + beta[0][t];
    for(i=1; i<NUMSTATES; i++) {
      /* myadd(alpha[i][t] + beta[i][t] - bIntronIntergene(t,i,slen),sum); */
      myadd(alpha[i][t] + beta[i][t],sum);
    }
    prob = exp(sum-likelihood);
    fprintf(testFile,"%f\t%f\t%f\n",sum,likelihood,prob);
  }
  fclose(testFile);
  printf("Wrote test.dat\n");

  alphaFile = fopen("alpha.dat","w");
  for(t=-1; t<=slen; t++) {
    for(i=0; i<NUMSTATES; i++)
      fprintf(alphaFile,"%2.10f\t",alpha[i][t]);
    fprintf(alphaFile,"\n");
  }
  fclose(alphaFile);
  printf("Wrote alpha.dat\n");

  betaFile = fopen("beta.dat","w");
  for(t=-1; t<=slen; t++) {
    for(i=0; i<NUMSTATES; i++)
      fprintf(betaFile,"%2.10f\t",beta[i][t]);
    fprintf(betaFile,"\n");
  }
  fclose(betaFile);
  printf("Wrote beta.dat\n");

  printf("\n");
  printf("Transitions...\n");
  for(i=0; i<NUMSTATES; i++) {
    for(j=0; j<NUMSTATES; j++)
      printf("%2.15f  ",exp(ptrans(i,j)));
    printf("\n");
  }

  printf("\n");
  for(i=0; i<NUMSTATES; i++)
    printf("pself[%d] = %2.15f\n",i,1-exp(pnotextra(i)));

  printf("\n");
  for(i=0,sum=0; i<NUMSTATES; i++) {
    sum += exp(pinit(i));
    printf("pinit[%d] = %2.15f   (%2.15f)\n",i,exp(pinit(i)),pinit(i));
  }
  printf("  sum(pinit(i)) = %f\n",sum);

  return;
}

/*
 * Simple function to round x to n significant figures.
 */
double round(double x, int n) {
  int xInt,rem;

  xInt = (int) (x * pow(10.0,(double)(n+1)));
  if(xInt > 0) {
    rem = xInt - 10 * (int) floor(xInt/10);
    if(rem > 4) ++xInt;
  } else {
    rem = 10 * (int) floor(xInt/10) - xInt;
    if(rem > 4) --xInt;
  }

  return(xInt/pow(10.0,(double)n));
}
