/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>

#define EXTERN extern
#include "full_common.h"
#undef EXTERN

#define EXTERN
#include "full_splice.h"
#undef EXTERN

#define MAX_WORD_SIZE 100
#define MAX_LINE_SIZE 1000
#define BEGINWORD     "BEGIN"
#define ENDWORD       "END"
#define SCOREWORD     "SCORES"

enum splice_models {SPLICE_MM, SPLICE_VLMC};

static double *fDonScore;       /* array to store forward donor scores */
static double *bDonScore;       /* array to store backward donor scores */
static int *fDonScoreIndex;  /* indices of forward donor scores in fDonScore */
static int *bDonScoreIndex;  /* indices of backward donor scores in bDonScore */
static int nFdon;            /* the number of forward donors currently scored */
static int nBdon;            /* the number of backward donors currently scored */
static int maxFdon;          /* the max possible number of forward donors to score */
static int maxBdon;          /* the max possible number of backward donors to score */

static double *fAccScore;       /* array to store forward acceptor scores */
static double *bAccScore;       /* array to store backward acceptor scores */
static int    *fAccScoreIndex;  /* indices of forward acceptor scores in fAccScore */
static int    *bAccScoreIndex;  /* indices of backward acceptor scores in fAccScore */
static int    nFacc;            /* the number of forward acceptors currently scored */
static int    nBacc;            /* the number of backward acceptors currently scored */
static int    maxFacc;          /* the max possible number of forward acceptors to score */
static int    maxBacc;          /* the max possible number of backward acceptors to score */

static int       acc_type;  /* the type of the acceptor model */
static Splice_mm acc_mm;    /* a markov model for acceptors */

static int       don_type;  /* the type of the donor model */
static Splice_mm don_mm;    /* a markov model for donors */


/*
** Read the file with the data for the acceptor splice site. 
** Input is just the pathname of the file to read.
** Note that this also sets the values of the global variables 
** acc_upstr and acc_dnstr
*/
void readAcceptorData(char *filename, Option *option) {
  acc_type = readSpliceData(filename,&acc_mm,&acc_upstr,&acc_dnstr,option);
}


/*
** Read the file with the data for the donor site.
** The input is the pathname of the file to read.
** Note that this also sets the values of the global variables 
** don_upstr and don_dnstr
*/
void readDonorData(char *filename, Option *option) {
  don_type = readSpliceData(filename,&don_mm,&don_upstr,&don_dnstr,option);
}


/*
** Reads in a splice site data file (donor or acceptor).  Currently the only format it 
** can handle is markov models, but more will be added.  The effect of this function
** is to store the model in the supplied Splice_mm object, and the type is returned.
*/
int readSpliceData(char *filename, Splice_mm *mm, int *up_str, int *dn_str, Option *option) {
  FILE *dataFile=NULL;
  char line[MAX_LINE_SIZE];
  char word[MAX_WORD_SIZE];
  int n,i,j,lineNo,type;
  int order,thisOrder,nContexts,thisPos,pos,siteLen;
  double num1,num2,num3,num4;

  if ((dataFile = fopen(filename,"r")) == NULL) {
    fprintf(stderr,"readSpliceData: can't open %s for read, exiting\n",filename);
    exit(1);
  }

  /* 
  ** Skip comments and blank lines before the starting signal
  ** and determine type.
  */
  for(lineNo=1; ;lineNo++) {
    if(NULL == fgets(line,MAX_LINE_SIZE,dataFile)) {
      fprintf(stderr,"readSpliceData: never found model type in %s, exiting\n",filename);
      exit(1);
    } else {
      n = sscanf(line,"%s",word);
      /* printf("I read %s\n",word); */
      if (n != 1) {
        /* printf("skipping blank line\n"); */
        continue;
      } else {
        if(word[0] == '#') {
          /* printf("skipping comment line\n"); */
          continue;
        } else if (strcmp(word,"MM")==0) {
          type = SPLICE_MM;
          if(option->verbose)  printf("readSpliceData: model type in %s is MM\n",filename);
          break;
        } else if (strcmp(word,"VLMC")==0) {
          acc_type = SPLICE_VLMC;
          if(option->verbose)  printf("readSpliceData: model in %s is VLMC\n",filename);
          break;
        } else {
          fprintf(stderr,"readSpliceData: bad format in line %d\n",lineNo);
          exit(EXIT_FAILURE);
        }
      }
    }
  }

  /* 
  ** read in the order of the model
  */
  for(lineNo++; ;lineNo++) {
    if(NULL == fgets(line,MAX_LINE_SIZE,dataFile)) {
      fprintf(stderr,"readSpliceData: never found model order in %s, exiting\n",filename);
      exit(EXIT_FAILURE);
    } else {
      n = sscanf(line,"%s %d",word,&order);
      if(n != 2) {
        fprintf(stderr,"readSpliceData: bad format in line %d of %s\n",lineNo,filename);
        exit(EXIT_FAILURE);
      } else {
        if(strcmp(word,"order")!=0) {
          fprintf(stderr,"readSpliceData: line %d of %s should begin with the word \"order\"\n",lineNo,filename);
          exit(EXIT_FAILURE);
        } else if (order < 0) {
          fprintf(stderr,"readSpliceData: negative order found in %s\n",filename);
          exit(EXIT_FAILURE);
        } else {
          break;
        }
      }
    }
  }
  if(option->verbose)  printf("readSpliceData: model order in %s is %d\n",filename,order);

  /* 
  ** read in the number of upstream bases
  */
  for(lineNo++; ;lineNo++) {
    if(NULL == fgets(line,MAX_LINE_SIZE,dataFile)) {
      fprintf(stderr,"readSpliceData: never found no of upstream bases in %s, exiting\n",filename);
      exit(EXIT_FAILURE);
    } else {
      n = sscanf(line,"%s %d",word,up_str);
      if(n != 2) {
        fprintf(stderr,"readSpliceData: bad format in line %d of %s\n",lineNo,filename);
        exit(EXIT_FAILURE);
      } else {
        if(strcmp(word,"upstream")!=0) {
          fprintf(stderr,"readSpliceData: line %d of %s should begin with the word \"upstream\"\n",lineNo,filename);
          exit(EXIT_FAILURE);
        } else if (*up_str < 0) {
          fprintf(stderr,"readSpliceData: negative upstream value found in %s\n",filename);
          exit(EXIT_FAILURE);
        } else {
          break;
        }
      }
    }
  }
  if(option->verbose)  printf("readSpliceData: %d upstream bases in %s\n",*up_str,filename);

  /* 
  ** read in the number of downstream bases
  */
  for(lineNo++; ;lineNo++) {
    if(NULL == fgets(line,MAX_LINE_SIZE,dataFile)) {
      fprintf(stderr,"readSpliceData: never found number of downstream bases in %s, exiting\n",filename);
      exit(EXIT_FAILURE);
    } else {
      n = sscanf(line,"%s %d",word,dn_str);
      if(n != 2) {
        fprintf(stderr,"readSpliceData: bad format in line %d of %s\n",lineNo,filename);
        exit(EXIT_FAILURE);
      } else {
        if(strcmp(word,"downstream")!=0) {
          fprintf(stderr,"readSpliceData: line %d of %s should begin with the word \"downstream\"\n",lineNo,filename);
          exit(EXIT_FAILURE);
        } else if (*dn_str < 0) {
          fprintf(stderr,"readSpliceData: negative value for dn_str in %s\n",filename);
          exit(EXIT_FAILURE);
        } else {
          break;
        }
      }
    }
  }
  if(option->verbose)  printf("readSpliceData: %d downstream bases in %s\n",*dn_str,filename);

  /*
  ** declare the memory for the model
  */
  siteLen = *up_str + *dn_str;
  switch(type) {
    case SPLICE_MM:
      mm->max_order = order;
      mm->nPos = siteLen;
      mm->pos = (MM *) malloc(siteLen * sizeof(MM));
      if(mm->pos == NULL) {
        fprintf(stderr,"readSpliceData: can't allocate mem for mm->pos\n");
        exit(EXIT_FAILURE);
      }
      for(i=0, thisOrder=0, nContexts=1; i < siteLen; i++) {
        mm->pos[i].order = thisOrder;
        mm->pos[i].nContexts = nContexts;
        mm->pos[i].context = (Probs *) malloc( nContexts * sizeof(Probs));
        if(mm->pos[i].context == NULL) {
          fprintf(stderr,"readSpliceData: can't allocate mem for mm->pos[%d].context\n",i);
          exit(EXIT_FAILURE);
        }
        if(thisOrder < order) {
          thisOrder++;
          nContexts *= N_BASES;
        }
      }
    break;

    case SPLICE_VLMC:
      fprintf(stderr,"readSpliceData: haven't written code for VLMC yet\n");
      exit(EXIT_FAILURE);
    break;
    
    default:
      fprintf(stderr,"readSpliceData: bad type (%d), exiting\n",type);
      exit(EXIT_FAILURE);
    break;
  }

  /*
  ** read in the data
  */
  lineNo++;
  for(i=0,thisPos=-*up_str,thisOrder=0,nContexts=1; i<siteLen ; i++,thisPos++) {
    for(j=0; j<nContexts; j++, lineNo++) {
      if(NULL == fgets(line,MAX_LINE_SIZE,dataFile)) {
        fprintf(stderr,"readSpliceData: (pos %d) bad format before line %d of %s\n",thisPos,lineNo,filename);
        exit(EXIT_FAILURE);
      }
      n = sscanf(line,"pos %d %lf %lf %lf %lf",&pos,&num1,&num2,&num3,&num4);

/* another problem - for cms I took an extra exp! */
/*
num1 = log(num1); num2 = log(num2); num3 = log(num3); num4 = log(num4);
*/

      if(n != 5) {
        fprintf(stderr,"readSpliceData: bad format in line %d of %s\n",lineNo,filename);
        exit(EXIT_FAILURE);
      } else {
        if(pos != thisPos) {
          fprintf(stderr,"readSpliceData: bad position in line %d of %s\n",lineNo,filename);
          fprintf(stderr,"  (pos is %d, should be %d)\n",pos,thisPos);
          exit(EXIT_FAILURE);
        } else if( (num1 < 0) || (num2 < 0) || (num3 < 0) || (num4 < 0) ) {
          fprintf(stderr,"readSpliceData: all entries should be nonnegative (line %d of %s)\n",lineNo,filename);
          exit(EXIT_FAILURE);
        } else {
          mm->pos[i].context[j].probs[0] = num1;
          mm->pos[i].context[j].probs[1] = num2;
          mm->pos[i].context[j].probs[2] = num3;
          mm->pos[i].context[j].probs[3] = num4;
          normalizeArray(mm->pos[i].context[j].probs,1,N_BASES);
        }
      }
    }
    if(thisOrder < order) {
      thisOrder++;
      nContexts *= N_BASES;
    }
    fgets(line,MAX_LINE_SIZE,dataFile);  /* skip a line */
  }
  fclose(dataFile);

  /* 
  ** print out the data just read in
  printf("\nSplice data from %s (after normalization)\n",filename);
  for(i=0,pos=-*up_str,nContexts=1,thisOrder=0; i<siteLen; i++,pos++) {
    printf("position %d:  order %d\n",pos,thisOrder);
    for(j=0; j<nContexts; j++) {
      for(k=0; k<N_BASES; k++) {
        if(mm->pos[i].context[j].probs[k] > -FLT_MAX)
          printf("  %3.2lf",mm->pos[i].context[j].probs[k]);
        else
          printf(" -Inf ");
      }
      printf("\n");
    }
    if(thisOrder < order) {
      thisOrder++;
      nContexts *= N_BASES;
    }
  }
  */

  if(option->verbose)  printf("\n");

  return(type);
}


/* 
** Normalizes array and takes logs.
** array is a dim1 x dim2 matrix of positive values, stored by row.
** Zero entries will end up with -FLT_MAX as their value.
*/
void normalizeArray(double *array, int dim1, int dim2) {
  int i,j;
  double sum;

  for(i=0; i<dim1; i++) {
    sum=0;

    for(j=0; j<dim2; j++) {
      sum += array[i*dim2+j];
/*       printf("%lf ",array[i*dim2+j]); */
    }
    /* printf(":  row[%d] sums to %lf\n",i,sum); */

    if(sum <= 0)  {
      fprintf(stderr,"normalizeArray: sum of row %d is not positive!\n",i);
      exit(1);
    }

    for(j=0; j<dim2; j++) {
      if(array[i*dim2+j] > 0) 
        array[i*dim2+j] = log(array[i*dim2+j]/sum);
      else
        array[i*dim2+j] = -FLT_MAX;
    }
  }
}

/* 
** Allocate space for the lookup table for scoring donors.  The lookup
** table consists of the following:
**
** double *fDonScore
** double *bDonScore
** int    *fDonScoreIndex
** int    *bDonScoreIndex
** int    nFdon;
** int    nBdon;
**
** For example, fDonScoreIndex[i] will have value "-1" if i is not the left end of a 
** donor site, otherwise it will give the index of the desired donor score in the 
** fDonScore array, ie the score is fDonScore[fDonScoreIndex[i]]
** 
*/
void donLookupInit(int nBases, int nFor, int nBak, Option *option) {
  int i;

  fDonScoreIndex = (int *) malloc(nBases * sizeof(int));
  if(fDonScoreIndex==NULL) {
    fprintf(stderr,"donLookupInit: can't allocate memory for fDonScoreIndex\n");
    exit(EXIT_FAILURE);
  }

  bDonScoreIndex = (int *) malloc(nBases * sizeof(int));
  if(bDonScoreIndex==NULL) {
    fprintf(stderr,"donLookupInit: can't allocate memory for bDonScoreIndex\n");
    exit(EXIT_FAILURE);
  }

  if(option->verbose) printf("Allocating memory for %d forward donors\n",nFor);
  fDonScore = (double *) malloc(nFor * sizeof(double));
  if(fDonScore==NULL) {
    fprintf(stderr,"donLookupInit: can't allocate memory for fDonScore\n");
    exit(EXIT_FAILURE);
  }

  if(option->verbose) printf("Allocating memory for %d backward donors\n",nBak);
  bDonScore = (double *) malloc(nBak * sizeof(double));
  if(bDonScore==NULL) {
    fprintf(stderr,"donLookupInit: can't allocate memory for bDonScore\n");
    exit(EXIT_FAILURE);
  }

  for(i=0; i<nBases; i++) {
    fDonScoreIndex[i] = -1;
    bDonScoreIndex[i] = -1;
  }

  maxFdon = nFor;
  maxBdon = nBak;
  nFdon   = 0;
  nBdon   = 0;

  if(option->verbose) printf("\n");

  return;
}


/*
** Allocate space for the lookup table for scoring acceptors.  The lookup
** table consists of the following:
**
** double *fAccScore
** double *bAccScore
** int    *fAccScoreIndex
** int    *bAccScoreIndex
** int    nFacc;
** int    nBacc;
**
** For example, bAccScoreIndex[i] will have value "-1" if i is not the left end of a 
** backward acceptor site, otherwise it will give the index of the desired acceptor
** score in the bAccScore array, ie the score is bAccScore[bAccScoreIndex[i]]
** 
*/
void accLookupInit(int nBases, int nFor, int nBak, Option *option) {
  int i;

  fAccScoreIndex = (int *) malloc(nBases * sizeof(int));
  if(fAccScoreIndex==NULL) {
    fprintf(stderr,"accLookupInit: can't allocate memory for fAccScoreIndex\n");
    exit(EXIT_FAILURE);
  }

  bAccScoreIndex = (int *) malloc(nBases * sizeof(int));
  if(bAccScoreIndex==NULL) {
    fprintf(stderr,"accLookupInit: can't allocate memory for bAccScoreIndex\n");
    exit(EXIT_FAILURE);
  }

  if(option->verbose) printf("Allocating memory for %d forward acceptors\n",nFor);
  fAccScore = (double *) malloc(nFor * sizeof(double));
  if(fAccScore==NULL) {
    fprintf(stderr,"accLookupInit: can't allocate memory for fAccScore\n");
    exit(EXIT_FAILURE);
  }

  if(option->verbose) printf("Allocating memory for %d backward acceptors\n",nBak);
  bAccScore = (double *) malloc(nBak * sizeof(double));
  if(bAccScore==NULL) {
    fprintf(stderr,"accLookupInit: can't allocate memory for bAccScore\n");
    exit(EXIT_FAILURE);
  }

  for(i=0; i<nBases; i++) {
    fAccScoreIndex[i] = -1;
    bAccScoreIndex[i] = -1;
  }

  maxFacc = nFor;
  maxBacc = nBak;
  nFacc   = 0;
  nBacc   = 0;

  if(option->verbose) printf("\n");

  return;
}


/*
** Returns the score of the acceptor.  Doesn't recalculate it if it has
** already been stored in the lookup table.  The actual calculation of
** the score is done by calcAccScore.
*/
double scoreAcceptor(char *seq, int pos, int dir) {
  double score=0;

  if(dir == FORWARD) {
    if(fAccScoreIndex[pos] < 0) {
      if(nFacc >= maxFacc) {
        fprintf(stderr,"scoreAcceptor: no more space in lookup table for forward acceptors\n");
        fprintf(stderr,"scoreAcceptor: (max allowed is %d)\n",maxFacc);
        exit(EXIT_FAILURE);
      }
      score = calcAccScore(seq,dir);
      fAccScore[nFacc] = score;
      fAccScoreIndex[pos] = nFacc;
      nFacc++;
    } else {
      score = fAccScore[fAccScoreIndex[pos]];
    }
  } else if(dir == BACKWARD) {
    if(bAccScoreIndex[pos] < 0) {
      if(nBacc >= maxBacc) {
        fprintf(stderr,"scoreAcceptor: no more space in lookup table for backward acceptors\n");
        fprintf(stderr,"scoreAcceptor: (max allowed is %d)\n",maxBacc);
        exit(EXIT_FAILURE);
      }
      score = calcAccScore(seq,dir);
      bAccScore[nBacc] = score;
      bAccScoreIndex[pos] = nBacc;
      nBacc++;
    } else {
      score = bAccScore[bAccScoreIndex[pos]];
    }
  } else {
    fprintf(stderr,"scoreAcceptor: bad direction passed in\n");
    exit(EXIT_FAILURE);
  }

  return(score);
}


/*
** Returns the score of the donor.  Doesn't recalculate it if it has
** already been stored in the lookup table.  The actual calculation of
** the score is done by calcDonScore.
*/
double scoreDonor(char *seq, int pos, int dir) {
  double score=0;

  if(dir == FORWARD) {
    if(fDonScoreIndex[pos] < 0) {
      if(nFdon >= maxFdon) {
        fprintf(stderr,"scoreDonor: no more space in lookup table for forward donors\n");
        fprintf(stderr,"scoreDonor: (max allowed is %d)\n",maxFdon);
        exit(EXIT_FAILURE);
      }
      score = calcDonScore(seq,dir);
      fDonScore[nFdon] = score;
      fDonScoreIndex[pos] = nFdon;
      nFdon++;
    } else {
      score = fDonScore[fDonScoreIndex[pos]];
    }
  } else if(dir == BACKWARD) {
    if(bDonScoreIndex[pos] < 0) {
      if(nBdon >= maxBdon) {
        fprintf(stderr,"scoreDonor: no more space in lookup table for backward donors\n");
        fprintf(stderr,"scoreDonor: (max allowed is %d)\n",maxBdon);
        exit(EXIT_FAILURE);
      }
      score = calcDonScore(seq,dir);
      bDonScore[nBdon] = score;
      bDonScoreIndex[pos] = nBdon;
      nBdon++;
    } else {
      score = bDonScore[bDonScoreIndex[pos]];
    }
  } else {
    fprintf(stderr,"scoreDonor: bad direction passed in\n");
    exit(EXIT_FAILURE);
  }

  return(score);
}


/*
** Calculates the score of the acceptor.
*/
double calcAccScore(char *seq, int dir) {
  double score=0;

  switch(acc_type) {
    case SPLICE_MM:
      /* score = score_mm(acc_mm,seq,2,acc_upstr-2,dir); */
      score = score_mm(acc_mm,seq,0,0,dir);
    break;

    case SPLICE_VLMC:
      fprintf(stderr,"calcAccScore: can't handle VLMC yet\n");
      exit(EXIT_FAILURE);
    break;

    default:
      fprintf(stderr,"calcAccScore: bad acc_type (%d)\n",acc_type);
      exit(EXIT_FAILURE);
    break;
  }

  return(score);
}


/*
** Calculates the score of the donor.
*/
double calcDonScore(char *seq, int dir) {
  double score=0;

  switch(don_type) {
    case SPLICE_MM:
      /* score = score_mm(don_mm,seq,2,don_upstr,dir); */
      score = score_mm(don_mm,seq,0,0,dir);
    break;

    case SPLICE_VLMC:
      fprintf(stderr,"calcDonScore: can't handle VLMC yet\n");
      exit(EXIT_FAILURE);
    break;

    default:
      fprintf(stderr,"calcDonScore: bad don_type (%d)\n",don_type);
      exit(EXIT_FAILURE);
    break;
  }

  return(score);
}

/*
** Calculates a markov model score, given a specific model and sequence to score.
** nSkip is the number of bases not to be scored (ie assigned probability 1), which
** will usually be 2 - the minimal GT or AG consensus.  skipOffset is the index of
** the first base to skip.  Set nSkip to zero to not do this at all.
*/
double score_mm(Splice_mm mm, char *seq, int nSkip, int skipOffset, int dir) {
  double score=0;
  int i,j,thisContext,thisOrder;
  char *thisSeq;
  static char *revSeq;
  static int revSeqLen=0;

  if(dir == FORWARD) {
    /* point thisSeq at the original sequence */
    thisSeq = seq;
  } else if (dir == BACKWARD) {
    /* allocate memory for reverse complement, if necessary */
    if(mm.nPos > revSeqLen) {
      if(revSeq != NULL)
        free(revSeq);
      revSeq = (char *) malloc(mm.nPos * sizeof(char));
      if(revSeq==NULL) {
        fprintf(stderr,"score_mm: can't allocate memory for revSeq\n");
        exit(EXIT_FAILURE);
      }
      revSeqLen = mm.nPos;
    }
    /* reverse complement the sequence */
    for(i=0, j=mm.nPos-1; i < mm.nPos; i++, j--)
      revSeq[i] = (3 - seq[j]);
    /* point thisSeq at the reverse complement */
    thisSeq = revSeq;
  } else {
    fprintf(stderr,"score_mm: bad direction specified\n");
    exit(EXIT_FAILURE);
  }

  /* score the sequence */
  for(i=0,thisContext=0,thisOrder=0; i < mm.nPos; i++) {
    thisContext = thisContext % (mm.pos[i].nContexts);
    if((nSkip > 0) && (i == skipOffset)) {
      ++skipOffset;
      --nSkip;
    } else {
      score += mm.pos[i].context[thisContext].probs[(int) thisSeq[i]];
      /* printf("  Added %lf for seq[%d] = %d\n",mm.pos[i].context[thisContext].probs[thisSeq[i]],i,thisSeq[i]); */
    }
    thisContext *= N_BASES;
    thisContext += thisSeq[i];
  }

  /*
  ** some debugging code - to run checks on the splice site scores
  if(mm.nPos == acc_upstr+acc_dnstr)
    printf("acc ");
  else 
    printf("don ");
  printf("%lf ",score);
  for(i=0; i<mm.nPos; i++)
    printf("%d ",thisSeq[i]);
  printf("\n");
  */

  return(score);
}
