/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#ifndef HEADER_COMMON
#define HEADER_COMMON 1

/*
** Global functions
*/
void *myAlloc(size_t nBytes);
void error(char *fmt, ...);


/* splice site lengths */
EXTERN int acc_upstr;
EXTERN int acc_dnstr;
EXTERN int don_upstr;
EXTERN int don_dnstr;


#define MAXLEN 256
#define N_BASES 4     /* 4 characters in the alphabet: ACGT */
#define INTRORDER 2   /* the order of the MM used in intron/intergene */
#define INTRPRED 16   /* 4^INTRORDER, number of possible intron/intergene contexts */
#define EXONTYPES 4   /* 4 types: single, initial, internal, terminal */
#define NSTOPS 3      /* the number of possible stop codons (3: TAA, TAG, TGA) */
#define STOPLEN 3     /* length of stop signal (always 3 bases) */


#define getsymb(y,file)\
do{\
int c;\
c = fgetc(file);\
switch(c){\
case 'a': case 'A': case 'n': case 'N':\
y = 0;\
break;\
case 'c': case 'C':\
y = 1;\
break;\
case 'g': case 'G':\
y = 2;\
break;\
case 't': case 'T':\
y = 3;\
break;\
case EOF:\
y = EOF;\
break;\
default:\
y = 4;\
break;\
}\
}\
while(y > 3)

enum types {TYPE_SNGL,TYPE_INIT,TYPE_INTL,TYPE_TERM};


#define MAX_PATH_LEN     500           /* max # of characters in a file's path */
#define MAX_FILENAME_LEN 300           /* max # of characters in a filename (not including path) */
#define ENV_PAR_PATH     "PHATPARS"  /* environment var giving location of data directory */
#define FALSE 0
#define STORDER 3
#define STCUT 64
#define NUMPOSNS 1024*128   /* max number of candidate donors/acceptors/start codons/stop codons */
#define SPLORDER 2
#define SPCUT 16
#define EXTRA 5
#define START_SIGNAL_LEN 3  /* length of the start signal, ATG */
#define STOP_SIGNAL_LEN  3  /* length of the stop signal, TAA/TAG/TGA */
#define N_BASES     4       /* number of genetic bases = 4 (ACGT) */
#define PHASES 3            /* 3 possible phases in which an exon can be started (0/1/2) */

#define MAX(x,y) ( (x) > (y) ) ? (x) : (y)  /* macro for max */
#define MIN(x,y) ( (x) < (y) ) ? (x) : (y)  /* macro for min */

#define isfstart(x) (x == 14)   /* start codon on direct  strand */
#define isbstart(x) (x == 19)   /* start codon on reverse strand */
#define isfend(x) ((x == 48) || (x == 50) || (x == 56))  /* stop codon on direct  strand */
#define isbend(x) ((x == 28) || (x == 52) || (x == 60))  /* stop codon on reverse strand */

#define isfacc(y) (y == 2)   /* candidate acceptor on direct strand */
#define isfdon(y) (y == 11)  /* candidate donor on direct strand */
#define isbacc(y) (y == 7)   /* candidate acceptor on reverse strand */
#define isbdon(y) (y == 1)   /* candidate acceptor on reverse strand */

#define NUMSTATES 4  /* number of I-states in the HMM */

/*
** Default extensions for output files
*/
#define SPLICE_FILE_EXT  ".spl"
#define MRNA_FILE_EXT    ".rna"
#define PROT_FILE_EXT    ".pep"
#define GFF_FILE_EXT     ".gff"
#define FWDPROB_FILE_EXT ".prf"
#define BAKPROB_FILE_EXT ".prb"

/*
** Default names for parameter files
*/
#define ACCEPTOR_FILE   "acceptor.dat"
#define DONOR_FILE      "donor.dat"
#define EXONFREQ_FILE "exon.freq.dat"
#define EXON_LEN_SING_FILE "exon.len.sing.dat"
#define EXON_LEN_INIT_FILE "exon.len.init.dat"
#define EXON_LEN_INTL_FILE "exon.len.intl.dat"
#define EXON_LEN_TERM_FILE "exon.len.term.dat"
#define INTRONFREQ_FILE "intron.freq.dat"
#define INTRON_LEN_FILE "intron.len.dat"
#define INTERGENE_FREQ_FILE  "intergene.freq.dat"
#define INTERGENE_LEN_FILE  "intergene.len.dat"
#define TRANSITIONS_FILE "transitions.dat"
#define INITPROBS_FILE "initprobs.dat"
#define STOP_PROBS_FILE "stopprobs.dat"

/*
** Stuff for storing gene locations 
*/
#define MAX_EXONS 250
#define MAX_GENES 700

#define BACKTRACKHDR "First and last positions of non-coding regions are "\
"shown\nN intergenic\nz phase 0 forward intron\no phase 1 forward intron\n"\
"t phase 2 forward intron\nZ phase 0 backward intron\nO phase 1 backward "\
"intron\nT phase 2 backward intron\n\n"

#define BADEXINFILE "is not an exon intron statistics file."
#define BADGENFILE "is not an intergenic statistics file."
#define NOMEM "Insufficient memory: aborting\n"
#define LONGSEQ "DNA sequence is too long. Sorry\n"

#define myadd(x,y) \
do{\
	(y) += log(1 + exp((x)-(y)));\
}while(FALSE)


typedef struct {
  /* output options */
  int  noprob;     /* if true, skip writing probability file */
  int  norna;      /* if true, skip writing predicted rna file */
  int  nopep;      /* if true, skip writing predicted peptide file */
  int  noextras;   /* if true, will only write to screen and gff file */
  int  oldprb;     /* if true, use old (big!) format for posterior probabiliy file(s). */

  int  twoPasses;  /* predict on the reverse strand as well as the forward */
  int  flip;       /* reverse complement the sequence before processing */
  int  verbose;    /* run in verbose mode (quite a bit of output) */
  int  debug;      /* run in debug mode (lots of output) */
  int  geometric;  /* use geometric distributions for E states */
  int  offset;     /* to be added to coordinates when writing output */
  int  noext;      /* don't add extension to seq name in output files */
  char outSeqName[MAX_FILENAME_LEN];   /* name for sequence in .gff file */
  char baseFileName[MAX_FILENAME_LEN]; /* base name for output files */


  /* the input file */
  char inFile[MAX_PATH_LEN + MAX_FILENAME_LEN];

  /* output files */
  char outSpliceFileName[MAX_FILENAME_LEN];
  char outMrnaFileName[MAX_FILENAME_LEN];
  char outProtFileName[MAX_FILENAME_LEN];
  char outGffFileName[MAX_FILENAME_LEN];
  char outFwdProbFileName[MAX_FILENAME_LEN];
  char outBakProbFileName[MAX_FILENAME_LEN];

  /* parameter files */
  char acceptorFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char donorFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenSingFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenInitFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenIntlFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char exonLenTermFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intronFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intronLenFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intergeneFreqFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char intergeneLenFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char transitionsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char initProbsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
  char stopProbsFile[MAX_PATH_LEN + MAX_FILENAME_LEN];
} Option;

typedef struct {
  int nExons;
  int start[MAX_EXONS];
  int stop[MAX_EXONS];
  int frame[MAX_EXONS];
  int phase[MAX_EXONS];
  int type[MAX_EXONS];
  double prob[MAX_EXONS];
} Gene;

enum states     {INTERGENE,I0,I1,I2};
enum fileFormat {FASTA,GENBANK,UNKNOWN_FORMAT};
enum exonType   {NADA,ESING,EINIT,E0,E1,E2,TERM0,TERM1,TERM2};


#endif
