/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define EXTERN extern
#include "half_common.h"
#include "half_seq.h"
#include "half_splice.h"
#include "half_output.h"
#include "half_probs.h"
#undef EXTERN

Option processArgs(int argc, char *argv[]);

int main(int argc,char *argv[]) {
  Option option;
  int i,last,nBases;

  int  nGenesF,nGenesB;
  int  *geneCounter;
  Gene gene[MAX_GENES];
  Gene *genePointer;

  /* process command-line options */
  option = processArgs(argc,argv);

  /* read in all parameters (splice sites/exon/intron/intergene) */
  readStats(&option);

  /* set all genes to zero exons */
  for(i=0; i<MAX_GENES; i++)
    gene[i].nExons = 0;

  /*
  ** Read in the sequence and initialize the sequence-dependent
  ** data structures.
  */
  if (option.verbose)  printf("Reading in the sequence...\n\n");
  nBases = readSequence(&option);

  if(option.flip) reverseSequence(&option);

  genePointer = gene;
  geneCounter = &nGenesF;

  if (option.verbose) printf("Locating candidate exon boundaries...\n\n");
  locateBounds(&option);

  if (option.verbose) printf("Running Viterbi algroithm and calculating forward varaibles...\n\n");
  last = viterbi(&option);

  if (option.verbose) printf("Calculating backward variables...\n\n");
  backward(&option);

  if(option.verbose) printf("Backtracking from state %d...\n\n",last);
  doBackTrack(last,-1,genePointer,geneCounter,&option);

  if( (!option.noextras) && (!option.noprob) )  {
    if(option.verbose) printf("Writing forward exon probabilities...\n\n");
    writeFwdProbFile(nBases,&option);
  }

  if(option.twoPasses) {
    if (option.verbose)   printf("Backward Pass...\n\n");

    if (option.verbose)   printf("Reversing the sequence...\n\n");
    reverseSequence(&option);
    genePointer = gene+nGenesF;
    geneCounter = &nGenesB;

    if (option.verbose) printf("Locating candidate exon boundaries...\n\n");
    locateBounds(&option);
  
    if (option.verbose) printf("Running Viterbi algroithm and calculating forward varaibles...\n\n");
    last = viterbi(&option);
  
    if (option.verbose) printf("Calculating backward variables...\n\n");
    backward(&option);
  
    if(option.verbose) printf("Backtracking from state %d...\n\n",last);
    doBackTrack(last,-1,genePointer,geneCounter,&option);
  
    if( (!option.noextras) && (!option.noprob) )  {
      if(option.verbose) printf("Writing backward exon probabilities...\n\n");
      writeBakProbFile(nBases,&option);
    }
  
    reverseSequence(&option);
  }



  /* Write output files */
  writeScreenOutput(gene,nGenesF,nGenesB,nBases,&option);
  if(option.verbose) printf("\n");
  writeGffFile(gene,nGenesF,nGenesB,nBases,&option);


  /* write the rest of the output files if the user didn't suppress them */
  if(! option.noextras) {
    if(!option.norna)  writeMrnaFile(gene,nGenesF,nGenesB,nBases,&option);
    if(!option.nopep)  writeProtFile(gene,nGenesF,nGenesB,nBases,&option);
  }

  writeGffFile(gene,nGenesF,nGenesB,nBases,&option);
  writeMrnaFile(gene,nGenesF,nGenesB,nBases,&option);
  writeProtFile(gene,nGenesF,nGenesB,nBases,&option);

  return EXIT_SUCCESS;
}


Option processArgs(int argc, char *argv[]) {
  Option opt;
  char *envParPath,*inFileName,*lastDot;
  char parPath[MAX_PATH_LEN];
  char accFile[MAX_FILENAME_LEN];
  char donFile[MAX_FILENAME_LEN];
  char *thisarg;
  int len,i;
  int gotInFile=0;

  /* 
  ** Set default values for options
  */
  opt.twoPasses = 0;
  opt.flip      = 0;
  opt.noprob    = 0;
  opt.norna     = 0;
  opt.nopep     = 0;
  opt.noextras  = 0;
  opt.oldprb    = 0;
  opt.debug     = 0;
  opt.verbose   = 0;
  opt.geometric = 0;
  opt.offset    = 0;
  opt.noext     = 0;
  opt.baseFileName[0] = '\0';
  opt.outSeqName[0]   = '\0';
  parPath[0]          = '\0';
  accFile[0]          = '\0';
  donFile[0]          = '\0';

  for (i=1; i<argc; i++) {
    thisarg = argv[i];
    if(thisarg[0] == '-') {
      if(0==strcmp(thisarg+1,"r")) {
        opt.flip = 1;
      } else if(0==strcmp(thisarg+1,"2")) {
        opt.twoPasses = 1;
      } else if(0==strcmp(thisarg+1,"acc")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide acceptor data file with option -acc\n");
          exit(1);
        } else
          strcpy(accFile,argv[i]);
      } else if(0==strcmp(thisarg+1,"debug")) {
        opt.debug = 1;
      } else if(0==strcmp(thisarg+1,"don")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide donor data file with option -don\n");
          exit(1);
        } else
          strcpy(donFile,argv[i]);
      } else if(0==strcmp(thisarg+1,"p")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide a directory name with option -p\n");
          exit(1);
        } else
          strcpy(parPath,argv[i]);
      } else if(0==strcmp(thisarg+1,"v")) {
        opt.verbose = 1;
      } else if(0==strcmp(thisarg+1,"geo")) {
        opt.geometric = 1;
      } else if(0==strcmp(thisarg+1,"offset")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide an offset with option -offset\n");
          exit(1);
        } else
          opt.offset = atoi(argv[i]);
      } else if (0==strcmp(thisarg+1,"o")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide an output name with option -o\n");
          exit(1);
        } else
          strcpy(opt.baseFileName,argv[i]);
      } else if(0==strcmp(thisarg+1,"seq")) {
        i++;
        if(i >= argc) {
          fprintf(stderr,"Error: must provide a sequence name with option -seq\n");
          exit(1);
        } else
          strcpy(opt.outSeqName,argv[i]);
      } else if (0==strcmp(thisarg+1,"noext")) {
        opt.noext=1;
      } else if (0==strcmp(thisarg+1,"noprob")) {
        opt.noprob=1;
      } else if (0==strcmp(thisarg+1,"norna")) {
        opt.norna=1;
      } else if (0==strcmp(thisarg+1,"nopep")) {
        opt.nopep=1;
      } else if (0==strcmp(thisarg+1,"oldprb")) {
        opt.oldprb=1;
      } else if (0==strcmp(thisarg+1,"noextras")) {
        opt.noextras=1;
      } else {
        fprintf(stderr,"Error: unknown option %s\n",thisarg);
        exit(1);
      }
    } else {
      if(!gotInFile) {
        strcpy(opt.inFile,thisarg);
        gotInFile = 1;
      } else {
        fprintf(stderr,"Warning: ignoring command line option %s\n",thisarg);
      }
    }
  }

  /* make sure we got a file to process */
  if(!gotInFile) {
    fprintf(stderr,"Error: no sequence file was specified\n");
    exit(1);
  }

  /* look for envParPath in environment if parPath wasn't supplied at command line */
  if(parPath[0] == '\0') {
    envParPath = getenv(ENV_PAR_PATH);
    if(envParPath == NULL) {
      fprintf(stderr,"Error: can't locate environment variable %s\n",ENV_PAR_PATH);
      exit(1);
    }
    strcpy(parPath,envParPath);
  }

  /* die if pathname is too long */
  len = strlen(parPath);
  if(len >= MAX_PATH_LEN) {
    fprintf(stderr,"Error: parameter path name is too long (%d chars, can be at most %d)\n",
      len,MAX_PATH_LEN);
    exit(1);
  }

  /* add in a '/' if not already there */
  if(parPath[len-1] != '/')
    strcat(parPath,"/");

  /* set the acceptor file to the command-line specified file, or the default */
  strcpy(opt.acceptorFile,parPath);
  if(accFile[0] == '\0')
    strcat(opt.acceptorFile,ACCEPTOR_FILE);
  else
    strcat(opt.acceptorFile,accFile);

  /* set the acceptor file to the command-line specified file, or the default */
  strcpy(opt.donorFile,parPath);
  if(donFile[0] == '\0')
    strcat(opt.donorFile,DONOR_FILE);
  else
    strcat(opt.donorFile,donFile);

  strcpy(opt.transitionsFile,parPath);
  strcat(opt.transitionsFile,TRANSITIONS_FILE);

  strcpy(opt.exonFreqFile,parPath);
  strcat(opt.exonFreqFile,EXONFREQ_FILE);

  strcpy(opt.intronFreqFile,parPath);
  strcat(opt.intronFreqFile,INTRONFREQ_FILE);

  strcpy(opt.intergeneFreqFile,parPath);
  strcat(opt.intergeneFreqFile,INTERGENE_FREQ_FILE);

  strcpy(opt.intergeneLenFile,parPath);
  strcat(opt.intergeneLenFile,INTERGENE_LEN_FILE);

  strcpy(opt.intronLenFile,parPath);
  strcat(opt.intronLenFile,INTRON_LEN_FILE);

  strcpy(opt.exonLenSingFile,parPath);
  strcat(opt.exonLenSingFile,EXON_LEN_SING_FILE);

  strcpy(opt.exonLenInitFile,parPath);
  strcat(opt.exonLenInitFile,EXON_LEN_INIT_FILE);

  strcpy(opt.exonLenIntlFile,parPath);
  strcat(opt.exonLenIntlFile,EXON_LEN_INTL_FILE);

  strcpy(opt.exonLenTermFile,parPath);
  strcat(opt.exonLenTermFile,EXON_LEN_TERM_FILE);

  strcpy(opt.initProbsFile,parPath);
  strcat(opt.initProbsFile,INITPROBS_FILE);

  strcpy(opt.stopProbsFile,parPath);
  strcat(opt.stopProbsFile,STOP_PROBS_FILE);

  if(opt.geometric && opt.verbose)
    printf("Using geometric length distribution for exon lengths\n");


  /*
  ** Determine base file name, if it wasn't already set...
  */
  if(opt.baseFileName[0] == '\0') {
    /* set inFileName to the file name (as opposed to full path name) */
    if((inFileName = strrchr(opt.inFile,'/')) != NULL)
      inFileName++;
    else
      inFileName = opt.inFile;

    /* pull off everything up till the last '.' (if any) */
    if((lastDot = strrchr(inFileName,'.')) == NULL)
      strcpy(opt.baseFileName,inFileName);
    else
      strncpy(opt.baseFileName,inFileName,lastDot-inFileName);
    /* printf("input file = %s, base name = %s\n",inFileName,opt.baseFileName); */
  }

  /* set the gff seq name to base name, if it wasn't explicitly set */
  if(opt.outSeqName[0] == '\0') {
    /* point inFileName to the last / in baseFileName, if any */
    if((inFileName = strrchr(opt.baseFileName,'/')) != NULL)
      inFileName++;
    else
      inFileName = opt.baseFileName;
    strcpy(opt.outSeqName,inFileName);
  }

  strcpy(opt.outSpliceFileName,opt.baseFileName);
  strcpy(opt.outMrnaFileName,opt.baseFileName);
  strcpy(opt.outProtFileName,opt.baseFileName);
  strcpy(opt.outGffFileName,opt.baseFileName);
  strcpy(opt.outFwdProbFileName,opt.baseFileName);
  strcpy(opt.outBakProbFileName,opt.baseFileName);

  strcat(opt.outSpliceFileName,SPLICE_FILE_EXT);
  strcat(opt.outMrnaFileName,MRNA_FILE_EXT);
  strcat(opt.outProtFileName,PROT_FILE_EXT);
  strcat(opt.outGffFileName,GFF_FILE_EXT);
  strcat(opt.outFwdProbFileName,FWDPROB_FILE_EXT);
  strcat(opt.outBakProbFileName,BAKPROB_FILE_EXT);

  /*
  printf("splice file = %s\n",opt.outSpliceFileName);
  printf("mrna file   = %s\n",opt.outMrnaFileName);
  printf("prot file   = %s\n",opt.outProtFileName);
  printf("gff file    = %s\n",opt.outGffFileName);
  */

  
  return opt;
}
