/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>

#define EXTERN extern
#include "half_common.h"
#include "half_splice.h"
#undef EXTERN

#define EXTERN
#include "half_probs.h"
#undef EXTERN

static int numExons[EXONTYPES];
static double *exonLenDensity[EXONTYPES];

static double exonP[EXONTYPES];
static double exonQ[EXONTYPES];

static int    exonOrder;
static double ***logExonFreq;

static int intergenicFreq[INTRPRED][N_BASES];
static int intronFreq[INTRPRED][N_BASES+1];

static double transitions[NUMSTATES][NUMSTATES];

static double logInit[NUMSTATES];
static double logStop[NSTOPS];
static double logP[NUMSTATES];
static double logQ[NUMSTATES];
static double logNS[INTRPRED][N_BASES];
static double logFS[INTRPRED][N_BASES];

static double *exonProbLUT[PHASES];
static double *intronProbs;
static double *intergenicProbs;


/* functions internal to full_probs.c */
double ***declare3dArray(int dim1,int dim2,int dim3);
void free3dArray( double ***array, int dim1, int dim2);


/*
** Reads in all the model parameters.
*/
void readStats(Option *option) {

  readAcceptorData(option->acceptorFile,option);
  readDonorData(option->donorFile,option);

  readIntronLen(option->intronLenFile); 
  readIntronFreq(option->intronFreqFile);

  readExonLen(option);
  readExonFreq(option->exonFreqFile);

  readIntergeneLen(option->intergeneLenFile);
  readIntergeneFreq(option->intergeneFreqFile);

  readInitProbs(option->initProbsFile);
  readTransitions(option->transitionsFile, option);
  readStopProbs(option->stopProbsFile, option);

  if(option->geometric)
    geometricFix(option);

  return;
}

/*
** Reads intron frequencies.  Zero counts are replaced by 1.
*/
void readIntronFreq(char *inFile) {
  FILE   *fp;
  char   line[MAXLEN];
  int    i,j,sum;
  int    off1,off2,freq;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntronFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAXLEN,fp) == NULL) {
    fprintf(stderr,"readIntronFreq: can't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  /*
  ** We could pick up some info here, but for now we assume the format,
  ** and so the lins below have been commented out.
  ** 
  **if(2 != sscanf(line,"%s %d",str,&num)) {
  **  fprintf(stderr,"readIntronFreq: bad format in first line of %s\n",inFile);
  **  exit(EXIT_FAILURE);
  **}
  **if(num != INTRPRED) {
  **  fprintf(stderr,"readIntronFreq: in %s, should have %d entries\n",inFile,INTRPRED);
  **  exit(EXIT_FAILURE);
  **}
  */
  
  for(i=0; i<INTRPRED; i++) {
    if(fgets(line,MAXLEN,fp) == NULL) {
      fprintf(stderr,"readIntronFreq: bad format in line %d of %s\n",i+3,inFile);
      exit(EXIT_FAILURE);
    }
    off1 = 0;
    for(j=0;j<N_BASES;j++) {
      sscanf(line+off1,"%d%n",&freq,&off2);
      if(freq == 0)
        freq = 1;
      intronFreq[i][j]=freq;
      off1 += off2;
    }
  }

  /* now do totals */
  for(i=0; i < INTRPRED; i++) {
    sum = 0;
    for(j=0; j < N_BASES; j++)
      sum += intronFreq[i][j];
    intronFreq[i][N_BASES] = sum;
  }

  fclose(fp);
  return;
}


/*
** Just reads the mean intron length, which determines the 
** geometric parameter used for intron lengths.
*/
void readIntronLen(char *inFile) {
  FILE   *fp;
  char   line[MAXLEN];
  int    i;
  double mean,p;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntronLen: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAXLEN,fp) == NULL) { fprintf(stderr,"readIntronLen: file %s ended too early\n",inFile);
    exit(EXIT_FAILURE);
  }
  if(1 != sscanf(line,"%lf",&mean)) {
    fprintf(stderr,"readIntronLen: bad format\n");
    exit(EXIT_FAILURE);
  }
  p = (mean-1)/mean;

  for(i=I0; i < I0+3; i++) {
    logQ[i] = log(p);
    logP[i] = log(1-p);
  }

  fclose(fp);
  return;
}

/*
** Reads the intergenic frequencies.  Zero counts will be replaced by 1.
*/
void readIntergeneFreq(char *inFile) {
  FILE   *fp;
  char   line[MAXLEN];
  int    i,j,sum;
  int    off1,off2,freq;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntergeneFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAXLEN,fp) == NULL) {
    fprintf(stderr,"readIntergeneFreq: can't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  /*
  ** We could do some checking based on info in this line, but for now
  ** we don't.
  **if(2 != sscanf(line,"%s %d",str,&num)) {
  **  fprintf(stderr,"readIntergeneFreq: bad format in first line of %s\n",inFile);
  **  exit(EXIT_FAILURE);
  **}
  **if(num != INTRPRED) {
  **  fprintf(stderr,"readIntergeneFreq: in %s, should have %d entries\n",inFile,INTRPRED);
  **  exit(EXIT_FAILURE);
  **}
  */
  
  for(i=0; i<INTRPRED; i++) {
    if(fgets(line,MAXLEN,fp) == NULL) {
      fprintf(stderr,"readIntergeneFreq: bad format in line %d of %s\n",i+3,inFile);
      exit(EXIT_FAILURE);
    }
    off1 = 0;
    for(j=0,sum=0; j<N_BASES; j++) {
      sscanf(line+off1,"%d%n",&freq,&off2);
      if(freq == 0)
        freq = 1;
      intergenicFreq[i][j]=freq;
      sum += freq;
      off1 += off2;
    }
    for(j=0; j<N_BASES; j++)
      logNS[i][j] = log((double)intergenicFreq[i][j] / (double) sum);
  }

  fclose(fp);
  return;
}


/*
** Juset reads in the mean intergenic length, which will
** be used to estimate a geometric parameter.
*/
void readIntergeneLen(char *inFile) {
  FILE   *fp;
  char   line[MAXLEN];
  double mean,p;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readIntergeneLen: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  if(fgets(line,MAXLEN,fp) == NULL) {
    fprintf(stderr,"readIntergeneLen: file %s ended too early\n",inFile);
    exit(EXIT_FAILURE);
  }
  if(1 != sscanf(line,"%lf",&mean)) {
    fprintf(stderr,"readIntergeneLen: bad format in line 1\n");
    exit(EXIT_FAILURE);
  }

  p = (mean-1) / mean;
  logP[INTERGENE] = log(1-p);
  logQ[INTERGENE] = log(p);

  fclose(fp);
  return;
}


/* 
** Reads in exon frequencies.  Zero counts will be set to 1.
** The sum for each will be computed, but the conversion to
** log probabilites is done elsewhere, in the lookup table.
*/
void readExonFreq(char *inFile) {
  FILE *fp;
  char line[MAXLEN];
  int freq,off1,off2,nContexts,lineNo,phase,base,context;
  double freqSum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readExonFreq: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  /* Pick up the order of the Markov model from the first line */ 
  if(fgets(line,MAXLEN,fp) == NULL)
    error((char *)"readExonFreq: couldn't read first line\n");
  if(1 != sscanf(line,"Order %d",&exonOrder))
    error((char *)"readExonFreq: bad format in first line of %s\n",inFile);
  nContexts = (int) pow((double) N_BASES, (double) exonOrder);

  /* free old mem (if any) and declare the new mem */
  if(logExonFreq != NULL)
    free3dArray(logExonFreq,PHASES,N_BASES);
  logExonFreq = declare3dArray(PHASES,N_BASES,nContexts);

  for(lineNo=2,phase=0; phase<PHASES; phase++) {
    for(context=0; context<nContexts; context++,lineNo++) {
      if(fgets(line,MAXLEN,fp) == NULL)
        error((char *)"readExonFreq: bad format in line %d:\n%s\ncontext=%d, phase=%d",lineNo,line,context,phase);
      off1 = 0;
      for(base=0; base < N_BASES; base++) {
        sscanf(line+off1,"%d%n",&freq,&off2);
        off1 += off2;
        if(freq == 0)
          freq = 1;
        logExonFreq[phase][base][context] = freq;
      }
    }

    /* normalize the frequencies and take logs */
    for(context=0; context<nContexts; context++) {
      /* get the sum for the current context */
      for(freqSum=0,base=0; base<N_BASES; base++)
        freqSum += logExonFreq[phase][base][context];
      /* normalize & take logs context */
      for(base=0; base<N_BASES; base++) {
        if(logExonFreq[phase][base][context] > 0)
          logExonFreq[phase][base][context] = log((double)(logExonFreq[phase][base][context]) / freqSum);
        else
          logExonFreq[phase][base][context] = -FLT_MAX;
      }
    }

    /* skip the first line of the next section (it just tells the phase) */
    if((phase < PHASES-1) && (fgets(line,MAXLEN,fp) == NULL))
      error((char *)"readExonFreq: file ended too early\n");
    lineNo++;
  }

  fclose(fp);
  return;
}


/* 
** Reads the four exon length densities into the exonLenDensity[][] 
** array.  If option->geometric is 1 then it will compute geometric
** parameters from these.
*/
void readExonLen(Option *option) {

  /* free possibly existing memory */
  if(exonLenDensity[TYPE_SNGL] != NULL)  free(exonLenDensity[TYPE_SNGL]);
  if(exonLenDensity[TYPE_INIT] != NULL)  free(exonLenDensity[TYPE_INIT]);
  if(exonLenDensity[TYPE_INTL] != NULL)  free(exonLenDensity[TYPE_INTL]);
  if(exonLenDensity[TYPE_TERM] != NULL)  free(exonLenDensity[TYPE_TERM]);

  /* read in the length distributions */
  exonLenDensity[TYPE_SNGL] = readLengthDist(numExons+TYPE_SNGL,option->exonLenSingFile,1,option->verbose);
  exonLenDensity[TYPE_INIT] = readLengthDist(numExons+TYPE_INIT,option->exonLenInitFile,1,option->verbose);
  exonLenDensity[TYPE_INTL] = readLengthDist(numExons+TYPE_INTL,option->exonLenIntlFile,1,option->verbose);
  exonLenDensity[TYPE_TERM] = readLengthDist(numExons+TYPE_TERM,option->exonLenTermFile,1,option->verbose);
  if(option->verbose) printf("\n");

  return;
}


/*
** Adjusts the parameters if we're using geometric exon lengths.
** Mean lengths are estimated from the supplied densities, and
** the transition probabilites are adjusted to account for the fact
** that the lengths are restricted modulo 3.
**
** NOTE: exonLenDensity[][] must have been initialized before this!
*/
void geometricFix(Option *option) {
  int i,j,phase;
  double mean[PHASES];
  double avgMean,p,sum;
  double PP,ZZ,OO,TT;

  /* estimate geometric probabilities from exonLenDensity[][] */
  for(i=0;i<EXONTYPES;i++) {
    for(j=0; j<PHASES; j++)
      mean[j]=0;
    for(j=0, phase=0; j<numExons[i]; j++,phase++) {
      phase = phase % PHASES;
      mean[phase] += (j+1) * exp(exonLenDensity[i][j]);
    }
    for(j=0,avgMean=0; j<PHASES; j++)
      avgMean += mean[j];
    avgMean = (double) avgMean / (double) PHASES;

    p = (avgMean-1) / avgMean;
    exonQ[i] = log(p);
    exonP[i] = log(1-p);
  }

  /*
  ** Adjust transition probailities, which corrects for 
  ** the fact that lenths are restricted modulo 3.
  */
  PP = exp(exonQ[TYPE_INTL]);
  ZZ = log(PP*PP/(1+PP + PP*PP));
  OO = log(1/(1+PP + PP*PP));
  TT = log(PP/(1+PP + PP*PP));
  transitions[I0][I0] -= ZZ;
  transitions[I1][I1] -= ZZ;
  transitions[I2][I2] -= ZZ;
  transitions[I0][I1] -= OO;
  transitions[I1][I2] -= OO;
  transitions[I2][I0] -= OO;
  transitions[I0][I2] -= TT;
  transitions[I1][I0] -= TT;
  transitions[I2][I1] -= TT;

  PP = exp(exonQ[TYPE_INIT]);
  transitions[INTERGENE][I0] -= log(PP*PP/(1+PP + PP*PP));
  transitions[INTERGENE][I1] -= log(1/(1+PP + PP*PP));
  transitions[INTERGENE][I2] -= log(PP/(1+PP + PP*PP));

  if(option->verbose) {
    printf("Mean exon lengths (using geometric distribution)\n");
    for(i=0; i<EXONTYPES; i++)
      printf("  type %d = %f\n", i, 1/(1-exp(exonQ[i])) );
    printf("\n");
   
    printf("Transitions (adjusted for geometric lengths)\n");
    printf("      N      I0      I1     I2\n  ");
    for(i=0; i<NUMSTATES; i++) {
      for(j=0,sum=0; j<NUMSTATES; j++) {
        printf("%2.5f ",exp(transitions[i][j]));
        sum += exp(transitions[i][j]);
      }
      printf("   (sum = %g)\n  ",sum);
    }
    printf("\n");
  }

  return;
}


/*
** read in the transition probabilities
*/
void readTransitions(char *inFile, Option *option) {
  FILE *fp;
  char line[MAXLEN];
  char str[MAXLEN];
  double freq,sum;
  int i,j,off1,off2;


  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readTransitions: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  /* skip the first line */
  if(fgets(line,MAXLEN,fp) == NULL) {
    fprintf(stderr,"readTransitions: couldn't read first line of %s\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0; i<NUMSTATES; i++) {
    /* read in the next line */
    if(fgets(line,MAXLEN,fp) == NULL) {
      fprintf(stderr,"readTransitions: couldn't read line %d of %s\n",i+1,inFile);
      exit(EXIT_FAILURE);
    }

    /* skip the initial string */
    if(1 != sscanf(line,"%s%n",str,&off1)) {
      fprintf(stderr,"readTransitions: bad start to line %d of %s\n",i+1,inFile);
      exit(EXIT_FAILURE);
    }

    /* read the transitions */
    for(j=0,sum=0; j<NUMSTATES; j++) {
      if(1 != sscanf(line+off1,"%lf%n",&freq,&off2)) {
        fprintf(stderr,"readTransitions: bad format, line %d of %s\n",i+1,inFile);
        fprintf(stderr,"  Here's the offending line...\n%s\n",line);
        exit(EXIT_FAILURE);
      }
      if(freq < 0) {
        fprintf(stderr,"readTransitions: all entries must be >= 0 in %s\n",inFile);
        fprintf(stderr,"  (entry[%d][%d] isn't)\n",i+1,j+1);
        exit(EXIT_FAILURE);
      } else if (freq == 0) {
        freq++;
      }
      /*
      ** the transitions file is really written for fullphat, so for halfphat
      ** we need to double the frequencies of transitions from intergene to 
      ** the initial exons to account for exons on the reverse strand.
      */
      if(i==0 && j > 0)
        freq += freq;
      transitions[i][j]=freq;
      off1 += off2;
      sum += freq;
    }

    /* normalize the transitions */
    for(j=0; j<NUMSTATES; j++)
      transitions[i][j] = log(transitions[i][j] / sum);
  }

  /*
  ** print out all transitions if verbose
  */
  if(option->verbose) {
    printf("Transitions\n");
    printf("      N      I0      I1     I2\n  ");
    for(i=0; i<NUMSTATES; i++) {
      for(j=0,sum=0; j<NUMSTATES; j++) {
        printf("%2.5f ",exp(transitions[i][j]));
        sum += exp(transitions[i][j]);
      }
      printf("   (sum = %g)\n  ",sum);
    }
    printf("\n");
  }

  return;
}


/*
** Read in the intial probabilities of the HMM.
*/
void readInitProbs(char *inFile) {
  FILE   *fp;
  char   line[MAXLEN];
  int    i;
  double sum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readInitProbs: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0,sum=0; i<NUMSTATES; i++) {
    if(fgets(line,MAXLEN,fp) == NULL) {
      fprintf(stderr,"readInitProbs: file %s ended too early\n",inFile);
      exit(EXIT_FAILURE);
    }
    if(1 != sscanf(line,"%lf",logInit+i)) {
      fprintf(stderr,"readInitProbs: bad format in line %d\n",i+1);
      exit(EXIT_FAILURE);
    }
    if(logInit[i] < 0) {
      fprintf(stderr,"readInitProbs: entries in %s must be >= 0\n",inFile);
      exit(EXIT_FAILURE);
    } else if (logInit[i] == 0) {
      logInit[i] += 1;
    }
    sum += logInit[i];
  }

  /*
  ** Double the intron probabilities, as the file is written
  ** for fullphat.
  */
  for(i=1; i<NUMSTATES; i++) {
    logInit[i] += logInit[i];
    sum += logInit[i];
  }

  /* normalize */
  for(i=0; i<NUMSTATES; i++)
    logInit[i] = log(logInit[i]/sum);

  fclose(fp);
  return;
}

/*
** read the probabilities of the 3 possible stop codons.
*/
void readStopProbs(char *inFile, Option *option) {
  FILE   *fp;
  char   line[MAXLEN];
  int    i;
  double sum;

  if ((fp = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"readStopProbs: can't open %s for read, exiting\n",inFile);
    exit(EXIT_FAILURE);
  }

  for(i=0,sum=0; i<NSTOPS; i++) {
    if(fgets(line,MAXLEN,fp) == NULL) {
      fprintf(stderr,"readStopProbs: file %s ended too early\n",inFile);
      exit(EXIT_FAILURE);
    }
    if(1 != sscanf(line,"%lf",logStop+i)) {
      fprintf(stderr,"readStopProbs: bad format in line %d\n",i+1);
      exit(EXIT_FAILURE);
    }
    if(logStop[i] <= 0) {
      fprintf(stderr,"readStopProbs: entries in %s must be positive\n",inFile);
      exit(EXIT_FAILURE);
    }
    sum += logStop[i];
  }

  for(i=0; i<NSTOPS; i++)
    logStop[i] = log(logStop[i]/sum);

  if(option->verbose) {
    printf("Stop codon probabilities (TAA/TAG/TGA)\n");
    for(i=0; i<NSTOPS; i++)
      printf("  stop[%d] = %f\n",i,exp(logStop[i]));
    printf("\n");
  }

  fclose(fp);
  return;
}


/*
** Returns initial probability of state j
*/
double pinit(int j) {
  return logInit[j];
}

/*
** Returns self-transition probability for state j
*/
double pextra(int j) {
  return logQ[j];
}

/*
** Returns non-self-transition probability for state j
*/
double pnotextra(int j) {
  return logP[j];
}


/*
** Returns the probability of a transition from x to y, 
** given that x doesn't have a self transition
*/
double ptrans(int x, int y) {
  return transitions[x][y];
}


/*
** Returns log(b_state(seq[pos])) for state = intron or intergene.
** b_state(seq[pos]) is defined as in Rabiner.
**
** The value is returned from the lookup table initialized in lookupIntronIntergene.
** 
*/
double bIntronIntergene(int pos, int state, int slen) {
  double p;

  if((pos < 0) || (pos >= slen)) {
    fprintf(stderr,"bIntronIntergene: pos=%d, should be in range [0,%d]\n",pos,slen-1);
    exit(EXIT_FAILURE);
  }

  switch(state) {
    case INTERGENE:
      p = intergenicProbs[pos];
    break;

    case I0:
    case I1:
    case I2:
      p = intronProbs[pos];
    break;

    default:
      fprintf(stderr,"bIntronIntergene: Invalid state: %d\n",state);
      p=-FLT_MAX;
    break;
  }

  return(p);
}


double exonProb(int start,int stop, int phase, int type, char *str, Option *option) {
  double prob;
  double donProb=0;
  double accProb=0;
  double exProb=0;
  double lenProb=0;
  int zphase;
  int exonstart,exonstop,exonLen;


  exonstart = start;
  exonstop  = stop;

  if(option->debug)
    printf("\nexonProb(start=%d, stop=%d, phase=%d, type=%d)\n",start,stop,phase,type);

  switch(type) {
    case TYPE_SNGL:
      if(option->debug) printf("  Single Exon\n");
      exonstart += 3;               /* cut out the ATG (no score needed, prob = 1) */

      /* cut out and score the stop signal */
      exonstop  -= 3;
      /* prob = log(1/3.0); */
      prob = getStopProb(str,exonstop+1);

      exonLen = (exonstop+3) - (exonstart-3) + 1;  /* compute length of coding portion */
      if((exonLen % PHASES) != 0)
        fprintf(stderr,"exonProb: Warning, passed a single exon of length %d\n",exonLen);
    break;

    case TYPE_INIT:
      if(option->debug) printf("  Initial Exon\n");
      exonstart += 3;               /* cut out the ATG (no score needed, prob = 1) */

      exonstop = stop-(don_upstr+don_dnstr);    /* cut out and score donor */
      donProb = scoreDonor(str+(exonstop+1),exonstop+1);
      prob = donProb;

      exonLen = (exonstop+don_upstr) - (exonstart-3) + 1;  /* compute length of coding portion */
    break;

    case TYPE_INTL:
      if(option->debug) printf("  Internal Exon\n");
      exonstop = stop-(don_upstr+don_dnstr);    /* cut out and score donor */
      donProb = scoreDonor(str+(exonstop+1),exonstop+1);
      prob = donProb;

      exonstart = start+(acc_upstr+acc_dnstr);  /* cut out and score acceptor */
      accProb = scoreAcceptor(str+start,start);
      prob += accProb;

      exonLen = (exonstop+don_upstr) - (exonstart-acc_dnstr) + 1;  /* compute length of coding portion */
    break;

    case TYPE_TERM:
      if(option->debug) printf("  Terminal Exon\n");
      exonstart = start+(acc_upstr+acc_dnstr);  /* cut out and score acceptor */
      accProb = scoreAcceptor(str+start,start);
      prob = accProb;

      exonstop  -= 3;               /* cut out and score the stop signal */
      /* prob += log(1/3.0); */
      prob += getStopProb(str,exonstop+1);

      exonLen = (exonstop+3) - (exonstart-acc_dnstr) + 1;  /* compute length of coding portion */
    break;

    default:
      fprintf(stderr,"exonProb: bad exon-type passed in\n");
      exit(1);
    break;
  }

  /* just a warning if exonstart > exonstop, but this shouldn't happen */
  if(exonstart > exonstop) {
    fprintf(stderr,"exonProb: exonstart >= exonstop!  Should not have happened\n");
    printf("  (start,stop) = (%d,%d)\n",start,stop);
    printf("  (exonstart,exonstop) = (%d,%d)\n",exonstart,exonstop);
    printf("  (phase,type) = (%d,%d)\n",phase,type);
    exit(1);
  }

  /* score the coding potential of the exon part */
  /* (this phase stuff needs to be fixed now that we have splice site models, though it
  ** still works as long as don_upstr and acc_dnstr are multiples of 3 */
  zphase = (exonstart + (PHASES -phase) ) % PHASES;
  zphase = (phase + PHASES -(exonstart % PHASES)) % PHASES;
  exProb = exonProbLUT[zphase][exonstop+1] - exonProbLUT[zphase][exonstart];
  prob += exProb;

  /* score the length */
  if(option->geometric) {
    /* use geometric length distribution */
    lenProb = exonP[type] + exonQ[type] * (exonLen);
  } else {
    /* use a general length distribution */
    lenProb = exonLenDensity[type][exonLen-1];
  }
  prob += lenProb;
  

  /* debugging code */
  if(option->debug) {
    printf("  exon[%d,%d] scores %2.15f (frame = %d)\n",exonstart,exonstop,prob,zphase);
    printf("  len[%d] scores %2.15f\n",exonLen,lenProb);
    if(donProb != 0)
      printf("  don[%d,%d] scores %2.15f\n",exonstop+1,exonstop+don_upstr+don_dnstr,donProb);
    if(accProb != 0)
      printf("  acc[%d,%d] scores %2.15f\n",start,start+acc_upstr+acc_dnstr-1,accProb);
    /*
    counter++;
    if(counter == 100)
      exit(1);
    */
  }

  return prob;
}


/*
** Scans through whole sequence to set the following two arrays...
**   P_intergene (Y_t|Y_0^t-1) = intergeneProbs[t]
**   P_intron    (Y_t|Y_0^t-1) = intronProbs[t]
** (indices are zero-based)
** 
** This takes up an extra chunk of memory, but avoids repeated calculation of contexts 
** (such as would otherwise happen in calculation of the backward variables, for example).
**
*/
void lookupIntronIntergene(char *str, int slen) {
  static int first=1;
  static int oldSlen=0;
  int i,context,c;
  double p,q;

  /* 
  ** If this is the first time we're in this function, or if
  ** the length of the sequence is different to last time, then
  ** reserve memory for intronProbs and intergenicProbs.
  */
  if(first || (slen != oldSlen) ) {
    oldSlen = slen;
    
    /* free the old memory, if any */
    if(!first) {
      free(intronProbs);
      free(intergenicProbs);
    } else {
      first=0;
    }

    /* allocate the new memory */
    intronProbs = (double *) malloc ((slen+1) * sizeof(double));
    if(intronProbs == NULL) {
      fprintf(stderr, "lookupIntronIntergene: can't allocate %d doubles for intronProbs\n",i);
      exit(1);
    }

    intergenicProbs = (double *) malloc ((slen+1) * sizeof(double));
    if(intergenicProbs == NULL) {
      fprintf(stderr, "lookupIntronIntergene: can't allocate %d doubles for intergenicProbs\n",i);
      exit(1);
    }

  }

  /* set up the initial context */
  for(context=0, i=-INTRORDER; i<0; i++) {
    context *= N_BASES;
    context += str[i];
  }

  for(i=0;i<slen;i++) {
    c = str[i];

    /* 
    ** Intergene: only take the log if the word actually occurs
    */
    p = logNS[context][c];
    if(p == 0) {
      q = (double) intergenicFreq[context][c] /
        (double) intergenicFreq[context][N_BASES];
      p = log(q);
      logNS[context][c] = p;
    }
    intergenicProbs[i] = logNS[context][c];

    /* 
    ** Intron: only take the log if the word actually occurs
    */
    p = logFS[context][c];
    if(p == 0) {
      q = (double) intronFreq[context][c] /
        (double) intronFreq[context][N_BASES];
      p = log(q);
      logFS[context][c] = p;
    }
    intronProbs[i] = logFS[context][c];

    /* advance the context */
    context *= N_BASES;
    context %= INTRPRED;
    context += c;
  }

  return;
}



/* 
** In exonProbLUT, we use element 0 for pos -1 and so on.
** Thus if b is a zero-based index, P(Y_0^b) is given by exonProbLUT[b+1] (ignoring phase)
*/
void lookupExon(char *str,int slen) {
  int i,nContexts,frame,context,c;
  int phase[PHASES];
  static int first=1;
  static int oldSlen=0;

  nContexts = (int) pow((double) N_BASES, (double) exonOrder);

  /* 
  ** If this is the first time we're in this function, or if
  ** the length of the sequence is different to last time, then
  ** reserve memory for exonProbLUT
  */
  if(first || (slen != oldSlen) ) {
    oldSlen = slen;
    
    /* free the old memory, if any */
    if(!first) {
      for(i=0; i<PHASES; i++) {
        free(exonProbLUT[i]);
      }
    } else {
      first=0;
    }

    /* allocate the new memory */
    for(i=0;i<PHASES;i++) {
      exonProbLUT[i] = (double *) malloc ((slen+1) * sizeof(double));
      if(exonProbLUT[i] == NULL) {
        fprintf(stderr, "lookupExon: can't allocate %d doubles for exonProbLUT[%d]\n",slen,i);
        exit(1);
      }
    }
  }

  for(i=0; i<PHASES; i++) {
    exonProbLUT[i][0] = 0;
    phase[i] = i;
  }

  /* get initial context */
  for(context=0, i=-exonOrder; i<0; i++) {
    context *= N_BASES;
    context += str[i];
  }

  /* score the sequence */
  for(i=0; i<slen; i++) {
    c = str[i];

    for(frame=0; frame<PHASES; frame++) /* get new probs */
      exonProbLUT[frame][i] = exonProbLUT[frame][i-1] + logExonFreq[phase[frame]][c][context];

    for(frame=0; frame<PHASES; frame++) {
      phase[frame]++;
      phase[frame] %= PHASES;
    }
    context *= N_BASES;
    context %= nContexts;
    context += c;
  }
  return;
}


/*
** Read a length density from filename.
** Returns a poitner to newly allocated memory holding the log of the density.
** If tertiaryNormalise is 1 then it will be normalized such that the sums
** over length modulo 3 will be 1.  Otherwise it is normalized in the usual manner.
** The max len of the density is written into *n.
*/
double *readLengthDist(int *n, char *fileName, int tertiaryNormalise, int verbose) {
  FILE *dataFile;
  char line[MAXLEN];
  int i,j;
  double sum[PHASES];
  double *lenDist,*lenDistPointer;

  if(verbose) printf("Reading length density from %s\n",fileName);
  
  if ((dataFile = fopen(fileName,"r")) == NULL) {
    fprintf(stderr,"readLengthDist: can't open %s for read, exiting\n",fileName);
    exit(1);
  }

  /* find out how many entries there are to be read */
  fgets(line,MAXLEN,dataFile);
  sscanf(line,"%d ",n);
  if(verbose) printf("  (max length is %d)\n",*n);

  /* declare the memory for lenDist */
  if(NULL == (lenDist = (double *) malloc ((*n) * sizeof(double))) ) {
    fprintf(stderr,"readLengthDist: unable to get mem for %d doubles, exiting\n",*n);
    exit(EXIT_FAILURE);
  }

  /* read in the entries */
  for(i=0,lenDistPointer=lenDist;  i<*n;  i++,lenDistPointer++) {
    if(! fgets(line,MAXLEN,dataFile)) {
      fprintf(stderr,"readLengthDist: file %s shorter than expected\n  (On line %d, expected %d)\n",fileName,i+2,*n+1);
      exit(EXIT_FAILURE);
    }
    if(1 != (sscanf(line,"%lf ",lenDistPointer)) ) {
      fprintf(stderr,"readLengthDist: couln't read a double from line %d of %s\n",i+2,fileName);
      exit(EXIT_FAILURE);
    }
    /* *lenDistPointer = exp(*lenDistPointer); */
  }
  fclose(dataFile);

  if(tertiaryNormalise) {
    /* Normalize the distribution such that  sum_{i mod PHASES = c} len_i = 1  for c=0,1,2 */
    /* This is for exon length distributions, as their length mod PHASES is fixed */

    /* initialize the sums to zero */
    for(j=0; j<PHASES; j++)
      sum[j] = 0;

    /* calculate the sums */
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++) {
      sum[j++] += *lenDistPointer++;
      j = j % PHASES;
    }

    /*
    for(j=0; j<PHASES; j++)
      printf("  sum[%d] = %g\n",j,sum[j]);
    */

    /* normalize the densities */
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++,lenDistPointer++) {
      *lenDistPointer /= sum[j++];
      *lenDistPointer = (*lenDistPointer > 0) ? log(*lenDistPointer) : -FLT_MAX;
      j = j % PHASES;
    }

    /* A check to see if everything was normalized OK
    for(j=0; j<PHASES; j++)
      sum[j] = 0;
    lenDistPointer=lenDist;
    for(i=0,j=0; i<*n; i++) {
      sum[j++] += exp(*lenDistPointer++);
      j = j % PHASES;
    }
    for(j=0; j<PHASES; j++)
      printf("  sum[%d] = %g\n",j,sum[j]);
    */

  } else {
    /* normalize the distribution in the usual manner */

    /* calculate the sum */
    sum[0] = 0;
    lenDistPointer=lenDist;
    for(i=0; i<*n; i++)
      sum[0] += *lenDistPointer++;
    /* printf("  sum = %g\n",sum[0]); */

    /* normalize the density */
    lenDistPointer=lenDist;
    for(i=0; i<*n; i++,lenDistPointer++) {
      *lenDistPointer /= sum[0];
      *lenDistPointer = (*lenDistPointer > 0) ? log(*lenDistPointer) : -FLT_MAX;
    }
  }

  return(lenDist);
}


/*
** return the max possible length of an exon of the given type.
*/
int maxExonLen(int type) {
  switch(type) {
    case TYPE_SNGL:
      return(numExons[TYPE_SNGL]);
    break;

    case TYPE_INIT:
      return(numExons[TYPE_INIT]);
    break;

    case TYPE_INTL:
      return(numExons[TYPE_INTL]);
    break;

    case TYPE_TERM:
      return(numExons[TYPE_TERM]);
    break;

    default:
      fprintf(stderr,"maxExonLen: bad type passed in: %d\n",type);
      exit(EXIT_FAILURE);
    break;
  }
}

double getStopProb(char *str, int loc) {
  int i,j,context;
  double ans;

  for(i=0,j=loc,context=0; i<STOPLEN; i++,j++) {
    context *= N_BASES;
    context += str[j];
  }

  switch(context) {
    case 48:            /* TAA */
      ans = logStop[0];  
    break;
    
    case 50:            /* TAG */
      ans = logStop[1];
    break;
    
    case 56:            /* TGA */
      ans = logStop[2];
    break;
    
    default:            /* none of the above */
      fprintf(stderr,"getStopProb: can't score a non-stop codon\n");
      fprintf(stderr,"  (This shouldn't have happened)\n");
      fprintf(stderr,"  loc is %d, context = %d\n",loc,context);
      fprintf(stderr,"  bases are %d, %d, %d\n",str[loc],str[loc+1],str[loc+2]);
      exit(EXIT_FAILURE);
    break;
  }
    
  return(ans);
}


void free3dArray( double ***array, int dim1, int dim2) {
  int i,j;
  if(array != NULL) {
    for(i=0; i<dim1; i++) {
      if(array[i] != NULL) {
        for(j=0; j<dim2; j++) {
          if(array[i][j] != NULL) {
            free(array[i][j]);
          }
        }
      }
      free(array[i]);
    }
    free(array);
  }
  return;
}


double ***declare3dArray(int dim1,int dim2,int dim3) {
  int i,j;
  double ***x;

  x = (double ***) myAlloc(dim1 * sizeof(double **));
  for(i=0; i<dim1; i++) {
    x[i] = (double **) myAlloc(dim2 * sizeof(double *));
    for(j=0; j<dim2; j++) {
      x[i][j] = (double *) myAlloc(dim3 * sizeof(double));
    }
  }

  return(x);
}

