/***********************************************************************
************************************************************************
**                                                                    **
** Phat - a system for predicting gene structures in DNA.             **
** Copyright (C) 2000 Simon Cawley and Anthony Wirth                  **
**                                                                    **
** This program is free software; you can redistribute it and/or      **
** modify it under the terms of the GNU General Public License        **
** as published by the Free Software Foundation; either version 2     **
** of the License, or (at your option) any later version.             **
**                                                                    **
** This program is distributed in the hope that it will be useful,    **
** but WITHOUT ANY WARRANTY; without even the implied warranty of     **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      **
** GNU General Public License for more details.                       **
**                                                                    **
** You should have received a copy of the GNU General Public License  **
** along with this program; if not, write to the Free Software        **
** Foundation, Inc.,                                                  **
** 59 Temple Place - Suite 330,                                       **
** Boston, MA  02111-1307, USA.                                       **
**                                                                    **
************************************************************************
***********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#define EXTERN extern
#include "half_common.h"
#include "half_splice.h"
#include "half_probs.h"
#undef EXTERN

#define EXTERN
#include "half_seq.h"
#undef EXTERN

#define PROB_SIG_FIG 2

typedef struct bt{
  int state;
  int loc;
  short type;
} backtrack;

#define MAX_WORD_SIZE 100
#define MAX_LINE_SIZE 1024

static char *sequence;
static char *seq;
static int slen;

static int fInitSt[NUMPOSNS];
static int fOtherSt[NUMPOSNS];
static int fOtherEnd[NUMPOSNS];
static int fTermEnd[NUMPOSNS];

static int fis;
static int fos;
static int foe;
static int fte;

static int curfis;
static int curfos;
static int curfoe;
static int curfte;


static double *fb[NUMSTATES];
static double *mygamma[NUMSTATES];
static backtrack *remember[NUMSTATES];
static double *al[NUMSTATES];
static double *alpha[NUMSTATES];
static double *be[NUMSTATES];
static double *beta[NUMSTATES];
static double likelihood;

static double seqProb;

int readSequence(Option *option) {
  FILE *seqFile=NULL;
  /* int y,i,done,c,n,cNum; */
  int base,i,cNum;
  int format;
  
  /*
  ** Allocate memory for sequence and set up the pointers.
  ** sequence will point to the start of memory, seq will point
  ** to the start of the sequence, 5 bases further on to allow for the
  ** cycling round of the last 5 bases to the start.
  */
  slen = getSeqLen(option->inFile);
  sequence = (char *) malloc((slen+EXTRA) * sizeof(char));
  if(sequence == NULL) {
    fprintf(stderr,"%s",NOMEM);
    exit(EXIT_FAILURE);
  }
  seq = sequence + EXTRA;


  /*
  ** determine format of sequence file
  */
  if( (format = getSeqFormat(option->inFile)) == UNKNOWN_FORMAT) {
    fprintf(stderr,"getSeqLen: bad format in %s",option->inFile);
    exit(EXIT_FAILURE);
  }

  /*
  ** read in the sequence 
  */
  openSeq(&seqFile,option->inFile,format);
  for( i=0, base = getNextBase(&seqFile,format); 
       base != EOF; 
       i++, base = getNextBase(&seqFile,format) ) {
    cNum = base2num(base);
    if(cNum > 3) {
      fprintf(stderr,"Error: readSequence: can't handle non-ACGT bases in %s\n",option->inFile);
      fprintf(stderr,"  (read character %c, treating it as \"A\")\n",base);
      cNum = 0;
    }
    seq[i] = cNum;
    if(i > slen) {
      fprintf(stderr,"readSequence: input sequence longer than I thought it was!\n");
      exit(EXIT_FAILURE);
    }
  }
  fclose(seqFile);

  for(i=-EXTRA;i<0;i++)
    seq[i] = seq[slen+i];

  /*
  ** Allocate the memory for the gamma (used in Viterbi), alpha (forward)
  ** and beta (backward) variables.
  */
  for(i=0;i<NUMSTATES;i++) {
    fb[i] = (double *) malloc((slen+2)*sizeof(double));
    remember[i] = (backtrack *) malloc((slen+1)*
        (sizeof(backtrack)));
    if(fb[i] == NULL || remember[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    mygamma[i] = fb[i]+1;
    al[i] = (double *) malloc((slen+2)*sizeof(double));
    if(al[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    alpha[i] = al[i]+1;
    be[i] = (double *) malloc((slen+2)*sizeof(double));
    if(be[i] == NULL) {
      fprintf(stderr,"%s",NOMEM);
      exit(EXIT_FAILURE);
    }
    beta[i] = be[i]+1;
  }

  lookupExon(seq,slen);
  lookupIntronIntergene(seq,slen);
  donLookupInit(slen,option);
  accLookupInit(slen,option);

  return slen;
}

/*
** returns the length of the sequence to be analyzed
*/
int getSeqLen(char *inFile) {
  FILE *seqFile;
  int format,seqLen;
  int base;

  if( (format = getSeqFormat(inFile)) == UNKNOWN_FORMAT) {
    fprintf(stderr,"getSeqLen: bad format in %s",inFile);
    exit(EXIT_FAILURE);
  }

  openSeq(&seqFile,inFile,format);
  base = getNextBase(&seqFile,format);
  seqLen = 0;
  while (base != EOF) {
    seqLen++;
    base = getNextBase(&seqFile,format);
  }
  fclose(seqFile);

  return(seqLen);
}


/*
** opens inFile (whose format is specified by the int format) and
** returns the fist base.  FILE *seqFile will now point to the next
** character in the file
*/
void openSeq(FILE **seqFile, char *inFile, int format) {
  char word[MAX_WORD_SIZE];
  char line[MAX_LINE_SIZE];

  if (((*seqFile) = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"openSeq: can't open %s for read, exiting\n",inFile);
    exit(1);
  }


  /* 
  ** figure out the format, skip over header to the start of sequence
  */
  if(NULL == fgets(line,MAX_LINE_SIZE,*seqFile)) {
    fprintf(stderr,"Error: openSeq: problem reading first line of %s \n",inFile);
    exit(1);
  }

  if(1 != sscanf(line,"%s",word)) {
    fprintf(stderr,"Error: openSeq: first line of %s is blank\n",inFile);
    exit(1);
  }

  if(format==FASTA) {
    /* already read first line, nothing more to do */
   return;
  } else if (format==GENBANK) {
    /* read everything up to the line starting with "ORIGIN" */
    for(;;) {
      /* read ahead to skip over the header */
      if(NULL == fgets(line,MAX_LINE_SIZE,*seqFile)) {
        fprintf(stderr,"openSeq: incorrect genbank format in %s\n",inFile);
        fprintf(stderr,"  (never found keyword \"ORIGIN\")\n");
        exit(1);
      }
      sscanf(line,"%s",word);
      if(0 == strcmp(word,"ORIGIN"))
        break;
    }
    return;
  } else {
    fprintf(stderr,"openSeq: don't know how to process format %d\n",format);
    exit(EXIT_FAILURE);
  }
}


/*
** return the next character that is supposed to be a DNA base in 
** the file.  This could still be an N or a * or something, checking 
** isn't done here.  EOF returned when EOF or the end of the sequence
** is reached.
*/
char getNextBase(FILE **seqFile, int format) {
  int c;

  if(format == FASTA) {
    c = fgetc(*seqFile);
    for( ; ; c = fgetc(*seqFile) ) {
      if(c == EOF) {
        break;
      } else if(isspace(c)) {
        continue;
      } else if(c == '>') {
        /* ignore second sequence, treat as EOF */
        c = EOF;
        break;
      } else {
        break;
      }
    }
  } else if (format == GENBANK) {
    c = fgetc(*seqFile);
    for( ; ; c = fgetc(*seqFile) ) {
      if( (c == EOF) || (c == '/') ) {
        c = EOF;
        break;
      } else if( isspace(c) || isdigit(c) ) {
        continue;
      } else {
        break;
      }
    }
  }

  return(c);
}


/*
** opens inFile and determines whether the format is FASTA, GENBANK
** or UNKNOWN_FORMAT
*/
int getSeqFormat(char *inFile) {
  FILE *seqFile;
  char word[MAX_WORD_SIZE];
  char line[MAX_LINE_SIZE];
  int seqFileFormat;

  if ((seqFile = fopen(inFile,"r")) == NULL) {
    fprintf(stderr,"getSeqFormat: can't open %s for read, exiting\n",inFile);
    exit(1);
  }

  /* 
  ** figure out format, skip over the header to the start of sequence
  */
  if(NULL == fgets(line,MAX_LINE_SIZE,seqFile)) {
    fprintf(stderr,"Error: getSeqFormat: problem reading first line of %s\n",inFile);
    exit(1);
  }

  if(1 != sscanf(line,"%s",word)) {
    fprintf(stderr,"Error: getSeqFormat: first line of %s is blank\n",inFile);
    exit(1);
  }

  if(word[0] == '>') {
    /* file is assumed to be in fasta format */
    seqFileFormat = FASTA;
  } else if (0 == strcmp(word,"LOCUS")) {
    /* file is assumed to be in genbank format */
    seqFileFormat = GENBANK;
  } else {
    seqFileFormat = UNKNOWN_FORMAT;
  }

  /* printf("fileFormat = %d\n",seqFileFormat); */
  fclose(seqFile);
  return(seqFileFormat);
}



/* 
** reverses complements the sequence and resets the lookup table
*/
void reverseSequence(Option *option) {
  int i,halfway,slenLess1,temp;

  if(slen <= 0) {
    fprintf(stderr,"Warning: reverseSequence: request to reverse an empty sequence\n");
    return;
  }

  /*
  if(option.debug) {
    printf("Sequence (before reversal)\n");
    for(i=0; i<slen; i++)
      printf("%d",seq[i]);
    printf("\n\n");
  }
  */
    
  halfway = (int) floor(slen/2.0);
  slenLess1 = slen-1;
  for(i=0; i<halfway; i++) {
    temp = 3-seq[i];
    seq[i] = 3-seq[slenLess1-i];
    seq[slenLess1-i] = temp;
  }
  if( (slen % 2) == 1 )
    seq[halfway] = 3-seq[halfway];
  
  /*
  if(option.debug) {
    printf("Sequence (after reversal)\n");
    for(i=0; i<slen; i++)
      printf("%d",seq[i]);
    printf("\n\n");
  }
  */
    
  lookupExon(seq,slen);
  lookupIntronIntergene(seq,slen);
  donLookupInit(slen,option);
  accLookupInit(slen,option);

  return;
}

/*
** locates all the potential start codons, stop codons, donor sites 
** and acceptor sites in the sequence, and records their numbers 
** and locations.
*/
void locateBounds(Option *option) {
  int i;
  int x,y;

  fis=0;
  fos=0;
  foe=0;
  fte=0;

  /*
  fInitSt[fis++]=0;
  fOtherSt[fos++]=0;
  */

  for(x=0, i=0; i<STORDER-1; i++) {
    x *= N_BASES;
    x += seq[i];
  }

  /* This bit is a hack assuming that SPLORDER = STORDER - 1 */
  y = x;

  if( (isfacc(y)) && (i-acc_upstr+1 > 0) )
    fOtherSt[fos++] = i-acc_upstr+1;
  if( (isfdon(y)) && (i-SPLORDER+don_dnstr+1 < slen) )
    fOtherEnd[foe++] = i-SPLORDER+don_dnstr+1;

  for(;i<slen;i++) {
    x *= N_BASES;
    x %= STCUT;
    x += seq[i];
    y = x % SPCUT;
    if(isfstart(x)) {
      if(fis >= NUMPOSNS) {
        fprintf(stderr,"Error: sequence has too many possible start codons.\n");
        fprintf(stderr,"  Shorten input, or recompile with a larger value for NUMPOSNS\n");
        exit(1);
      }
      fInitSt[fis] = i-STORDER+1;
      fis++;
    }
    if(isfend(x)) {
      if(fte >= NUMPOSNS) {
        fprintf(stderr,"Error: sequence has too many possible stop codons.\n");
        fprintf(stderr,"  Shorten input, or recompile with a larger value for NUMPOSNS\n");
        exit(1);
      }
      fTermEnd[fte] = i+1;
      fte++;
    }
    if( (isfacc(y)) && (i-acc_upstr+1 > 0) ) {
      if(fos >= NUMPOSNS) {
        fprintf(stderr,"Error: sequence has too many possible acceptor sites.\n");
        fprintf(stderr,"  Shorten input, or recompile with a larger value for NUMPOSNS\n");
        exit(1);
      }
      fOtherSt[fos] = i-acc_upstr+1;
      fos++;
    }
    if( (isfdon(y)) && (i-SPLORDER+don_dnstr+1 < slen) ) {
      if(foe >= NUMPOSNS) {
        fprintf(stderr,"Error: sequence has too many possible donor sites.\n");
        fprintf(stderr,"  Shorten input, or recompile with a larger value for NUMPOSNS\n");
        exit(1);
      }
      fOtherEnd[foe] = i-SPLORDER+don_dnstr+1;
      foe++;
    }
  }
  /*
  fTermEnd[fte++] = slen;
  fOtherEnd[foe++] = slen;
  */

  curfis = fis-1;
  curfos = fos-1;
  curfoe = 0;
  curfte = 0;

  if (option->debug) {
    printf("\nlocateBounds: on exit, have located:\n");
    printf("%7d possible start codons\n",fis);
    printf("%7d possible stop codons\n",fte);
    printf("%7d possible donors\n",foe);
    printf("%7d possible acceptors\n",fos);
    printf("in %d bases of input sequence\n",slen);
  }
}


int viterbi(Option *option) {
  int i,j;
  double f,g,h,maxq,maxp;
  int last_state;

  /* initialization */
  for(j=0;j<NUMSTATES;j++) {
    f = pinit(j);
    g = bIntronIntergene(0,j,slen);

    mygamma[j][-1] = f;
    alpha[j][-1] = f;
    remember[j][-1].state = j;
    remember[j][-1].loc = -1;
    remember[j][-1].type = NADA;

    mygamma[j][0] = f + g;
    alpha[j][0] = f+g;
    remember[j][0].state = j;
    remember[j][0].loc = -1;
    remember[j][0].type = NADA;
  }

  /* iteration */
  for(i=1;i<slen;i++) {
  /*
  if(i == 29)
    printf("######\n######\n######\n######\n######\n");
  */
    for(j=0;j<NUMSTATES;j++) {
      /* probabilities if continuing state  */
      h = pextra(j);
      g = bIntronIntergene(i,j,slen);
        /* g factors out as a multiplier in the forward and viterbi formulae, */
        /* so can be added in later */
      f = mygamma[j][i-1] + h;
      alpha[j][i] = alpha[j][i-1]+h;

      maxq = -FLT_MAX;
      switch(j) {
        case INTERGENE:
        maxq = endforward(i,maxq,option);
        break;

        case I0:
        case I1:
        case I2:
        maxq = donorforward(i,j,maxq,option);
        break;

        default:
        fprintf(stderr,"Error in viterbi: bad state\n");
        break;
      }
      if(f > maxq) {
        remember[j][i].state = j;
        remember[j][i].loc = i-1;
        remember[j][i].type = NADA;
        mygamma[j][i] = f+g;
      } else {
        mygamma[j][i] = maxq+g;
        if(option->debug) {
          printf("Local best(%d,%5d): ",j,i+1);
          printBestExon(j,i);
        }
      }
      alpha[j][i] +=g;
      /* printf("adding in %2.15lf to alpha[%d][%d], value is now %2.15lf\n",g,j,i,alpha[j][i]); */
    }
/* printf("  %7.2lf\t%7.2lf\t%7.2lf\t%7.2lf\n",alpha[0][i],alpha[1][i],alpha[2][i],alpha[3][i]); */
  }

  /* termination */
  maxp = -FLT_MAX;
  for(j=0;j<NUMSTATES;j++) {
    h = pextra(j);
    f = mygamma[j][i-1] + h;
    alpha[j][slen] = alpha[j][slen-1] + h;

    maxq = -FLT_MAX;
    switch(j) {
      case INTERGENE:
      maxq = endforward(slen,maxq,option);
      break;

      case I0:
      case I1:
      case I2:
      maxq = donorforward(slen,j,maxq,option);
      break;

      default:
      fprintf(stderr,"Error in viterbi (termination part): bad state\n");
      break;
    }
    if(f > maxq) {
      remember[j][slen].state = j;
      remember[j][slen].loc = slen-1;
      remember[j][slen].type = NADA;
      mygamma[j][slen] = f;
    } else {
      mygamma[j][slen] = maxq;
    }
    if(mygamma[j][slen] > maxp) {
      maxp = mygamma[j][slen];
      last_state = j;
    }
  }
  
  /*
  ** Compute the likelihood.
  ** This actually computes P(Y_1,...,Y_T, X_T = some intron state) which is just 
  ** a little different to the usual likelihood computed, P(Y_1,...,Y_T)
  */
  for(likelihood=alpha[0][slen],j=1; j<NUMSTATES; j++)
    myadd(alpha[j][slen],likelihood);

  /* printf("######\n######\nEND OF FORWARD ALGORITHM\n######\n######\n"); */

  return last_state;
}

/*
** Calculate beta, the backward variables as in the usual Rabiner notation
*/
void backward(Option *option) {
  int i,j;
  double test1,test2;

  /* initialization */
  for(j=0;j<NUMSTATES;j++) {
    beta[j][slen] = 0;
    beta[j][slen-1] = pextra(j);
  }

  /* iteration */
  for(i=slen-2; i >= -1; i--) {
    for(j=0;j<NUMSTATES;j++) {
      /* probability if continuing state  */
      beta[j][i] = pextra(j) + beta[j][i+1] + bIntronIntergene(i+1,j,slen);
  /* printf("beta[%d][%d] = %2.15lf (before exons)\n",j,i,beta[j][i]); */

      switch(j) {
        case INTERGENE:
	startforward(i+1,option);
        break;

        case I0:
        case I1:
        case I2:
	acceptorforward(i+1,j,option);
        break;

        default:
        fprintf(stderr,"Error in backward: bad state\n");
        break;
      }
  /* printf("beta[%d][%d] = %2.15lf (after exons)\n\n",j,i,beta[j][i]); */
    }
  }

  /* as a check, the following 2 quantities should be equal...
  ** sum_j (exp(beta[j][0] + alpha[j][0]))
  ** sum_j alpha[j][slen]
  */
  if(option->verbose) {
    test1 = beta[0][0] + alpha[0][0];
    test2 = alpha[0][slen];
    for(j=1; j<NUMSTATES; j++) {
      myadd(beta[j][1] + alpha[j][1],test1);
      myadd(alpha[j][slen],test2);
    }
    printf("likelihood from alphas = %2.15f\n",test1);
    printf("likelihood from betas  = %2.15f\n",test2);
  }

  if(option->verbose) printf("\n");
  
  return;
}



/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible terminal and single exons.
*/
double endforward(int pos,double maxp,Option *option) {
  int i,j,phase,seqLen,exonLen,minLen;
  double p;
  double ap;

  if(pos != fTermEnd[curfte] || curfte >= fte)
    return -FLT_MAX;
  else
    curfte++;
  /* pos points to the first base _after_ the stop codon */

  /* Sum over all possible single exons ending at pos-1 */
  minLen = START_SIGNAL_LEN + STOP_SIGNAL_LEN+1;
  for(i=0;i<fis;i++) {
    j = fInitSt[i];
    seqLen = pos-j;

    /* skip to next if exon too long */
    if( seqLen > maxExonLen(TYPE_SNGL)) continue;

    /* stop if the single exon is too short (as all the rest will be too) */
    if( seqLen < minLen) break;

    /* skip if length not a multiple of 3 */
    if((pos-j) % PHASES != 0) continue;

    phase = 0;
    p = exonProb(j,pos-1,phase,TYPE_SNGL,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,SINGLE) = %2.15f\n",j,pos-1,phase,p);
    /*
    printf("for the whole exon, scored %2.15lf\n",p);
    */
    /* the pnotextra below is to account for not having another self-transition 
       in the intron state */
    /*
    printf("ptrans(intergene,intergene) = %2.15lf\n",ptrans(INTERGENE,INTERGENE));
    printf("pnotextra(intergene) = %2.15lf\n",pnotextra(INTERGENE));
    */
    p += ptrans(INTERGENE,INTERGENE) + pnotextra(INTERGENE);
    /*
    printf("  with transitions, %2.15lf\n",p);
    */

    ap = p + alpha[INTERGENE][j-1];
    myadd(ap,alpha[INTERGENE][pos]);

    p += mygamma[INTERGENE][j-1];
    if(p > maxp) {
      maxp = p;
      (remember[INTERGENE][pos]).state=INTERGENE;
      (remember[INTERGENE][pos]).loc=j-1;
      (remember[INTERGENE][pos]).type=ESING;
    }
  }
  /*
  printf("alpha[INTERGENE][%d] is now %2.15lf\n",pos,alpha[INTERGENE][pos]);
  */

  /* Sum over all possible terminal exons ending at pos-1 */
  minLen = acc_upstr + acc_dnstr + STOP_SIGNAL_LEN + 1;
  for(i=0;i<fos;i++) {
    j = fOtherSt[i];
    seqLen = pos-j;

    /* skip if the terminal exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM) ) continue;

    /* stop if the terminal exon would be too short */
    if( seqLen < minLen) break;

    exonLen = seqLen-acc_upstr;
    phase = (PHASES - (exonLen % PHASES) ) % PHASES;  /* phase is phase of this exon */
    p = exonProb(j,pos-1,phase,TYPE_TERM,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,TERMINAL) = %2.15f\n",j,pos-1,phase,p);
    /*
    printf("exonLen = %d\n",exonLen);
    printf("for the whole exon, scored %2.15lf\n",p);
    */
    /* the pnotextra below is to account for not having another self-transition 
       in the intron state */
    /*
    printf("ptrans(I0+%d,intergene) = %2.15lf\n",phase,ptrans(I0+phase,INTERGENE));
    printf("pnotextra(I0+%d) = %2.15lf\n",phase,pnotextra(I0+phase));
    */
    p += ptrans(I0+phase,INTERGENE) + pnotextra(I0+phase);
    /*
    printf("  with transitions, %2.15lf\n",p);
    */

    /*
    printf("adding %2.15lf to %2.15lf\n",p+alpha[I0+phase][j-1],alpha[INTERGENE][pos]);
    */
    ap = p + alpha[I0+phase][j-1];
    myadd(ap,alpha[INTERGENE][pos]);

    p += mygamma[I0+phase][j-1];
    if(p > maxp) {
      maxp = p;
      remember[INTERGENE][pos].state=I0+phase;
      remember[INTERGENE][pos].loc=j-1;
      remember[INTERGENE][pos].type=TERM0+phase;
    }
  }

  /*
  printf("alpha[INTERGENE][%d] is now %2.15lf\n",pos,alpha[INTERGENE][pos]);
  */
  return maxp;
}


/*
** Required for Viterbi algorithm and for calculation of forward variables.
** Sums over possible initial and internal exons.
*/
double donorforward(int pos,int state,double maxp,Option *option) {
  int i,j,phase,seqLen,exonLen,minLen;
  double p;
  double ap;

  if( (pos != fOtherEnd[curfoe]) || (curfoe > foe) )
    return -FLT_MAX;
  if(state == I2)
    curfoe++;
  /* pos now points to the first base _after_ the end of the donor signal */

  /* Sum over all possible initial exons */
  minLen = don_upstr+don_dnstr+START_SIGNAL_LEN+1;
  for(i=0;i<fis;i++) {
    j = fInitSt[i];
    seqLen = pos-j;

    /* skip if the initial exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT) ) continue;

    /* stop if the initial exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen - don_dnstr; 
    /* skip if this exon not the right length */
    if( (exonLen % PHASES) != (state - I0) ) continue;

    phase = 0;
    p = exonProb(j,pos-1,phase,TYPE_INIT,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,INITIAL) = %2.15f\n",j,pos-1,phase,p);
    p += ptrans(INTERGENE,state) + pnotextra(INTERGENE);

    ap = p + alpha[INTERGENE][j-1];
    myadd(ap,alpha[state][pos]);

    p += mygamma[INTERGENE][j-1];
    if(p > maxp) {
      maxp = p;
      remember[state][pos].state=INTERGENE;
      remember[state][pos].loc=j-1;
      remember[state][pos].type=EINIT;
    }
  }

  /* Sum over all possible internal exons */
  minLen = acc_upstr+acc_dnstr+don_upstr+don_dnstr+1;
  for(i=0; i<fos;i++) {
    j = fOtherSt[i];
    seqLen = pos - j;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the internal exon would be too short */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    phase = ( (state-I0) + PHASES - (exonLen % PHASES)) % PHASES;
    p = exonProb(j,pos-1,phase,TYPE_INTL,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,INTERNAL) = %2.15f\n",j,pos-1,phase,p);
    p += ptrans(I0+phase,state) + pnotextra(I0+phase);

    ap = p + alpha[I0+phase][j-1];
    myadd(ap,alpha[state][pos]);

    p += mygamma[I0+phase][j-1];
    if(p > maxp) {
      maxp = p;
      remember[state][pos].state=I0+phase;
      remember[state][pos].loc=j-1;
      remember[state][pos].type=E0+phase;
    }
  }
  return maxp;
}


/*
** Required for calculation of backward variables.
** Sums over possible initial and single exons.
*/
void startforward(int pos, Option *option) {
  int seqLen,exonLen,minLen,nextState,phase;
  int i,j;
  double p;

  if( (pos != fInitSt[curfis]) || (curfis < 0) )
    return;
  else
    curfis--;

  /* sum over all possible single exons */
  minLen = START_SIGNAL_LEN+STOP_SIGNAL_LEN+1;
  for(i=fte-1; i>=0; i--) {
    j = fTermEnd[i];   /* j points to the first base _after_ the stop codon */
    exonLen = j-pos;

    /* skip if single exon is too long */
    if( exonLen > maxExonLen(TYPE_SNGL) ) continue;

    /* stop if the single exon is too short (all the rest will be too) */
    if( exonLen < minLen ) break;

    /* skip if length not a multiple of 3 */
    if((exonLen % PHASES) != 0) continue;

    phase = 0;
    p  = exonProb(pos,j-1,phase,TYPE_SNGL,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,SINGLE) = %2.15f\n",pos,j-1,phase,p);
    p += pnotextra(INTERGENE) + ptrans(INTERGENE,INTERGENE);
    p += beta[INTERGENE][j];
    if(j < slen)
      p += bIntronIntergene(j,INTERGENE,slen);

    myadd(p,beta[INTERGENE][pos-1]);
  }

  /* sum over all possible initial exons */
  minLen = START_SIGNAL_LEN + don_upstr + don_dnstr + 1;
  for(i=foe-1; i>=0; i--) {
    j = fOtherEnd[i];   /* j points to the first base _after_ the end of the donor signal */
    seqLen = j-pos;

    /* skip if the initial exon is too long */
    if( seqLen > maxExonLen(TYPE_INIT) ) continue;

    /* stop if the initial exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-don_dnstr;

    phase = 0;
    nextState = I0 + (exonLen % PHASES);
    p  = exonProb(pos,j-1,phase,TYPE_INIT,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,INITIAL) = %2.15f\n",pos,j-1,phase,p);
    p += pnotextra(INTERGENE) + ptrans(INTERGENE,nextState);
    p += beta[nextState][j];
    if(j < slen)
      p += bIntronIntergene(j,nextState,slen);

    myadd(p,beta[INTERGENE][pos-1]);
  }

  return;
}


/*
** Required for calculation of backward variables.
** Sums over possible internal and terminal exons.
*/
void acceptorforward(int pos, int state, Option *option) {
  int exonLen,seqLen,minLen,nextState,phase;
  int i,j;
  double p;

  if( (pos != fOtherSt[curfos]) || (curfos < 0) )
    return;
  else if (state == I2)
    curfos--;

  phase = (state-I0);  /* the phase in which we'll begin */

  /* sum over all possible internal exons */
  minLen = don_upstr+don_dnstr+acc_upstr+acc_dnstr+1;
  for(i=foe-1; i>=0; i--) {
    j = fOtherEnd[i];   /* j points to the first base _after_ the stop codon */
    seqLen = j-pos;

    /* skip if the internal exon is too long */
    if( seqLen > maxExonLen(TYPE_INTL) ) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr-don_dnstr;
    nextState = I0 + ((phase + exonLen) % PHASES);

    p = exonProb(pos,j-1,phase,TYPE_INTL,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,INTERNAL) = %2.15f\n",pos,j-1,phase,p);
    p += pnotextra(state) + ptrans(state,nextState);
    p += beta[nextState][j];
    if(j < slen)
      p += bIntronIntergene(j,nextState,slen);

    myadd(p,beta[state][pos-1]);
  }

  /* sum over all possible terminal exons */
  minLen = acc_upstr+acc_dnstr+STOP_SIGNAL_LEN+1;
  for(i=fte-1; i>=0; i--) {
    j = fTermEnd[i];   /* j points to the first base _after_ the stop codon */
    seqLen = j-pos;

    /* skip if terminal exon is too long */
    if( seqLen > maxExonLen(TYPE_TERM)) continue;

    /* stop if the exon is too short (all the rest will be too) */
    if( seqLen < minLen ) break;

    exonLen = seqLen-acc_upstr;
    /* skip to the next stop codon if this one isn't the right length */
    if( 0 != ((phase+exonLen) % PHASES)) continue;

    p  = exonProb(pos,j-1,phase,TYPE_TERM,seq,option);
    if(option->debug) printf("  NOTE: total for exonProb(%d,%d,%d,TERMINAL) = %2.15f\n",pos,j-1,phase,p);
    p += pnotextra(state) + ptrans(state,INTERGENE);
    p += beta[INTERGENE][j];
    if(j < slen) {
      p += bIntronIntergene(j,INTERGENE,slen);
      /* printf("adding bIntonIntergene(%d,INTERGENE) = %2.15lf\n",j,bIntronIntergene(j,INTERGENE,slen)); */
    }

    /*
    printf("adding %2.15lf for probs\n",pnotextra(state) + ptrans(state,INTERGENE));
    printf("adding beta[INTERGENE,%d] = %2.15lf\n",j,beta[INTERGENE][j]);
    printf("total is %2.15lf, adding that to %2.15lf\n",p,beta[state][pos-1]);
    */
    myadd(p,beta[state][pos-1]);
  }

  return;
}


/* 
** backtracking part of the viterbi algorithm.  The backtrack will start 
** from state startstate in position startpos.  If startpos is <= 0 then
** it will start from the last state.
*/
void doBackTrack(int startstate, int startpos, Gene *gene, int *nGenes, Option *option) {
  short type;
  int pos;
  int state;
  int oldstate;
  int oldpos;
  int i;
  double x;

  x = alpha[0][slen];
  for(i=1;i<NUMSTATES;i++) {
    myadd(alpha[i][slen],x);
  }
  seqProb = x;

  /* start from last state if startpos <= 0 */
  if(startpos <= 0)
    startpos = slen;

  for(state=startstate,pos=startpos,oldpos=startpos;;) {
    /* printf("in Backtrack: pos = %d\n",pos); */
    if(pos == -1) {
      if(oldpos != -1) {
        fprintf(stdout, "%d\n",pos+2);
      }
      break;
    }

    /* skip over intron, ie skip states of duration length 1 */
    oldstate = state;
    oldpos = pos;
    state = remember[state][pos].state;
    pos = remember[oldstate][pos].loc;
    type = remember[oldstate][oldpos].type;
    while(oldpos == pos+1) {
      oldstate = state;
      oldpos = pos;
      state = remember[state][pos].state;
      pos = remember[oldstate][pos].loc;
      type = remember[oldstate][oldpos].type;
    }
    if(oldpos != -1) {
      /* printBestExon(oldstate,oldpos); */
      if((*nGenes) < MAX_GENES) {
        gene = storeGene(gene,oldstate,oldpos,nGenes,option);
      } else {
        fprintf(stderr,"Warning: ran out of space for storing genes\n");
        fprintf(stderr,"  Have space for %d in total\n",MAX_GENES);
        fprintf(stderr,"  (Maybe recompile with higher MAX_GENES value)\n");
        return;
      }
    }
  }
  /* must increment the gene count if we finished with part of a gene */
  if( (*gene).nExons != 0)
    (*nGenes)++;

  /* printf("about to exit doBacktrack\n"); */
  return;
}


/* 
** print out the details of the current exon
*/
void printBestExon(int i,int t) {
  int start,stop,prev,cur,type,phase,frame;
  char name[30];

  /* compute everything zero-based indices first */
  start = remember[i][t].loc + 1;
  stop  = t-1;
  cur   = i;
  prev  = remember[i][t].state;
  type  = remember[i][t].type;

  if(type == NADA)
    return;

  if(type == ESING) {
    strcpy(name,"Single   ");
    phase = 0;
  } else if(type == EINIT) { 
    strcpy(name,"Initial  ");
    phase = 0;
    stop -= don_dnstr;
  } else if ((type >= E0) && (type <= E2)) {
    strcpy(name,"Internal ");
    phase = type-E0;
    stop -= don_dnstr;
    start += acc_upstr;
  } else {
    strcpy(name,"Terminal ");
    phase = type-TERM0;
    start += acc_upstr;
  }
  frame = ((start + phase) % PHASES);

  printf("%s %6d -> %6d,  %4d bp   ph %d  fr %d  (st %d to %d)\n",
    name,start+1,stop+1,stop-start+1,phase,frame,prev,cur);
}


/*
** Records the exon in a Gene structure.  geneCount will be incremented by one
** and gene will be set to point to the next element in the Gene array if we
** add in a single or an initial exon.
*/
Gene *storeGene(Gene *gene, int i, int t, int *geneCount, Option *option) {
  int phase,thisIndex;
  int prevState,nextState;
  int peb;                   /* prevExtraBases */
  int type;                  /* one of TYPE_SNGL,TYPE_INIT,TYPE_INTL,TYPE_TERM */
  int thisExonType;          /* one of NADA,ESING,EINIT,E0,E1,E2,TERM0,TERM1,TERM2 */
  int start,stop;            /* these coordinates include the splice site signals, for scoring */
  int exonStart,exonStop;    /* these are for the start and stop of the actual exon */
  double thisExonProb;
  double part1,part2,part3,part4,allparts;

  /* compute everything zero-based indices first */
  prevState = remember[i][t].state;
  nextState = i;
  start = remember[i][t].loc + 1;
  stop  = t-1;
  thisExonType  = remember[i][t].type;

  if(thisExonType == NADA) {
    fprintf(stderr,"storeGene: request to store intergenic region as gene!\n");
    exit(1);
  }

  if(thisExonType == ESING) {
    peb       = 0;
    exonStart = start;
    exonStop  = stop;
    type      = TYPE_SNGL;
  } else if(thisExonType == EINIT) {
    peb       = 0;
    exonStart = start;
    exonStop  = stop - don_dnstr;
    type      = TYPE_INIT;
  } else if ((thisExonType >= E0) && (thisExonType <= E2)) {
    peb       = thisExonType-E0;
    exonStart = start + acc_upstr;
    exonStop  = stop - don_dnstr;
    type      = TYPE_INTL;
  } else {
    peb       = thisExonType-TERM0;
    exonStart = start + acc_upstr;
    exonStop  = stop;
    type      = TYPE_TERM;
  }
  phase = (PHASES-peb) % PHASES;

  /* get the probability this is a real exon */
  part1 = alpha[prevState][start-1];
  part2 = pnotextra(prevState) + ptrans(prevState,nextState);
  part3 = exonProb(start,stop,peb,type,seq,option);
  part4 = beta[nextState][stop];
  allparts  = part1+part2+part3+part4;
  thisExonProb  = exp(allparts-likelihood);

  if( option->verbose && ((thisExonProb < 0) || (thisExonProb > 1)) ) {
    fprintf(stderr,"Warning: computed an invalid exon probability!\n");
    fprintf(stderr,"  Exon [%d,%d], type %d\n",exonStart,exonStop,thisExonType);
    fprintf(stderr,"  leftState  =  %d\n",prevState);
    fprintf(stderr,"  rightState =  %d\n",nextState);
    fprintf(stderr,"  prob     = %f\n",thisExonProb);
    fprintf(stderr,"  part1    = %f\n",part1);
    fprintf(stderr,"  part2    = %f\n",part2);
    fprintf(stderr,"  part3    = %f\n",part3);
    fprintf(stderr,"  part4    = %f\n",part4);
    fprintf(stderr,"  allparts = %f\n",allparts);
    fprintf(stderr,"  likelihood = %f\n",likelihood);
    thisExonProb = (thisExonProb > 1) ? 1 : 0;
  }


  if((*gene).nExons >= MAX_EXONS) {
    fprintf(stderr,"Warning: storeGene: ran out of space for exon storage\n");
    fprintf(stderr,"  Can store at most %d\n",MAX_EXONS);
    fprintf(stderr,"  (Maybe recompile with higher value of MAX_EXONS)\n");
    return gene;
  } else {
    thisIndex = (*gene).nExons;
    (*gene).start[thisIndex] = option->offset + exonStart;
    (*gene).stop[thisIndex]  = option->offset + exonStop;
    (*gene).frame[thisIndex] = (exonStart + option->offset + phase) % PHASES;
    (*gene).phase[thisIndex] = phase;
    (*gene).type[thisIndex]  = type;
    (*gene).prob[thisIndex]  = thisExonProb;
    (*gene).nExons++;
  }

  if( (thisExonType == EINIT) || (thisExonType == ESING) ){
    (*geneCount)++;
    return gene+1;
  } else {
    return gene;
  }
}

/* 
** returns
** 0 if c is a or A
** 1 if c is c or C
** 2 if c is g or G
** 3 if c is t or T
** 4 otherwise
*/
int base2num(int c) {

  switch(c) {
    case 'a':
    case 'A':
      return 0;
    break;

    case 'c':
    case 'C':
      return 1;
    break;

    case 'g':
    case 'G':
      return 2;
    break;

    case 't':
    case 'T':
      return 3;
    break;

    default:
      return 4;
    break;
  }

}

char *getGenomicSeqPointer(void) {
  return(seq);
}


/* 
** Write out the marginal probabilities of each base being some kind of exon
** for the forward pass.
*/
void writeFwdProbFile(int nBases, Option *opt) {
  FILE *outProbFile = NULL;
  int t,i;
  double numerator,prob,lastProb,diff;

  if((outProbFile = fopen(opt->outFwdProbFileName,"w")) == NULL) {
    fprintf(stderr,"writeProbFile: can't open %s for write\n",opt->outFwdProbFileName);
    exit(1);
  }
  lastProb = -1;
  diff = pow(10.0, (double)(-(PROB_SIG_FIG+1)));

  for(t=0; t < nBases; t++) {
    numerator = alpha[0][t] + beta[0][t];
    for(i=1; i < NUMSTATES; i++) {
      myadd(alpha[i][t] + beta[i][t], numerator);
    }
    prob = round(1-exp(numerator-likelihood),PROB_SIG_FIG);

    /* this should be impossible, but just in case... */
    if(prob < 0)
      prob = 0;
    else if(prob > 1)
      prob = 1;

    if(opt->oldprb) {
      fprintf(outProbFile,"%g\n",prob);
    } else {
      if( (fabs(prob-lastProb) > diff) || (t == (nBases-1)) ) {
        fprintf(outProbFile,"%d %g\n",t+1,prob);
        lastProb = prob;
      }
    }
  }
  fclose(outProbFile);

  return;
}


/* 
** Write out the marginal probabilities of each base being some kind of exon
** for the reverse pass.
*/
void writeBakProbFile(int nBases, Option *opt) {
  FILE *outProbFile = NULL;
  int t,i;
  double numerator,prob,lastProb,diff;

  if((outProbFile = fopen(opt->outBakProbFileName,"w")) == NULL) {
    fprintf(stderr,"writeProbFile: can't open %s for write\n",opt->outBakProbFileName);
    exit(1);
  }
  lastProb = -1;
  diff = pow(10.0, (double)(-(PROB_SIG_FIG+1)));

  for(t=nBases-1; t >= 0; t--) {
    numerator = alpha[0][t] + beta[0][t];
    for(i=1; i < NUMSTATES; i++) {
      myadd(alpha[i][t] + beta[i][t], numerator);
    }
    prob = round(1-exp(numerator-likelihood),PROB_SIG_FIG);

    /* this should be impossible, but just in case... */
    if(prob < 0)
      prob = 0;
    else if(prob > 1)
      prob = 1;

    if(opt->oldprb) {
      fprintf(outProbFile,"%g\n",prob);
    } else {
      if( (fabs(prob-lastProb) > diff) || (t == 0) ) {
        fprintf(outProbFile,"%d %g\n",nBases-t,prob);
        lastProb = prob;
      }
    }
  }
  fclose(outProbFile);

  return;
}


void testFunction(void) {
  int t,i,j;
  double sum,prob;
  FILE *testFile;
  FILE *alphaFile;
  FILE *betaFile;

  testFile = fopen("test.dat","w");
  if(testFile == NULL) {
    fprintf(stderr,"testFunction: can't open output file\n");
    exit(EXIT_FAILURE);
  }
  for(t=0; t<slen; t++) {
    /* sum = alpha[0][t] + beta[0][t] - bIntronIntergene(t,0,slen); */
    sum = alpha[0][t] + beta[0][t];
    for(i=1; i<NUMSTATES; i++) {
      /* myadd(alpha[i][t] + beta[i][t] - bIntronIntergene(t,i,slen),sum); */
      myadd(alpha[i][t] + beta[i][t],sum);
    }
    prob = exp(sum-likelihood);
    fprintf(testFile,"%f\t%f\t%f\n",sum,likelihood,prob);
  }
  fclose(testFile);
  printf("Wrote test.dat\n");

  alphaFile = fopen("alpha.dat","w");
  for(t=-1; t<=slen; t++) {
    for(i=0; i<NUMSTATES; i++)
      fprintf(alphaFile,"%2.10f\t",alpha[i][t]);
    fprintf(alphaFile,"\n");
  }
  fclose(alphaFile);
  printf("Wrote alpha.dat\n");

  betaFile = fopen("beta.dat","w");
  for(t=-1; t<=slen; t++) {
    for(i=0; i<NUMSTATES; i++)
      fprintf(betaFile,"%2.10f\t",beta[i][t]);
    fprintf(betaFile,"\n");
  }
  fclose(betaFile);
  printf("Wrote beta.dat\n");

  printf("\n");
  printf("Transitions...\n");
  for(i=0; i<NUMSTATES; i++) {
    for(j=0; j<NUMSTATES; j++)
      printf("%2.15f  ",exp(ptrans(i,j)));
    printf("\n");
  }

  printf("\n");
  for(i=0; i<NUMSTATES; i++)
    printf("pself[%d] = %2.15f\n",i,1-exp(pnotextra(i)));

  printf("\n");
  for(i=0,sum=0; i<NUMSTATES; i++) {
    sum += exp(pinit(i));
    printf("pinit[%d] = %2.15f   (%2.15f)\n",i,exp(pinit(i)),pinit(i));
  }
  printf("  sum(pinit(i)) = %f\n",sum);

  return;
}

/*
 * Simple function to round x to n significant figures.
 */
double round(double x, int n) {
  int xInt,rem;

  xInt = (int) (x * pow(10.0,(double)(n+1)));
  if(xInt > 0) {
    rem = xInt - 10 * (int) floor(xInt/10);
    if(rem > 4) ++xInt;
  } else {
    rem = 10 * (int) floor(xInt/10) - xInt;
    if(rem > 4) --xInt;
  }

  return(xInt/pow(10.0,(double)n));
}
