#!/usr/local/bin/perl -w
##
## A wrapper script for halfphat/fullphat.
##
## Usage:
##   phat.pl [ options ] seq1 seq2 ... seqN
##
## Option Argument     Effect
## -v     (none)       Passes verbose option on to doublephat
## -h     (none)       Prints this help screen
## -type  half/full    Specifies halfphat/fullphat (default = fullphat)
## -ph    \"options\"    Pass options on to halfphat/fullphat (default = -noprob)
## -cpu   nCPUs        The assumed number of CPUs (default = $nCPUs)
## -org   [pf/hum]     Organism.
##

########################################################################
########################################################################
##                                                                    ##
## Phat - a system for predicting gene structures in DNA.             ##
## Copyright (C) 2000 Simon Cawley and Anthony Wirth                  ##
##                                                                    ##
## This program is free software; you can redistribute it and/or      ##
## modify it under the terms of the GNU General Public License        ##
## as published by the Free Software Foundation; either version 2     ##
## of the License, or (at your option) any later version.             ##
##                                                                    ##
## This program is distributed in the hope that it will be useful,    ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of     ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ##
## GNU General Public License for more details.                       ##
##                                                                    ##
## You should have received a copy of the GNU General Public License  ##
## along with this program; if not, write to the Free Software        ##
## Foundation, Inc.,                                                  ##
## 59 Temple Place - Suite 330,                                       ##
## Boston, MA  02111-1307, USA.                                       ##
##                                                                    ##
########################################################################
########################################################################



use File::Basename;
use Getopt::Long;
use Bio::Seq;
use Bio::SeqIO;


$BIN   = ".";       # directory with halfphat/fullphat binaries
$LIB   = "./Pars";  # directory with parameters
$nCPUs = 1;    # the number of CPUs on the system

$halfPhatExe   = "$BIN/halfphat";
$fullPhatExe   = "$BIN/fullphat";

$phatPars = {
  "pf"  => "$LIB/Plasmodium.falciparum",
  "pv"  => "$LIB/Plasmodium.vivax",
  "hum" => "$LIB/Homo.sapiens",
};

$tempExt  = "phat.temp.fasta";         # extension for temporary files

#
# The %formats hash defines how file extensions imply formats
# on your system.  You may want to edit this according to your
# system's naming conventions.  In the case of an unrecognized
# format or no extension, $defaultFormat will be assumed.
#
%formats = (
  Fasta   => [ "fasta", "fa" ],
  GenBank => [ "genbank", "gbk", "gb" ],
  GCG     => [ "gcg" ],
);
$defaultFormat = "Fasta";



##
## You will probably only need to alter stuff above this line
##


# make sure we can execute the required programs
&checkExes($halfPhatExe,$fullPhatExe);

$option{"v"}     = 0;
$option{"h"}     = 0;
$option{"type"}  = "full";
$option{"ph"}    = "";
$option{"org"}   = "";        # You can put in a default organism here if you want.
$option{"cpu"}   = $nCPUs;
if (!(&GetOptions(\%option,"v","h","org=s","type=s","ph=s","cpu=s"))) {
  &usage();
  die("Error: problem processing command-line options\n");
}

# help option
if($option{"h"}) {
  &usage();
  exit(0);
}

if(scalar(@ARGV) <= 0) {
  &usage();
  die("Error: you need to specify some sequence files to work on\n");
}

# sets phatExe to "half" or "full"
$option{"type"} = lc($option{"type"});
if($option{"type"} eq "half") {
  $phatExe = $halfPhatExe;
} elsif ($option{"type"} eq "full") {
  $phatExe = $fullPhatExe;
} else {
  &usage();
  die "Error: -type option must specify either \"half\" or \"full\"\n";
}
  
# check on the number of cpus
if(! ($option{"cpu"} =~ /^([1-9]\d*)$/)) {
  &usage();
  die "Error: -cpu option must specify a positive integer\n";
} else {
  $nCPUs = $option{"cpu"};
}

# Ensure an organism was specified.
if($option{"org"} eq "") {
  warn "You must specify an organism with the -org option.\n";
  warn "Available types are:\n";
  foreach $type (sort keys %$phatPars) {
    printf STDERR ("  -org %4s      (uses parameters in %s)\n",$type,$phatPars->{$type});
  }
  die;
}

# verify we recognize the organism
if(! exists($phatPars->{$option{"org"}})) {
  die "Error: invalid organism type: " . $option{"org"} . "\n";
}

# options that get passed through to halfphat/fullphat
if($option{"ph"} ne "") {
  print "Passing on options \"" . $option{"ph"} . "\" to $phatExe\n";
}


#
# analyze the sequences
#
%analysed = (); # hash of files already processed
@jobList = ();
$batch = 1;
foreach $seqFile (@ARGV) {
  # Determine sequence format and the number of sequences.
   @fileParts = fileparse($seqFile,'\..*');
  ($base,$ext) = ($fileParts[0],$fileParts[2]);
  $format = &getFormat($ext,\%formats,"$base$ext",$defaultFormat);
  $seqin = Bio::SeqIO->new( '-format' => $format , -file => $seqFile);
  $nSeq=0;
  while($seqobj = $seqin->next_seq()) { $nSeq++; }
  $seqin->close;

  $seqin = Bio::SeqIO->new( '-format' => $format , -file => $seqFile);
  while($seqobj = $seqin->next_seq()) {
    # make sure there is some sequence to analyze
    if(! defined($seqobj->seq)) {
      print "  $base$ext contains no sequence\n";
      next;
    }

    # Determine $outBase, base name for phat output files
    $seqId = $seqobj->display_id;
    ($seqId) = ($seqId =~ /^(\S+)\s*/);
    $outBase = ($nSeq > 1) ? "$base.$seqId" : $base;
    $tempFile = "$outBase.$tempExt";

    # Make sure we haven't already written this file
    if(exists($analysed{$outBase})) {
      print STDERR "\nWarning: already wrote output for $outBase.\*\n";
      print STDERR "  (Not processing $seqFile)\n\n";
    } else {
      $analysed{$outBase} = 1;

      # extract next sequence, write it to temporary single-fasta file
      $seqout= Bio::SeqIO->new( '-format' => 'Fasta', -file => ">$tempFile" );
      $seqout->write_seq($seqobj);
      $seqout->close;
  
      # if we've dispatched $nCPUs jobs then wait for them all to finish
      if(scalar(@jobList) == $nCPUs) {
        foreach $thing (@jobList) {
          ($pid,$oldTempFile) = @$thing;
          waitpid($pid,0);
          unlink($oldTempFile);
          #print "done with job $pid, $oldTempFile\n";
        }
        print "Finished batch $batch\n" if($nCPUs > 1);
        $batch++;
        @jobList = ();
      }

      # run phat
      $command = "$phatExe -p " . $phatPars->{$option{"org"}};
      $command .=  " \"$tempFile\" -o \"$outBase\" -seq \"$seqId\" ";
      $command .= $option{"ph"} . " 1> /dev/null";
      print "  Analyzing $outBase\n";
      &forkJob($command,$tempFile,\@jobList);
    }
  }

}

# wait for the final processes to end and delete their temp files
foreach $thing (@jobList) {
  ($pid,$oldTempFile) = @$thing;
  waitpid($pid,0);
  #print "done with job $pid, $oldTempFile\n";
  unlink($oldTempFile);
}
print "Finished batch $batch\n" if($nCPUs > 1);
exit(0);




sub forkJob {
  my($command) = shift;
  my($tempFile) = shift;
  my($jobList) = shift;

  if($pid = fork) {
    push(@$jobList,[$pid, $tempFile]);
    #print "starting job $pid, $tempFile\n";
  } elsif (defined $pid) {
    exec $command;
  } elsif ($! =~ /No more process/) {
    print "Should redo fork\n";
  } else {
    die "Can't fork: $!\n";
  }
}
    
    
exit(0);




##
## Return the format implied by the extension $thisExt, according to the
## formats defined in \%formats
##
sub getFormat {
  my($thisExt) = shift;
  my($formats) = shift;
  my($fileName) = shift;
  my($default) = shift;
  my($found,$format,$ext,$ans);

  $found = 0;
  foreach $format (keys %$formats) {
    foreach $ext (@{$formats->{$format}}) {
      if($thisExt eq ".$ext") {
        $found = 1;
        $ans = $format;
        last;
      }
    }
    last if($found);
  }
  
  if(! $found) {
    print STDERR "\nWarning: can't determine sequence format from name $fileName\n";
    print STDERR "  Assuming Fasta format for now, if that's not correct you might\n";
    print STDERR "  consider altering the extension conventions in $0\n\n";
    $ans = $default;
  }

  return($ans);
}

##
## Verify we have execute authority on the names files
##
sub checkExes {
  my(@files) = @_;
  my($problem) = 0;

  foreach $file (@files) {
    if(! -x $file) {
      print STDERR "Can't execute $file\n";
      $problem = 1;
    }
  }
  
  die if($problem);
  return;
}


#
# Usage function - will print out commented lines at the top of this 
# file until it encounters a non-commented line.  (Skips the first.)
#
sub usage {
  open(MYSELF,$0) || die "$0: can't open $0 to print usage!\n";
  while(<MYSELF>) {
    if(/^\#!/) {
      next;
    } elsif (/^\s*\#(.*)$/) {
      eval "print STDERR \"$1\n\"";
    } else {
      last;
    }
  }
  close MYSELF;
}
